/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017,2018,2021,2023 Harald Foerster	*/
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <glib/gstdio.h>

#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#else
#include "getopt.h"
#endif

#include "cff_macros-gtk.h"
#include "cff_accel-gtk.h"
#include "cff_gprop.h"
#include "cff_nls.h"
#include "cff_charconv.h"

#ifdef G_OS_WIN32
#include <fcntl.h>
#include <windows.h>
#endif

/*
#include <gtk/gtk.h>
#include "cff_tools-gtk.h"
*/


#define kPrintVersion	(1 << 0)
#define kPrintHelp	(1 << 1)


enum
{
	CFF_FIRST,
	CFF_MENUACCEL = CFF_FIRST,
	CFF_PROP,
	CFF_LAST = CFF_PROP
};

const gchar *gCffMainCutterffPackageString = PACKAGE;
const gchar *gCffMainCutterffVersionString = VERSION;

static const char gCffMainCutterffRsrcFileString[] =
#ifdef G_OS_WIN32
			"cutterff.rc";
#else
			".cutterffrc";
#endif

static const char *gCffMainCfgRsrcNameString[] =
{
	"MenuAccel:",
	"Property:"
};

#if GTK_MAJOR_VERSION > 2
static GSList *gMainAccelSlist = NULL;
#endif

#ifdef G_OS_WIN32
/* Prototyp */
static int cff_main_open_console(void);
#endif


#ifdef G_OS_WIN32

#if GLIB_MAJOR_VERSION == 2 && GLIB_MINOR_VERSION >= 40

#define cff_main_win32_get_command_line g_win32_get_command

#else

static gchar **cff_main_win32_get_command_line(void)
{
	/* Stolen from 'glib-2.51.1' */

	gchar  **result;
	LPWSTR *args;
	gint   i, n;

	args   = CommandLineToArgvW(GetCommandLineW(), &n);
	result = g_malloc((n + 1) * sizeof(gchar*));

	for (i = 0; i < n; i++)
	{
		result[i] = g_utf16_to_utf8(args[i], -1, NULL, NULL, NULL);
	}

	result[i] = NULL;

	LocalFree(args);

	return result;

} /* static gchar **cff_main_win32_get_command_line(void) */

#endif /* #if GLIB_MAJOR_VERSION == 2 && GLIB_MINOR_VERSION >= 40 .. #else */

#endif /* G_OS_WIN32 */


static void cff_main_fprn_stderr(int parse_err, const char *detail)
{
	const char *format, *mesg;

#ifdef G_OS_WIN32
	if ( ! cff_main_open_console())
	{
		return;
	}
#endif
	if (parse_err == FALSE)
	{
		mesg	= gettext_nop("Illegal argument");
		format	= "%s : %s\n";
	}

	else
	{
		mesg	= gettext_nop("Unable to parse");
		format	= "%s : '%s'\n";
	}

	fprintf(stderr, format, gettext(mesg), detail);

} /* static void cff_main_fprn_stderr(int, const char*) */

static FILE *cff_main_open_rc(const char *modus)
{
	FILE		*fd;
	char		*filename;
	const char	*config_dir, *rc_name;

	rc_name = gCffMainCutterffRsrcFileString;

#if HAVE_GCONFIGDIR

	config_dir = g_get_user_config_dir();

	if (config_dir != NULL)
	{

#ifndef G_OS_WIN32
		/* Skip '.' */
		rc_name++;
#endif
	}
	else

#endif /* HAVE_GCONFIGDIR */

	{
		config_dir = g_get_home_dir();

		if (config_dir == NULL)
		{
			return NULL;
		}

	} /* if (config_dir != NULL) .. else */

	filename = g_build_filename(config_dir, rc_name, NULL);

	if (filename == NULL)
	{
		return NULL;
	}

	fd = fopen(filename, modus);
	g_free(filename);

	return fd;

} /* static FILE *cff_main_open_rc(const char*) */

char *cff_main_save_rc(void)
{
	FILE *fd = cff_main_open_rc("w");

	if (fd == NULL)
	{
		return gettext_nop("Unable to open the cutterffrc file");
	}

	fprintf(fd, "# %s\n# %s\n\n",
			gettext("This resource file was generated by 'CutterFF'"),
					gettext("You may edit it if you're careful!"));

	cff_accel_save_data(fd, gCffMainCfgRsrcNameString[CFF_MENUACCEL]);
	cff_gprop_save_data(fd, gCffMainCfgRsrcNameString[CFF_PROP]);

	fclose(fd);

	return NULL;

} /* char *cff_main_save_rc(void) */

static void cff_main_load_rc(void)
{
	FILE *fd;
	char *line, pom[2048];

	fd = cff_main_open_rc("r");

	if (fd == NULL)
	{
		return;
	}

	while ((line = fgets(pom, sizeof(pom), fd)) != NULL)
	{
		static const char white_spaces[] = " \t";

		int i;

		line += strspn(line, white_spaces);

		if (*line == '#' || (i = strcspn(line, "\n")) == 0)
		{
			continue;
		}

		line[i]	= '\0';
		i	= CFF_LAST;

		do
		{
			int		len;
			const char	*str;

			str = gCffMainCfgRsrcNameString[i];
			len = strlen(str);

			if (g_ascii_strncasecmp(line, str, len) == 0)
			{
				line += len;
				line += strspn(line, white_spaces);
				break;
			}
		}
		while (--i >= CFF_FIRST);

		switch (i)
		{
			case CFF_MENUACCEL:
			{

#if GTK_MAJOR_VERSION > 2
				gMainAccelSlist =
					g_slist_append(gMainAccelSlist,
								g_strdup(line));
#else
				if (cff_accel_parse_str(line) == FALSE)
				{
					cff_main_fprn_stderr(TRUE, line);
				}
#endif
				break;
			}

			case CFF_PROP:
			{
				if (cff_gprop_parse_str(line) == FALSE)
				{
					cff_main_fprn_stderr(TRUE, line);
				}

				break;
			}

			default:
			{
				cff_main_fprn_stderr(TRUE, line);

				break;
			}

		} /* switch (i) */

	} /* while ((line = fgets(pom, sizeof(pom), fd)) != NULL) */

	fclose(fd);

} /* static void cff_main_load_rc(void) */

static void cff_main_print_version_help(int what, const char *prgname)
{

#ifdef G_OS_WIN32
	if ( ! cff_main_open_console())
	{
		return;
	}
#endif
	if (what & kPrintVersion)
	{
		printf("%s-%s\n",
			gCffMainCutterffPackageString,
				gCffMainCutterffVersionString);
	}

	if ( ! (what & kPrintHelp))
	{
		return;
	}

	printf("%s %s %s %s\n",
			gettext("Usage:"), prgname,
				gettext("[OPTION]"),
					gettext("[FILENAME]"));

	printf(
		"  -l, --loglevel=LOGNAME  %s\n"
		"                             quiet (%s)\n"
		"                             panic\n"
		"                             fatal\n"
		"                             error\n"
		"                             warning\n"
		"                             info\n"
		"                             verbose\n"
		"                             debug\n"
		"  -x, --threads=[%s|auto] %s\n"
		"                             <0   = %s\n"
		"                              0   = %s\n"
		"                             >0   = %s\n"
		"                             auto = %s\n"
#ifdef HAVE_GETTEXT
		"      --locale=[C|de_DE]\n"
		"      --nls-dir=%s  %s\n"
#endif
		"  -v, --version            %s\n"
		"  -h, --help               %s\n",
		gettext("Loglevel for printing messages to the console"),
		gettext("default"),
		gettext("NUM"),
			gettext("Set number of threads to NUM (default: auto)"),
			gettext("disable multi-threading"),
			gettext("automatic detection"),
			gettext("number of threads"),
			gettext("automatic detection"),
#ifdef HAVE_GETTEXT
		gettext("DIR      "),
			gettext("Set the nlsdir (eg. /usr/share/locale)"),
#endif
		gettext("Display version information and exit"),
		gettext("Display this help and exit")

	); /* printf() */

	return;

} /* static void cff_main_print_version_help(int, const char*) */


#ifdef G_OS_WIN32

static int cff_main_window_console = 0x00;

static int cff_main_connect_console(void)
{
	int	newline;
	DWORD	which;

	newline	= 0;
	which	= STD_OUTPUT_HANDLE;

	for (;;)
	{
		int		fnr;
		HANDLE		stdHandle;
		FILE		*stdFp;

		stdFp = (which == STD_OUTPUT_HANDLE) ? stdout : stderr;

		if ((fnr = fileno(stdFp)) == -1 ||
			(stdHandle = (HANDLE) _get_osfhandle(fnr)) ==
							INVALID_HANDLE_VALUE ||
				GetFileType(stdHandle) == FILE_TYPE_UNKNOWN)
		{
			int	osfHandle;
			FILE	*fp;

			stdHandle = GetStdHandle(which);

			if (stdHandle == INVALID_HANDLE_VALUE)
			{
				break;
			}

			osfHandle = _open_osfhandle((long) stdHandle, _O_TEXT);

			if (osfHandle == -1)
			{
				break;
			}

			fp = _fdopen(osfHandle, "w" );

			if (fp == NULL)
			{
				break;
			}

			*stdFp = *fp;

			if (newline == FALSE)
			{
				newline = TRUE;
				fputs("\n", stdFp);
			}
		}

		if (which != STD_OUTPUT_HANDLE)
		{
			return TRUE;
		}

		which = STD_ERROR_HANDLE;

	} /* for (;;) */

	return FALSE;

} /* static int cff_main_connect_console(void) */

static int cff_main_attach_console(void)
{
	if (AttachConsole(ATTACH_PARENT_PROCESS))
	{
		cff_main_window_console = 0x10;

		if (cff_main_connect_console())
		{
			cff_main_window_console = 0x11;
			return 0x01;
		}
	}

	return 0x00;

} /* static int cff_main_attach_console(void) */

static int cff_main_open_console(void)
{
	if ( ! (cff_main_window_console & 0x10) && ! cff_main_attach_console())
	{
		cff_main_window_console = 0x10;

		if ( ! AllocConsole())
		{
			return 0x00;
		}

		if (cff_main_connect_console())
		{
			cff_main_window_console = 0x11;
		}

	} /*  if ( ! (cff_main_window_console & 0x10) && ! cff_main_attach_console()) */

	return cff_main_window_console & 0x01;

} /* static int cff_main_open_console(int) */

#endif /* G_OS_WIN32 */


static int cff_main_cmd_setup(int argc, char **argv, cff_init_param_t *param)
{
	const struct option options[] =
	{
		{"loglevel", 1, NULL, 'l'},
		{"threads", 1, NULL, 'x'},
#ifdef HAVE_GETTEXT
		{"locale", 1, NULL, 'g'},
		{"nls-dir", 1, NULL, 'g'},
#endif
		{"version", 0, NULL, 'v'},
		{"help", 0, NULL, 'h'},
		/* !!! GTK&GDK OPTIONS !!! */
		{"gtk-module", 1, NULL, 'g'},
		{"g-fatal-warnings", 0, NULL, 'g'},
		{"gtk-debug", 2, NULL, 'g'},
		{"gtk-no-debug", 2, NULL, 'g'},
		{"gdk-debug", 2, NULL, 'g'},
		{"gdk-no-debug", 2, NULL, 'g'},
		{"display", 1, NULL, 'g'},
		{"sync", 0, NULL, 'g'},
		{"no-xshm", 0, NULL, 'g'},
		{"name", 1, NULL, 'g'},
		{"class", 1, NULL, 'g'},
		{"gxid_host", 1, NULL, 'g'},
		{"gxid_port", 1, NULL, 'g'},
		{"xim-preedit", 1, NULL, 'g'},
		{"xim-status", 1, NULL, 'g'},
		{NULL, 0, NULL, 0},
	};

	int c, index, print_info;

#ifdef G_OS_WIN32
	if (argc <= 1 || ! cff_main_attach_console())
	{
		/* disable error messages - no 'stderr' */
		opterr = 0;
	}
#endif

	print_info = 0;

	while ((c = getopt_long(argc, argv, "l:x:vh",
					options, &index)) != - 1)
	{
		switch (c)
		{
			case 'l':
				param->loglevel = optarg;
				break;
			case 'x':
				if (strcmp(optarg, "auto") != 0)
				{
					char *eptr;
					param->threads = strtol(optarg, &eptr, 10);
					if (eptr[0] != '\0')
					{
						cff_main_fprn_stderr(FALSE, optarg);
						return EINVAL;
					}
				}
				break;
			case 'v':
				print_info |= kPrintVersion;
				break;
			case 'h':
				print_info |= kPrintHelp;
				break;
			case 'g':
				break;

			default:
				return EINVAL;

		} /* switch (c) */

	} /* while ((c = getopt_long(argc, argv, ... )) != - 1) */

	if (print_info)
	{
		cff_main_print_version_help(print_info, argv[0]);
		exit(0);
	}

	if (optind < argc)
	{
		if (argc > optind + 1)
		{
			cff_main_fprn_stderr(FALSE, argv[optind]);
			return EINVAL;
		}

		param->fname = argv[optind];
	}

	cff_main_load_rc();

	return 0;

} /* static int cff_main_cmd_setup(int, char**, cff_init_param_t*) */

static void cff_main_sigterm(int signum)
{
	int save_flag;

	/* Store settings */
	if ( ! cff_gprop_get_bool(kPropBoolOptSaveConfig, &save_flag) || save_flag)
	{
		cff_main_save_rc();
	}

	if (signum == SIGTERM)
	{
		_exit(0);
	}

} /* static void cff_main_sigterm(int) */

static void cff_main_atexit(void)
{
	/* User lost interrest */
	cff_main_sigterm(SIGQUIT);
}

static void cff_main_signal_init(void)
{

#ifdef HAVE_SIGACTION

	struct sigaction sigact;

	/* Block other signals, while executing 'sa_handler' */
	sigfillset(&sigact.sa_mask);
	sigact.sa_flags = 0;

	/* 'CutterFF' was terminated */
	sigact.sa_handler = cff_main_sigterm;
	sigaction(SIGTERM, &sigact, NULL);
#else
	signal(SIGTERM, cff_main_sigterm);
#endif

} /* static void cff_main_signal_init(void) */

#ifdef HAVE_GSETENV
#define CFF_SETENV(name, value) g_setenv((name), (value), TRUE)
#else
#define CFF_SETENV(name, value) putenv(g_strconcat((name), "=", (value), NULL))
#endif

static void cff_main_init_locale(int argc, const char **argv)
{
	const char *arg, *locale, *nlsdir;

	locale = nlsdir = NULL;

	while ((arg = *(argv++)) != NULL)
	{
		if (strncmp(arg, "--locale", 8) == 0)
		{
			if (arg[8] == '=')
			{
				/* --locale=LOCALE */
				locale = &arg[9];
			}

			else
			{
				/* --locale LOCALE */
				locale = *(argv++);
			}
		}

		else if (strncmp(arg, "--nls-dir", 9) == 0)
		{
			if (arg[9] == '=')
			{
				/* --nls-dir=DIR */
				nlsdir = &arg[10];
			}

			else
			{
				/* --nls-dir DIR */
				nlsdir = *(argv++);
			}
		}
	}

	if (locale)
	{
		const char *s;

		CFF_SETENV("LANG", locale);

		s = strchr(locale, '.');

		if (s)
		{
			s = g_strndup(locale, s - locale);
		}

		else
		{
			s = locale;
		}

		CFF_SETENV("LANGUAGE", s);
		CFF_SETENV("LC_ALL", locale);

		if (s != locale)
		{
			g_free((char *) s);
		}
	}

	CFF_NLS_INIT_LOCALE(nlsdir);

} /* static void cff_main_init_locale(int, const char**) */


#ifdef HAVE_GMENU

static void cff_main_startup(GtkApplication *app, GSList *accel_slist)
{
	/* Parse accelerators  after 'gtk_init' */

	while (accel_slist != NULL)
	{
		gchar *line = accel_slist->data;

		if (cff_accel_parse_str(line) == FALSE)
		{
			cff_main_fprn_stderr(TRUE, line);
		}

		g_free(line);
		accel_slist = g_slist_remove(accel_slist, line);

	} /* while (accel_slist != NULL) */

} /* static void cff_main_startup(GtkApplication*, GSList*) */

#endif /* HAVE_GMENU */


int main(int argc, char **argv)
{
	/* cff_ffmpeg.c */
	extern int cff_ffmpeg_init(const char*, int);

	int			error;
	char			**strv;
	cff_init_param_t	param;
#ifdef HAVE_GMENU
	GtkApplication		*app;
	gint			status;
#endif


	/*
		------
		 INIT
		------
	*/

#ifdef G_OS_WIN32
	strv = cff_main_win32_get_command_line();
#else
	strv = argv;
#endif

	cff_main_init_locale(argc, (const char **) strv);

	param.fname	= NULL;
	param.loglevel	= NULL;
	param.threads	= 0;

	error = cff_main_cmd_setup(argc, strv, &param);

	if (error)
	{
		return error;
	}

	if (cff_ffmpeg_init(param.loglevel, param.threads) < 0)
	{
		cff_main_fprn_stderr(FALSE, param.loglevel);
		return EINVAL;
	}

	cff_main_signal_init();
	atexit(cff_main_atexit);

#ifdef G_OS_WIN32

	if (param.fname != NULL)
	{
		param.fname = g_strdup(param.fname);
	}

	g_strfreev(strv);

#endif /* G_OS_WIN32 */


	/*
		-------
		 START
		-------
	*/

#ifdef HAVE_GMENU

	app = gtk_application_new(NULL, G_APPLICATION_FLAGS_NONE);

	GuiSignalConnect(app, "startup", cff_main_startup, gMainAccelSlist);
	GuiSignalConnect(app, "activate", cff_gui_init, &param);

	status = g_application_run(G_APPLICATION(app), 0, NULL);

	g_object_unref(app);

	return status;

#else /* ! HAVE_GMENU */

	gtk_init(&argc, &argv);

#if glib_major_version == 2 && glib_minor_version < 32

	if ( ! g_thread_supported())
	{
		g_thread_init(NULL);
	}
#endif

#if GTK_MAJOR_VERSION > 2

	while (gMainAccelSlist != NULL)
	{
		gchar *line = gMainAccelSlist->data;

		if (cff_accel_parse_str(line) == FALSE)
		{
			cff_main_fprn_stderr(TRUE, line);
		}

		g_free(line);
		gMainAccelSlist = g_slist_remove(gMainAccelSlist, line);

	} /* while (gMainAccelSlist != NULL) */
#endif

	cff_gui_init(&param);
	gtk_main();

	return 0;

#endif /* #ifdef HAVE_GMENU .. #else */

} /* int main(int, char**) */

/* EOF */
