/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017..2019,2021,2023,2024 Harald Foerster	*/
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/

#include <string.h>
#include <glib/gstdio.h>

#include "cff_macros-gtk.h"
#include "cff_gprop.h"
#include "cff_nls.h"
#include "cff_charconv.h"

/*
#include <gtk/gtk.h>
#include "cff_tools-gtk.h"
*/


#if GTK_MAJOR_VERSION == 2
#define Gui_DIALOG_ICON_INFO		GTK_STOCK_DIALOG_INFO
#define Gui_DIALOG_ICON_WARNING		GTK_STOCK_DIALOG_WARNING
#define Gui_DIALOG_ICON_ERROR		GTK_STOCK_DIALOG_ERROR
#else
#define Gui_DIALOG_ICON_INFO		"dialog-information"
#define Gui_DIALOG_ICON_WARNING		"dialog-warning"
#define Gui_DIALOG_ICON_ERROR		"dialog-error"
#endif

/* Global */
cff_windata_t gWindata;

#if GTK_MAJOR_VERSION == 4
GMainLoop *gMainLoop = NULL;
#endif

static const gchar gCutterFFString[] = "CutterFF";
static const gchar gOkayString[]     = gettext_nop("OK");
static const gchar gCancelString[]   = gettext_nop("Cancel");


void cff_guitl_window_set_title_filename(GtkWidget *window, const gchar *title)
{
	gchar *str, buf[2048];

	if (title == NULL)
	{
		title = gettext_nop("No video");
		str   = cff_charconv_utf8_gettext(title);
	}

	else
	{
		/* Already UTF-8 */
		str = (gchar *) title;
	}

	snprintf(buf, sizeof(buf), "%s: %s", gCutterFFString, str);
	CFF_CHARCONV_UTF8_FREE(str, title);

	gtk_window_set_title(GTK_WINDOW(window), buf);

} /* void cff_guitl_window_set_title_filename(GtkWidget*, const gchar*) */

static void cff_guitl_window_set_title_gettext(GtkWidget *window, const gchar *title)
{
	gchar *str;

	if (title != NULL)
	{
		str = cff_charconv_utf8_gettext(title);
		gtk_window_set_title(GTK_WINDOW(window), str);
		CFF_CHARCONV_UTF8_FREE(str, title);
	}

} /* static void cff_guitl_window_set_title_gettext(GtkWidget*, const gchar*) */


#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
static gint cff_guitl_window_focus_in(GtkWidget *window,
				GdkEventFocus *e, cff_windata_t *windata)
#else /* GTK_MAJOR_VERSION == 4 */
static void cff_guitl_window_focus_in(GtkEventControllerFocus *controller,
							cff_windata_t *windata)
#endif
{
	GSList *slist;

#if GTK_MAJOR_VERSION == 4
	GtkWidget *window =
		gtk_event_controller_get_widget(GTK_EVENT_CONTROLLER(controller));
#endif

	windata->topwin = window;
	slist = windata->slist;

	if (slist != NULL)
	{
		slist = g_slist_remove(slist, window);
	}

	windata->slist = g_slist_prepend(slist, window);

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
	return FALSE;
#endif

} /* static gint cff_guitl_window_focus_in(GtkWidget*, GdkEventFocus*, windata_t*) */

static void cff_guitl_window_destroy(GtkWidget *window, cff_windata_t *windata)
{
	/*
		If no window of CutterFF is on top, then
		the 'focus-in' signal will not send.
		The pointer of the destroyed window/dialog
		is now invalid and should removed.
		Used in 'cff_guitl_dialog_icon' for setting
		the transient parent.
	*/

	GSList *slist = windata->slist;

	if (slist != NULL)
	{
		slist = g_slist_remove(slist, window);
		windata->slist  = slist;
	}

	windata->topwin = (slist == NULL) ? windata->mainwin : slist->data;

} /* static void cff_guitl_window_destroy(GtkWidget*, cff_windata_t*) */

GtkWidget *cff_guitl_window_new(GtkWidget *child, gint border)
{
	GtkWidget *window = GuiWindowNew();

	GuiWindowAddChild(window, child);

	if (border > 0)
	{
		GuiContainerSetBorderWidth(child, border);
	}

	cff_guitl_window_set_topwin(window);

	return window;

} /* GtkWidget *cff_guitl_window_new(GtkWidget*, gint) */

GtkWidget *cff_guitl_window_new_gettext(const gchar *title, GtkWidget *child, gint border)
{
	GtkWidget *window;

	window = cff_guitl_window_new(child, border);
	cff_guitl_window_set_title_gettext(window, title);

	return window;

} /* GtkWidget *cff_guitl_window_new_gettext(const gchar*, GtkWidget*, gint) */


#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
static void cff_guitl_window_resize(GtkWidget *widget, GdkRectangle *a, gpointer data)
#else
static gboolean cff_guitl_window_resize(GtkWidget *widget, gpointer data)
#endif
{
	gint prop = GPOINTER_TO_INT(data);
	cff_guitl_widget_size_to_prop(widget, prop);

#if GTK_MAJOR_VERSION == 4
	return FALSE;
#endif

} /* static void cff_guitl_window_resize(GtkWidget *w, GdkRectangle*, gpointer) */

static gboolean cff_guitl_window_test_size(GtkWindow *window, gint width, gint height)
{

#ifdef HAVE_GMENU

	GdkDisplay	*display;
	GdkMonitor	*monitor;
	GdkRectangle	rect;

#if GTK_MAJOR_VERSION == 3
	display = gtk_widget_get_display(GTK_WIDGET(window));
	monitor = gdk_display_get_monitor_at_window(display,
				gtk_widget_get_window(GTK_WIDGET(window)));
#else
	GdkSurface	*surface;
	GtkNative	*native;

	native  = gtk_widget_get_native(GTK_WIDGET(window));
	surface = gtk_native_get_surface(native);
	display = gdk_surface_get_display(surface);
	monitor = gdk_display_get_monitor_at_surface(display, surface);
#endif
	gdk_monitor_get_geometry(monitor, &rect);

	if (rect.width < width || rect.height < height)
	{
		return FALSE;
	}

#else /* ! HAVE_GMENU */

	GdkScreen *screen = gtk_window_get_screen(window);

	if (gdk_screen_get_width(screen) < width ||
			gdk_screen_get_height(screen) < height)
	{
		return FALSE;
	}

#endif /* #ifdef HAVE_GMENU .. #else */

	return TRUE;

} /* static gboolean cff_guitl_window_test_size(GtkWindow *window, gint, gint) */

void cff_guitl_window_setup(GtkWidget *widget, gint prop)
{
	gint		width, height;
	GtkWindow	*window;

#if GTK_MAJOR_VERSION == 4
	gtk_widget_show(widget);
#endif
	window = GTK_WINDOW(widget);

	if ( (cff_gprop_get_int(prop, &width) == FALSE || width <= 0) ||
		(cff_gprop_get_int(prop + 1, &height) == FALSE || height <= 0) ||
			cff_guitl_window_test_size(window, width, height) == FALSE )
	{
		width  = -1;
		height = -1;
	}

	gtk_window_set_resizable(window, TRUE);

	if (width != -1 && height != - 1)
	{
		gtk_window_set_default_size(window, width, height);
	}

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
	GuiSignalConnect(widget, "size-allocate",
			cff_guitl_window_resize, GINT_TO_POINTER(prop));
#else
	GuiSignalConnect(widget, "close-request",
			cff_guitl_window_resize, GINT_TO_POINTER(prop));
#endif

} /* void cff_guitl_window_setup(GtkWidget*, gint) */

void cff_guitl_window_set_topwin(GtkWidget *window)
{

#if GTK_MAJOR_VERSION == 4
	GtkEventController *controller;

	controller = gtk_event_controller_focus_new();
	gtk_widget_add_controller(window, controller);
	GuiSignalConnect(controller, "enter",
				cff_guitl_window_focus_in, &gWindata);
#else
	GuiSignalConnect(window, "focus-in-event",
				cff_guitl_window_focus_in, &gWindata);
#endif

	GuiSignalConnect(window, "destroy",
				cff_guitl_window_destroy, &gWindata);

} /* void cff_guitl_window_set_topwin(GtkWidget*) */

GtkWidget *cff_guitl_scrolled_window_new(gint wmin, gint hmin, GtkWidget *child)
{
	GtkWidget *swin = GuiScrolledWindowNew();

	if (child != NULL)
	{
		GuiScrolledWindowAddChild(swin, child);
	}

	gtk_widget_set_size_request(swin, wmin, hmin);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(swin),
				GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
#if GTK_MAJOR_VERSION == 3
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(swin),
								GTK_SHADOW_IN);
#endif

#if GTK_MAJOR_VERSION > 2
	gtk_widget_set_hexpand(swin, TRUE);
	gtk_widget_set_vexpand(swin, TRUE);
#endif

	return swin;

} /* GtkWidget *cff_guitl_scrolled_window_new(gint, gint, GtkWidget*) */

void cff_guitl_dialog_set_child(GtkWidget *dialog, GtkWidget *child)
{
	GtkWidget *content;

	/* Allocate the size needed by child */
	GuiWidgetShowAll(child);

	content	= gtk_dialog_get_content_area(GTK_DIALOG(dialog));
	GuiBoxAddChild(content, child);

} /* void cff_guitl_dialog_set_child(GtkWidget*, GtkWidget*) */

static void cff_guitl_dialog_modal_destroy(GtkWidget *w, GtkWidget **mbar)
{
	if (mbar != NULL && *mbar != NULL)
	{
		GuiWidgetUnsetStateSelected(*mbar);
	}

} /* static void cff_guitl_dialog_modal_destroy(GtkWidget*, GtkWidget**) */

void cff_guitl_dialog_set_modal(GtkWidget *dialog, GtkWidget *parent, GtkWidget **mbar)
{
	if (mbar != NULL && *mbar != NULL)
	{
		GuiWidgetSetStateSelected(*mbar);
		GuiSignalConnect(dialog, "destroy",
				cff_guitl_dialog_modal_destroy, mbar);
	}

	GuiWindowSetModal(dialog, parent);

} /* void cff_guitl_dialog_set_modal(GtkWidget*, GtkWidget*, GtkWidget**) */

static gint cff_guitl_dialog_icon_run(GtkWidget *dialog, const gchar *icon, const gchar *text, GtkWidget **mbar)
{
	gint		retv;
	GtkWidget	*frame, *hbox, *image, *label;

	if (mbar != NULL && *mbar != NULL)
	{
		GuiWidgetSetStateSelected(*mbar);
	}

#if GTK_MAJOR_VERSION == 4
	gtk_widget_show(dialog);
#endif
	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);

	hbox  = GuiHBoxNew(FALSE, 0);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, 4, hbox);

	image = GuiImageGetDialogIcon(icon);

#if GTK_MAJOR_VERSION == 4
	gtk_image_set_icon_size(GTK_IMAGE(image), GTK_ICON_SIZE_LARGE);
#endif
	GuiMiscSetPadding(image, 8);
	GuiBoxAddChild(hbox, image);

	label = cff_guitl_label_new_gettext(text);
	GuiMiscSetPadding(label, 8);
	GuiBoxAddChild(hbox, label);

	cff_guitl_dialog_set_child(dialog, frame);

	GuiSignalConnect(dialog, "destroy",
			cff_guitl_window_destroy, &gWindata);

	retv = GuiDialogRun(dialog);

	GuiDialogClose(dialog, retv);

	if (mbar != NULL && *mbar != NULL)
	{
		GuiWidgetUnsetStateSelected(*mbar);
	}

	return retv;

} /* static gint cff_guitl_dialog_icon_run(GtkWidget*, const gchar *, const gchar*, GtkWidget**) */

static void cff_guitl_dialog_icon(const gchar *icon, const gchar *str, GtkWidget **mbar)
{
	gchar		*accept, *text;
	GtkWidget	*dialog;

	accept = cff_charconv_utf8_gettext(gOkayString);
	dialog = gtk_dialog_new_with_buttons(gCutterFFString,
		GTK_WINDOW(gWindata.topwin),
		GTK_DIALOG_MODAL,
		accept, GTK_RESPONSE_ACCEPT,
		NULL);

	CFF_CHARCONV_UTF8_FREE(accept, gOkayString);

	text = cff_charconv_utf8_gettext(str);
	cff_guitl_dialog_icon_run(dialog, icon, text, mbar);
	CFF_CHARCONV_UTF8_FREE(text, str);

} /* static void cff_guitl_dialog_icon(const gchar*, const gchar*, GtkWidget**) */

void cff_guitl_dialog_info(const gchar *text, GtkWidget **mbar)
{
	cff_guitl_dialog_icon(Gui_DIALOG_ICON_INFO, text, mbar);

} /* void cff_guitl_dialog_info(const gchar*, GtkWidget**) */

void cff_guitl_dialog_error(const gchar *text, GtkWidget **mbar)
{
	cff_guitl_dialog_icon(Gui_DIALOG_ICON_ERROR, text, mbar);

} /* void cff_guitl_dialog_error(const gchar*, GtkWidget**) */

void cff_guitl_dialog_warning(const gchar *text, GtkWidget **mbar)
{
	cff_guitl_dialog_icon(Gui_DIALOG_ICON_WARNING, text, mbar);

} /* void cff_guitl_dialog_warning(const gchar*, GtkWidget**) */

gboolean cff_guitl_dialog_warning_2btn(const gchar *text, GtkWidget *parent, GtkWidget **mbar)
{
	GtkWidget *dialog;

	dialog = cff_guitl_dialog_new_gettext_2btn(gCutterFFString,
						gCancelString,
						gettext_nop("Yes"),
						parent);
	if (cff_guitl_dialog_icon_run(dialog,
		Gui_DIALOG_ICON_WARNING, text, mbar) == GTK_RESPONSE_ACCEPT)
	{
		return TRUE;
	}

	return FALSE;

} /* gboolean cff_guitl_dialog_warning_2btn(const gchar*, GtkWidget*, GtkWidget**) */

GtkWidget *cff_guitl_dialog_new_gettext_2btn(const gchar *title, const gchar *btn_no, const gchar *btn_yes, GtkWidget *parent)
{
	gchar		*accept, *cancel, *title_utf8;
	GtkWidget	*widget;

	cancel     = cff_charconv_utf8_gettext(btn_no);
	accept     = cff_charconv_utf8_gettext(btn_yes);
	title_utf8 = cff_charconv_utf8_gettext(title);

	widget = gtk_dialog_new_with_buttons(title_utf8,
		GTK_WINDOW(parent),
		GTK_DIALOG_MODAL,
		cancel, GTK_RESPONSE_CANCEL,
		accept, GTK_RESPONSE_ACCEPT,
		NULL);

	CFF_CHARCONV_UTF8_FREE(title_utf8, title);
	CFF_CHARCONV_UTF8_FREE(accept, btn_yes);
	CFF_CHARCONV_UTF8_FREE(cancel, btn_no);

	return widget;

} /* GtkWidget *cff_guitl_dialog_new_gettext_2btn(const gchar*, const gchar*, const gchar*, GtkWidget*) */

GtkWidget *cff_guitl_dialog_new_cancel_ok(const gchar *title, GtkWidget *parent)
{
	return cff_guitl_dialog_new_gettext_2btn(title,
						gCancelString,
						gOkayString,
						parent);

} /* GtkWidget *cff_guitl_dialog_new_cancel_ok(const gchar*, GtkWidget*) */

void cff_guitl_dialog_okay_run(const gchar *title, GtkWidget *child, GtkWidget **mbar, int prop)
{
	gchar		*accept, *title_utf8;
	GtkWidget	*dialog;
#if GTK_MAJOR_VERSION == 4
	GtkResponseType	response;
#endif

	if (mbar != NULL && *mbar != NULL)
	{
		GuiWidgetSetStateSelected(*mbar);
	}

	accept     = cff_charconv_utf8_gettext(gOkayString);
	title_utf8 = cff_charconv_utf8_gettext(title);

	dialog = gtk_dialog_new_with_buttons(title_utf8,
		GTK_WINDOW(gWindata.mainwin),
		GTK_DIALOG_MODAL,
		accept, GTK_RESPONSE_ACCEPT,
		NULL);

	if (prop >= 0)
	{
		cff_guitl_window_setup(dialog, prop);
	}

	else
	{
		gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);

#if GTK_MAJOR_VERSION == 4
		gtk_widget_show(dialog);
#endif
	} /* if (prop >= 0) .. else */

	if (title != NULL)
	{
		CFF_CHARCONV_UTF8_FREE(title_utf8, title);
	}

	CFF_CHARCONV_UTF8_FREE(accept, gOkayString);

	if (child != NULL)
	{
		cff_guitl_dialog_set_child(dialog, child);
	}

	gtk_dialog_set_default_response(GTK_DIALOG(dialog),
						GTK_RESPONSE_ACCEPT);
#if GTK_MAJOR_VERSION == 4
	response =
#endif
	GuiDialogRun(dialog);

	GuiDialogClose(dialog, response);

	if (mbar != NULL && *mbar != NULL)
	{
		GuiWidgetUnsetStateSelected(*mbar);
	}

} /* void cff_guitl_dialog_okay_run(const gchar*, GtkWidget*, GtkWidget**, int) */

GtkWidget *cff_guitl_label_new_gettext(const gchar *str)
{
	gchar		*s;
	GtkWidget	*label;

	s = cff_charconv_utf8_gettext(str);
	label = gtk_label_new(s);
	CFF_CHARCONV_UTF8_FREE(s, str);

	return label;

} /* GtkWidget *cff_guitl_label_new_gettext(const gchar*) */

GuiMenuItem *cff_guitl_menu_item_new_gettext(const gchar *label)
{
	gchar		*str;
	GuiMenuItem	*mi;

	str = cff_charconv_utf8_gettext(label);
	mi  = GuiMenuItemNewWithLabel(str);
	CFF_CHARCONV_UTF8_FREE(str, label);

	return mi;

} /* GuiMenuItem *cff_guitl_menu_item_new_gettext(const gchar*) */

GuiMenuItem *cff_guitl_check_menu_item_new_gettext(const gchar *label)
{
	gchar		*str;
	GuiMenuItem	*mi;

	str = cff_charconv_utf8_gettext(label);
	mi = GuiCheckMenuItemNewWithLabel(str);
	CFF_CHARCONV_UTF8_FREE(str, label);

	return mi;

} /* GuiMenuItem *cff_guitl_check_menu_item_new_gettext(const gchar*) */

GtkWidget *cff_guitl_frame_new_add_child_gettext(const gchar *title, gint border, GtkWidget *child)
{
	gchar		*str;
	GtkWidget	*frame;

	str = cff_charconv_utf8_gettext(title);
	frame = gtk_frame_new(str);
	CFF_CHARCONV_UTF8_FREE(str, title);

	GuiContainerSetBorderWidth(frame, border);
	GuiFrameAddChild(frame, child);

	return frame;

} /* GtkWidget *cff_guitl_frame_new_add_child_gettext(const gchar*, gint, GtkWidget*) */

GtkWidget *cff_guitl_check_button_new_gettext(const gchar *label)
{
	gchar		*str;
	GtkWidget	*button;

	str = cff_charconv_utf8_gettext(label);
	button = gtk_check_button_new_with_label(str);
	CFF_CHARCONV_UTF8_FREE(str, label);

	return button;

} /* GtkWidget *cff_guitl_check_button_new_gettext(const gchar*) */

GtkWidget *cff_guitl_vbox_add_check_button_gettext(GtkWidget *vbox, const gchar *label)
{
	GtkWidget *button, *hbox;

	hbox = GuiHBoxNew(TRUE, 2);

	GuiContainerSetBorderWidth(hbox, 2);
	GuiVBoxPackStart(vbox, hbox, FALSE, FALSE, 2);

	button = cff_guitl_check_button_new_gettext(label);

	GuiHBoxPackStart(hbox, button, TRUE, FALSE, 2);

	return button;

} /* GtkWidget *cff_guitl_vbox_add_check_button_gettext(GtkWidget*, const gchar*) */

GtkWidget *cff_guitl_hbox_add_button(GtkWidget *hbox, const gchar *text, GCallback func, gpointer data)
{
	GtkWidget *button = gtk_button_new_with_label(text);

	GuiHBoxPackStart(hbox, button, FALSE, FALSE, 1);

	if (func != NULL)
	{
		GuiSignalConnect(button, "clicked", func, data);
	}

	return button;

} /* GtkWidget *cff_guitl_hbox_add_button(GtkWidget*, const gchar*, GCallback, gpointer) */

GtkWidget *cff_guitl_vbox_add_button(GtkWidget *vbox, const gchar *text, GCallback func, gpointer data)
{
	GtkWidget *button = gtk_button_new_with_label(text);

	GuiVBoxPackStart(vbox, button, FALSE, FALSE, 1);

	if (func != NULL)
	{
		GuiSignalConnect(button, "clicked", func, data);
	}

	return button;

} /* GtkWidget *cff_guitl_hbox_add_button(GtkWidget*, const gchar*, GCallback func, gpointer data) */

static GtkWidget *cff_guitl_button_gettext(const gchar *text, GCallback func, gpointer data)
{
	gchar		*str;
	GtkWidget	*button;

	str = cff_charconv_utf8_gettext(text);
	button = gtk_button_new_with_label(str);
	CFF_CHARCONV_UTF8_FREE(str, text);

	if (func != NULL)
	{
		GuiSignalConnect(button, "clicked", func, data);
	}

	return button;

} /* static GtkWidget *cff_guitl_button_gettext(const gchar*, GCallback, gpointer) */

GtkWidget *cff_guitl_hbox_add_button_gettext(GtkWidget *hbox, const gchar *text, GCallback func, gpointer data)
{
	GtkWidget *button = cff_guitl_button_gettext(text, func, data);

	GuiHBoxPackStart(hbox, button, FALSE, FALSE, 1);

	return button;

} /* GtkWidget *cff_guitl_hbox_add_button(GtkWidget*, const gchar*, GCallback, gpointer) */

GtkWidget *cff_guitl_vbox_add_button_gettext(GtkWidget *vbox, const gchar *text, GCallback func, gpointer data)
{
	GtkWidget *button = cff_guitl_button_gettext(text, func, data);

	GuiVBoxPackStart(vbox, button, FALSE, FALSE, 1);

	return button;

} /* GtkWidget *cff_guitl_hbox_add_button(GtkWidget*, const gchar*, GCallback, gpointer) */

GtkWidget *cff_guitl_hbutton_box_new(GtkWidget *parent, gint spacing)
{
	GtkWidget *hbutton_box = GuiHButtonBoxNew();

	if (parent != NULL)
	{

#if GTK_MAJOR_VERSION == 2
		if (GTK_IS_VBOX(parent))
		{
			gtk_box_pack_start(GTK_BOX(parent),
						hbutton_box, FALSE, FALSE, 0);
		}
		else
#endif
		{
			GuiBoxAddChild(parent, hbutton_box);
		}

	} /* if (parent != NULL) */

	gtk_box_set_spacing(GTK_BOX(hbutton_box), spacing);

#if GTK_MAJOR_VERSION == 4
	gtk_widget_set_vexpand(hbutton_box, FALSE);
	gtk_widget_set_halign(hbutton_box, GTK_ALIGN_END);
#else
	gtk_button_box_set_layout(GTK_BUTTON_BOX(hbutton_box), 
						GTK_BUTTONBOX_END);
#endif

	return hbutton_box;

} /* GtkWidget *cff_guitl_hbutton_box_new(GtkWidget*, gint) */

GtkWidget *cff_guitl_button_box_add_button(GtkWidget *button_box, const gchar *text, GCallback func, gpointer data)
{
	GtkWidget *button = cff_guitl_button_gettext(text, func, data);

	GuiBoxAddChild(button_box, button);
	GuiContainerSetBorderWidth(button, 2);

	GuiWidgetCanDefault(button);

	return button;

} /* GtkWidget *cff_guitl_button_box_add_button(GtkWidget*, const gchar*, GCallback, gpointer) */

void cff_guitl_beep(void)
{
	gint active;

	if (cff_gprop_get_bool(kPropBoolOptBeep, &active) && active)
	{
		GdkDisplay *display;
		GtkWidget  *widget;

		widget = gWindata.topwin;

		if (widget == NULL)
		{
			display = gdk_display_get_default();
		}

		else
		{
			display = gtk_widget_get_display(widget);
		}

		gdk_display_beep(display);
	}

} /* void cff_guitl_beep(void) */

void cff_guitl_widget_set_tooltip_gettext(GtkWidget *widget, const gchar *text)
{
	gchar *str = cff_charconv_utf8_gettext(text);

	gtk_widget_set_tooltip_text(widget, str);

	if (text != NULL)
	{
		CFF_CHARCONV_UTF8_FREE(str, text);
	}

} /* void cff_guitl_widget_set_tooltip_gettext(GtkWidget*, const gchar*) */

void cff_guitl_widget_size_to_prop(GtkWidget *widget, gint prop)
{
	/*
		either main window, info dialogs or filechooser
		-----------------------------------------------
		kPropIntMainWinWidth         + kPropIntMainWinHeight
		kPropIntFileChooserWinWidth  + kPropIntFileChooserWinHeight
		kPropIntFileSaveWinWidth     + kPropIntFileSaveWinHeight
		kPropIntFileImageWinWidth    + kPropIntFileImageWinHeight
		kPropIntFileInfoWinWidth     + kPropIntFileInfoWinHeight
		kPropIntHelpFFmpegWinWidth   + kPropIntHelpFFmpegWinHeight
		kPropIntHelpCutterFFWinWidth + kPropIntHelpCutterFFWinHeight
		kPropIntLogWinWidth          + kPropIntLogWinHeight
	*/

	if (widget != NULL)
	{
		gint width, height;

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
		gtk_window_get_size(GTK_WINDOW(widget), &width, &height);
#else
		gtk_window_get_default_size(GTK_WINDOW(widget), &width, &height);
#endif
		cff_gprop_set_int(prop, width);
		cff_gprop_set_int(prop + 1, height);

	} /* if (widget != NULL) */

} /* void cff_guitl_widget_size_to_prop(GtkWidget*, gint) */


#if GTK_MAJOR_VERSION > 2

GtkWidget *cff_guitl_box_new(GtkOrientation orientation,
				gboolean homogeneous, gint spacing)
{
	GtkWidget *box;

	box = gtk_box_new(orientation, spacing);
	gtk_box_set_homogeneous(GTK_BOX(box), homogeneous);

	return box;

} /* GtkWidget *cff_guitl_box_new(GtkOrientation, gboolean, gint) */

#endif /* GTK_MAJOR_VERSION > 2 */


#if GTK_MAJOR_VERSION == 4

void cff_guitl_widget_set_border_width(GtkWidget *widget, gint width)
{
	gtk_widget_set_margin_start(widget, width);
	gtk_widget_set_margin_end(widget, width);
	gtk_widget_set_margin_top(widget, width);
	gtk_widget_set_margin_bottom(widget, width);

} /*  void cff_guitl_widget_set_border_width(GtkWidget*, gint) */

static void cff_guitl_dialog_response(GtkDialog *dialog, gint response_id, gint *retv)
{
	*retv = response_id;
	cff_guitl_main_level_quit();

} /* static void cff_guitl_dialog_response(GtkDialog*, gint, gint*) */

gint cff_guitl_dialog_run(GtkWidget *dialog)
{
	volatile gint	retv;
	guint		sigid;

	retv  = GTK_RESPONSE_CANCEL;
	sigid = GuiSignalConnect(dialog, "response",
				cff_guitl_dialog_response, &retv);

	cff_guitl_main_level_start();

	if (retv != GTK_RESPONSE_DELETE_EVENT)
	{
		GuiSignalHandlerDisconnect(dialog, sigid);
	}

	return retv;

} /* gint cff_guitl_dialog_run(GtkWidget*) */

void cff_guitl_main_level_start(void)
{
	GMainLoop *loop, *prev;

	prev      = gMainLoop;
	loop      = g_main_loop_new(NULL, FALSE);
	gMainLoop = loop;

	g_main_loop_run(loop);

	gMainLoop = prev;
	g_main_loop_unref(loop);

} /* void cff_guitl_main_level_start(void) */

void cff_guitl_main_level_quit(void)
{
	/* Stops a 'GMainLoop' from running */

	if (gMainLoop != NULL)
	{
		g_main_loop_quit(gMainLoop);
	}

} /* void cff_guitl_main_level_quit(void) */

#endif /* GTK_MAJOR_VERSION == 4 */


#ifdef HAVE_GMENU

char *cff_guitl_create_action_name(gchar *dst, gsize len, const gchar *src)
{
	if (dst != NULL)
	{
		if (src == NULL || len < CFF_GACTION_PREFIX_LENGTH + 1)
		{
			dst[0] = '\0';
		}

		else
		{
			guint  chr;
			gchar *ptr;

			/* Set prefix */
			ptr = strcpy(dst, CFF_GACTION_PREFIX_STRING);

			/* Skip prefix */
			ptr += CFF_GACTION_PREFIX_LENGTH;
			len -= CFF_GACTION_PREFIX_LENGTH;

			while (--len > 0 && (chr = *(src++)) != '\0')
			{
				if (g_ascii_isalnum(chr))
				{
					*(ptr++) = g_ascii_tolower(chr);
				}
			}

			/* Terminate the string */
			*ptr = '\0';
		}

	} /* if (dst != NULL) */

	return dst;

} /* char *cff_guitl_create_action_name(gchar*, gsize, const gchar*) */

#endif /* HAVE_GMENU */


/* Misc */

gboolean cff_tl_is_ascii_str(const gchar *str)
{
	if (str != NULL)
	{
		guint		c;
		const guchar	*p;

		p = (const guchar *) str;

		while ((c = *(p++)) != '\0')
		{
			if (c < 0x20 || c > 0x7e)
			{
				return FALSE;
			}
		}

	} /* if (str != NULL) */

	return TRUE;

} /* gboolean cff_tl_is_ascii_str(const gchar*) */

gchar *cff_tl_replace_non_utf8_chars_in_str(gchar *str)
{
	/* Replace control characters with '?' */

	if (str != NULL)
	{
		guint  chr;
		guchar *ptr;

		ptr = (guchar *) str;

		while ((chr = *(ptr++)) != '\0')
		{
			if (chr > 0x7e)
			{
				ptr--;

				if (g_utf8_get_char_validated((const gchar *) ptr, -1) == -1)
				{
					*(ptr++) = '?';
				}

				else
				{
					ptr = (guchar *)
						g_utf8_find_next_char((const gchar *) ptr, NULL);
				}
			}

			else if (chr < 0x08 || (chr > 0x0D && chr < 0x20))
			{
				ptr[-1] = '?';
			}

		} /* while ((chr = *(ptr++)) != '\0') */

	} /* if (str != NULL */

	return str;

} /* gchar *cff_tl_replace_non_utf8_chars_in_str(gchar*) */

void cff_tl_file_path_local_to_unix(gchar *path)
{
	if (path != NULL)
	{
		while ((path = strchr(path, '\\')) != NULL)
		{
			*(path++) = '/';
		}
	}

} /* void cff_tl_file_path_local_to_unix(gchar*) */

void cff_tl_file_path_unix_to_local(gchar *path)
{
	if (path != NULL)
	{
		guint chr;

#ifdef G_OS_WIN32
		chr = '/';
#else
		chr = '\\';
#endif
		while ((path = strchr(path, chr)) != NULL)
		{
			*(path++) = G_DIR_SEPARATOR;
		}
	}

} /* void cff_tl_file_path_unix_to_local(gchar*) */

const char *cff_tl_file_get_basename(const gchar *fname)
{
	if (fname != NULL)
	{
		const gchar *sep, *s1;

		sep = strrchr(fname, '/');
		s1  = strrchr(fname, G_DIR_SEPARATOR);

		if (s1 > sep)
		{
			sep = s1;
		}

		if (sep != NULL)
		{
			return &sep[1];
		}

	} /* if (fname != NULL) */

	return fname;

} /* const char *cff_tl_file_get_basename(const gchar*) */

GFile *cff_tl_file_get_directory(const gchar *fname)
{
	GFile *file;

	if (fname == NULL)
	{
		return NULL;
	}

	file = g_file_parse_name(fname);

	while (file != NULL)
	{
		GFile *tmp;

		if (cff_tl_file_is_directory(file))
		{
			break;
		}

		tmp = g_file_get_parent(file);
		g_object_unref(G_OBJECT(file));
		file = tmp;

	} /* while (file != NULL) */

	return file;

} /* GFile *cff_tl_file_get_directory(const gchar*) */

gboolean cff_tl_file_is_directory(GFile *file)
{
	if (file != NULL)
	{
		if (g_file_query_file_type(file,
			G_FILE_QUERY_INFO_NONE, NULL) ==
						G_FILE_TYPE_DIRECTORY)
		{
			return TRUE;
		}

	} /* if (file != NULL) */

	return FALSE;

} /* gboolean cff_tl_file_is_directory(GFile*) */

static GFileInfo *cff_tl_file_info_etag(GFile *file)
{
	GFileInfo *info;
	GError    *error;

	if (file == NULL)
	{
		return NULL;
	}

	error = NULL;
	info  = g_file_query_info(file,
				G_FILE_ATTRIBUTE_ETAG_VALUE,
				G_FILE_QUERY_INFO_NONE,
				NULL,
				&error);
	if (error != NULL)
	{
		g_error_free(error);
	}

	return info;

} /* static GFileInfo *cff_tl_file_info_etag(GFile*) */

gboolean cff_tl_file_is_equal(GFile *file_src, GFile *file_dst)
{
	/* Test whether input and output are equal (symlink or hardlink) */

	gboolean  retv;
	GFileInfo *info_dst, *info_src;

	info_dst = cff_tl_file_info_etag(file_dst);

	if (info_dst == NULL)
	{
		return FALSE;

	}

	info_src = cff_tl_file_info_etag(file_src);
	retv     = FALSE;

	if (info_src != NULL)
	{
		const gchar *etag_dst = g_file_info_get_etag(info_dst);

		if (etag_dst != NULL && etag_dst[0] != '\0')
		{
			const gchar *etag_src = g_file_info_get_etag(info_src);

			if (etag_src != NULL && strcmp(etag_dst, etag_src) == 0)
			{
				retv = TRUE;
			}
		}

		g_object_unref(G_OBJECT(info_src));

	} /* if (info_src != NULL) */

	g_object_unref(G_OBJECT(info_dst));

	return retv;

} /* gboolean cff_tl_file_is_equal(GFile*, GFile*) */


#if GTK_MAJOR_VERSION == 4

GFile *cff_tl_file_append_fname(GFile *dir, const gchar *fname)
{
	GFile *file;

	if (dir == NULL)
	{
		return NULL;
	}

	file = NULL;

	if (fname != NULL && fname[0] != '\0')
	{
		GError *error;

		error = NULL;
		file  = g_file_get_child_for_display_name(dir, fname, &error);

		if (error != NULL)
		{
			g_error_free(error);
		}

	} /* if (fname != NULL && fname[0] != '\0') */

	g_object_unref(G_OBJECT(dir));

	return file;

} /* GFile *cff_tl_file_append_fname(GFile*, const gchar*) */

#endif /* GTK_MAJOR_VERSION == 4 */

/* EOF */
