/************************************************************************/
/* 		This code is part of 'CutterFF'				*/
/*		a tool for cutting video files with 'FFmpeg' and 'GTK'	*/
/*		Copyright (c) 2017 .. 2019,2021 .. 2024 Harald Foerster	*/
/*		(harald_foerster@users.sourceforge.net)			*/
/*		Distributed under GPL 2 or later			*/
/************************************************************************/

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#include <gdk/gdkkeysyms.h>
#include <glib/gstdio.h>

#include "cff_macros-gtk.h"
#include "cff_accel-gtk.h"
#include "cff_gprop.h"
#include "cff_nls.h"
#include "cff_charconv.h"
#include "cff_ffmpeg.h"

#ifdef HAVE_SYS_NPROCESSORS

#ifdef G_OS_WIN32
#include <windows.h>
#elif _MACOS
#include <sys/param.h>
#include <sys/sysctl.h>
#else
#include <unistd.h>
#endif

#endif /* HAVE_SYS_NPROCESSORS */

/*
#include <stdio.h>
#include <gtk/gtk.h>
#include <sys/types.h>
#include "cff_tools-gtk.h"
*/

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
#ifndef G_OS_WIN32
/* Application icon */
#include "../cutterff.xpm"
#endif
#endif

#if GTK_MAJOR_VERSION > 2

#ifndef HAVE_CAIRO
#define HAVE_CAIRO			1
#endif

#endif /* GTK_MAJOR_VERSION > 2 */


#ifdef G_OS_WIN32
#define CFF_IS_DIR_SEPARATOR(chr)	((chr) == '/' || (chr) == '\\')
#else
#define CFF_IS_DIR_SEPARATOR(chr)	((chr) == '/')
#endif

#ifdef GTK_TEXT_VIEW_PRIORITY_VALIDATE
#define CFF_TEXT_VIEW_PRIORITY_VALIDATE	GTK_TEXT_VIEW_PRIORITY_VALIDATE
#else
#define CFF_TEXT_VIEW_PRIORITY_VALIDATE	Gui_PRIORITY_REDRAW + 5
#endif


typedef struct
{
	GThread			*thread;
	GThreadFunc		func;
	gpointer		data;
	gpointer		result;
	guint			max;
	guint			num;
	volatile gint		step;
	GuiMutex		mutex;

} cff_thread_t;

static cff_thread_t gThread = { NULL };

typedef struct
{
	GtkWidget	*some;
	GtkWidget	*all;
	cff_ListStore_t	list;

} cff_streamlist_t;

typedef struct
{
	GtkWidget	*ts_from_duration;
	GtkWidget	*ts_adjust;
	GtkWidget	*bsf_option;
	cff_bsflist_t	*bsf_list;

} cff_mediaopts_t;

typedef struct
{
	GtkWidget		*dialog;
	GtkEntry		*entry;
	cff_video_t		*video;
	gchar			*filechooser_utf8str;
	gint			path_prop;
	guint			okay_bits;

	struct
	{
		GtkWidget	*option;
		gchar		*extension_orig;

		union
		{
			cff_fmtlist_t	*list;   /* Video */
			gchar		**names; /* Image */
		} u;

	} format;

} cff_save_common_t;

typedef struct
{
	cff_save_common_t	common;
	GtkWidget		*up;
	GtkWidget		*down;
	GtkWidget		*shortest;
	GtkWidget		*stream_sync;
	GtkWidget		*ts_keep_start;

	struct
	{
		GtkWidget	*option;
		GtkWidget	*both_gops;
		gchar		**encoder_names;

	} softcut;

	cff_streamlist_t	avail;
	cff_streamlist_t	out;
	cff_mediaopts_t		media[kCFFMediaTotal];

} cff_save_video_t;

typedef struct
{
	gint			bsfname_idx;      /* Filter to use */
	gboolean		ts_from_duration; /* Compute timestamps from duration */
	gboolean		ts_adjust;        /* Adjust broken timestamps */

} cff_save_media_prev_t;

typedef struct
{
	gint			*stream_idx;	/* List of stream id's */
	gint			nb_streams;	/* Number of streams */
	gint			format_idx;	/* Muxer to use */
	gint			encoder_idx;	/* Softcut => decode / encode */
	gboolean		shortest;	/* Finish with shortest stream duration */
	gboolean		ts_keep_start;	/* Keep original start timestamps */
	gboolean		stream_sync;	/* For better synchron */
	gboolean		both_gops;	/* Encode both (end/start) GOPs */
	cff_save_media_prev_t	media[kCFFMediaTotal];

} cff_save_video_prev_options_t;

static cff_save_video_prev_options_t *gSaveVideoPrevOptionsPtr = NULL;


typedef struct
{
	GtkSpinButton		*widget;
	gint			prev_value;

} cff_spinbutton_t;

typedef struct
{
	cff_save_common_t	common;
	cff_spinbutton_t	width;
	cff_spinbutton_t	height;
	GtkWidget		*keep_dimension;
	GtkWidget		*keep_ratio;
	GtkWidget		*gray;

} cff_save_image_t;

typedef struct
{
	volatile GtkProgressBar	*widget;
	GuiMutex		*mutex;

	union
	{
		gdouble		progress;

		struct
		{
			gint		decoded;
			gint		encoded;
			gboolean	show_text;

		} softcut;
	} u;

} cff_progressbar_callback_t;

typedef struct
{
	int (*func)(cff_video_t*, double*, const char**);

	cff_video_t			*video;
	const gchar			*title;
	GtkWidget			**mbar;
	GuiMutex			*mutex;
	GtkProgressBar			*progress_bar;
	
	/*
		Lock and data for 'cff_gui_dialog_progress_set_fraction'
		and 'cff_gui_dialog_progress_set_text'
	*/

	cff_progressbar_callback_t	cb;
	
	const gchar			*error;
	gint				retv;
	volatile gboolean		is_locked;

} cff_progressbar_t;

typedef struct
{
	GtkWidget	*window;
	GtkWidget	*textview;

	struct
	{
		GtkAdjustment	*h_widget;
		gdouble		v_upper;
		gdouble		v_pagesize;

	} scrollbar;

	struct
	{
		gchar		*str;
		guint		len;
		guint		max;
		guint		lastchar;
		guint		cr_len;
		gboolean	overwrite;

	} utf8;

	guint		nThreads;
	guint		timeout_id;
	gint		busy_loop;
	GuiMutex	mutex;

} cff_logwindow_t;

static cff_logwindow_t gLogWindow = { NULL };

static const gchar gLogWinTextColor[]       = "#00EE00";
static const gchar gLogWinBackgroundColor[] = "#000000";


/* Open, Program or Save */
#define kIsBusyFile		1
/* 'cff_gui_video_frame' */
#define kIsBusyVideoFrame	2

static gint gIsBusy = 0;


/* Ignore these events if busy */
#define kMouseButtonPress	0x01
#define kMouseButtonRelease	0x02

#define kMouseButtonAny		0
#define kMouseButtonLeft	1
#define kMouseButtonRight	3


/* Filter */

#define kFileFilterEntriesMax	10
#define kFileFilterCharsMax	15

typedef struct
{
	GtkWidget	*dialog;
	GtkFileChooser	*chooser;
	GtkWidget	*btn_add;
	GtkWidget	*btn_remove;
	GtkFileFilter	*buf[kFileFilterEntriesMax];
	guint		num;
	const gchar	*ext;
	gint		prop;

} cff_filefilter_t;

static const gchar cff_video_ext[] = "* avi mov mp4 mpg mpeg mkv ts webm";
static const gchar cff_image_ext[] = "* jpg";

typedef struct
{
	enum
	{
		kFrameSeekNext,
		kFrameSeekCutpoint,
		kFrameSeekSliderClick,
		kFrameSeekSliderMove

	} type;

	union
	{
		enum cff_VideoFrame	e_next;
		unsigned int		i_next;
		gdouble			d_next;

	} value;

} cff_frame_seek_t;


/*
	------------
	 PROTOTYPES
	------------
*/

#ifdef HAVE_GMENU
#define CFF_MENU_PROTOTYPE_ASTERIX(func) \
		func(GSimpleAction *a, GVariant *v, gpointer d)
#define CFF_MENU_PROTOTYPE_OBELIX(func) \
		func(GSimpleAction *a, GVariant *v, gpointer *d)
#else
#define CFF_MENU_PROTOTYPE_ASTERIX(func) \
		func(GtkWidget *w, gpointer d)
#define CFF_MENU_PROTOTYPE_OBELIX(func) \
		func(GtkWidget *w, gpointer *d)
#endif

/* File Menu */
static void CFF_MENU_PROTOTYPE_OBELIX(cff_gui_file_open);
static void CFF_MENU_PROTOTYPE_OBELIX(cff_gui_file_save);
static void CFF_MENU_PROTOTYPE_OBELIX(cff_gui_file_image);
static void CFF_MENU_PROTOTYPE_OBELIX(cff_gui_file_close);
static void CFF_MENU_PROTOTYPE_OBELIX(cff_gui_file_info);
static void CFF_MENU_PROTOTYPE_ASTERIX(cff_gui_file_save_rc);
static void CFF_MENU_PROTOTYPE_OBELIX(cff_gui_file_quit);

/* Edit Menu */
static void CFF_MENU_PROTOTYPE_ASTERIX(cff_gui_edit_undo);
static void CFF_MENU_PROTOTYPE_ASTERIX(cff_gui_edit_redo);
static void CFF_MENU_PROTOTYPE_OBELIX(cff_gui_edit_load);
static void CFF_MENU_PROTOTYPE_OBELIX(cff_gui_edit_save);
static void CFF_MENU_PROTOTYPE_OBELIX(cff_gui_edit_clear);
static void CFF_MENU_PROTOTYPE_ASTERIX(cff_gui_edit_add);
static void CFF_MENU_PROTOTYPE_ASTERIX(cff_gui_edit_remove);

/* Options Menu */
static void CFF_MENU_PROTOTYPE_OBELIX(cff_gui_opt_accel_dialog);
static void CFF_MENU_PROTOTYPE_ASTERIX(cff_gui_opt_toggle_bool);

/* Help Menu */
static void CFF_MENU_PROTOTYPE_OBELIX(cff_gui_help_ffmpeg);
static void CFF_MENU_PROTOTYPE_OBELIX(cff_gui_help_cutterff);
static void CFF_MENU_PROTOTYPE_OBELIX(cff_gui_help_about);

/* Submenus */
static void CFF_MENU_PROTOTYPE_ASTERIX(cff_gui_build_frame_submenu);
static void CFF_MENU_PROTOTYPE_ASTERIX(cff_gui_build_program_submenu);
static void CFF_MENU_PROTOTYPE_ASTERIX(cff_gui_build_loglevel_submenu);

/* Called from 'cff_gaccel.c' and 'cff_gui-gtk.c' */
gboolean CFF_MENU_PROTOTYPE_ASTERIX(cff_gui_adjust_menu);


/*
	-------------
	 Some Macros
	-------------
*/

#ifdef HAVE_GMENU
#define GuiSignalConnectMouseButton(widget, button, type, func, data) \
	cff_gui_signal_connect_mouse_button((widget), (button), \
			(type), (GCallback) (func), (data))
#else
#define GuiSignalConnectMouseButton(widget, button, type, func, data) \
	cff_gui_signal_connect_mouse_button((widget), \
			(type), (GCallback) (func), (data))
#endif

#ifdef HAVE_GMENU
#define GuiBuildMenuItems(item, menu, window, group) \
	cff_gui_build_menu_items((item), (menu), (window), (group))
#else
#define GuiBuildMenuItems(item, menu, window, group) \
	cff_gui_build_menu_items((item), (menu), (window))
#endif

#ifdef HAVE_GMENU
#define GuiBuildSubmenuItem(str, func, idx, disable, group) \
	cff_gui_build_submenu_item((str), (GCallback) (func), (idx), (disable), (group))
#else
#define GuiBuildSubmenuItem(str, func, idx, disable, group) \
	cff_gui_build_submenu_item((str), (GCallback) (func), (idx), (disable))
#endif

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

/* Prototype needed for 'cff_gui_calculate_frame_rects' */
static void cff_gui_widget_queue_draw_area(GtkWidget *widget);

#define GuiWidgetQueueDrawArea(widget) \
	cff_gui_widget_queue_draw_area(widget)
#else

/*
	Only 'gtk_widget_queue_draw()' is left
	to mark a widget as needing redraw
*/

#define GuiWidgetQueueDrawArea(widget) \
	gtk_widget_queue_draw(widget)

#endif /* #if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 .. #else */

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
#define GuiLogWindowFree(widget, event, data) \
	cff_gui_logwindow_free((widget), (event), (data))
#else
#define GuiLogWindowFree(widget, event, data) \
	cff_gui_logwindow_free((widget), (data))
#endif


/* cff_main.c */
extern char *cff_main_save_rc(void);


/* Function for drawing rectangles */

#ifdef HAVE_CAIRO

#if GTK_MAJOR_VERSION == 2

#define GuiDrawRectangle(cr, event, color, fill, rect) \
		cff_gui_cairo_draw_rect_gtk2((cr), \
			(GdkEventExpose *) (event), (color), (fill), (rect))
#else

#define GuiDrawRectangle(cr, event, color, fill, rect) \
		cff_gui_cairo_draw_rectangle((cr), (color), (fill), (rect))
#endif

#else /* ! HAVE_CAIRO */

#define GuiDrawRectangle(cr, event, color, fill, rect) \
		cff_gui_gdk_draw_rectangle((GdkEventExpose *) (event), \
						(color), (fill), (rect))

#endif /* #ifdef HAVE_CAIRO .. #else */


/*
	--------
	 CURSOR
	--------
*/

typedef struct
{

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
	GdkWindow	*window;
#else
	GtkWidget	*window;
#endif
	GdkCursor	*busy;

} cff_cursor_t;

static cff_cursor_t gCursor = { NULL };


/*
	-------------
	 VIDEO FRAME
	-------------
*/

typedef struct
{
	GtkWidget	*scale;
	GtkAdjustment	*adjust;
	gdouble		value;
	gdouble		x;
	gboolean	is_slider;
	volatile gint	button;

} cff_slider_t;

static cff_slider_t gSlider = { NULL };


typedef struct
{
	GtkWidget	*drawingArea;
	GtkLabel	*info;
	GtkLabel	*time;
	const guchar	**dataRGB24;
#ifdef HAVE_CAIRO
	cairo_surface_t	*surfaceRGB24;
#else
	gint		rowstrideRGB24;
#endif
	GdkRectangle	rectRGB24;
	GdkRectangle	rectLeftTop;
	GdkRectangle	rectRightBottom;

} cff_video_frame_t;

static cff_video_frame_t gCurrentVideoFrame;


static GtkWidget *gDrawingAreaBar;

enum
{
	kColorBlack,
	kColorDarkGrey,
	kColorGrey,
	kColorGreen,
	kColorRed,
	kColorTotal
};

#if GTK_MAJOR_VERSION == 2

#ifdef HAVE_CAIRO
static GdkColor	gColor[kColorTotal];
#else
static GdkGC *gContext[kColorTotal];
#endif

static const gchar gSizeAllocationStr[] = "size-allocate";
static const gchar gExposeEventStr[]	= "expose-event";

#else /* GTK_MAJOR_VERSION > 2 */

static GdkRGBA gRGBA[kColorTotal];

#if GTK_MAJOR_VERSION == 3
static const gchar gSizeAllocationStr[] = "size-allocate";
static const gchar gExposeEventStr[]    = "draw";
#else /* GTK_MAJOR_VERSION == 4 */
static const gchar gSizeAllocationStr[] = "resize";
#endif

#endif /* #if GTK_MAJOR_VERSION == 2 .. #else */


#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

static const gchar gButtonPressEventStr[] = "button-press-event";
static const gchar gDeleteEventStr[]      = "delete-event";

#else /* GTK_MAJOR_VERSION == 4 */

static const gchar gButtonPressEventStr[] = "clicked";
static const gchar gDeleteEventStr[]      = "close-request";

#endif /* #if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 .. #else */


/*
	----------------
	 BUTTON WIDGETS
	----------------
*/

enum
{
	kButtonVideoFramePrev4,
	kButtonVideoFramePrev3,
	kButtonVideoFramePrev2,
	kButtonVideoFramePrev1,
	kButtonVideoFrameNext1,
	kButtonVideoFrameNext2,
	kButtonVideoFrameNext3,
	kButtonVideoFrameNext4,
	kButtonVideoFrameTotal
};

static gchar *gButtonVideoFrameLabel[] =
{
	"|<",
	"<<<",
	"<<",
	"<",
	">",
	">>",
	">>>",
	">|"
};

static GtkWidget *gui_button_video_frame[kButtonVideoFrameTotal];

enum
{
	kButtonVideoCutAdd,
	kButtonVideoCutRemove,
	kButtonVideoCutTotal
};

static GtkWidget *gButtonVideoCut[kButtonVideoCutTotal];

#ifdef HAVE_GMENU
#define GuiEditAddCutFunc	cff_gui_edit_add_clicked
#define GuiEditRemoveCutFunc	cff_gui_edit_remove_clicked
#else
#define GuiEditAddCutFunc	cff_gui_edit_add
#define GuiEditRemoveCutFunc	cff_gui_edit_remove
#endif


/*
	--------------
	 MENU WIDGETS
	--------------
*/

enum
{
	kMenuBarFile,
	kMenuBarEdit,
	kMenuBarOption,
	kMenuBarHelp,
	kMenuBarTotal
};

/* Enbled or disabled */

enum
{
	kMenuFileSave,
	kMenuFileImage,
	kMenuFileClose,
	kMenuFileInfo,
	kMenuEditUndo,
	kMenuEditRedo,
	kMenuEditLoad,
	kMenuEditSave,
	kMenuEditClear,
	kMenuEditAdd,
	kMenuEditRemove,
	kMenuEditGoto,
	kMenuEditProgram,
	kMenuItemsTotal
};

enum
{
	kPopupFileSave,
	kPopupEditUndo,
	kPopupEditRedo,
	kPopupEditLoad,
	kPopupEditSave,
	kPopupEditClear,
	kPopupEditGoto,
	kPopupItemsTotal
};

static GtkWidget     *gMenuBarTitle[kMenuBarTotal] = { NULL };
static GuiMenuAction *gMenuAction[kMenuItemsTotal];
static GuiMenuAction *gPopupAction[kPopupItemsTotal];

/* kPropBoolOptKeyFrame, kPropBoolOptIFrame and kPropBoolOptPFrame */
static GuiMenuAction *gCheckmenuFrameAction[3];
static GuiMenuAction *gPopupCheckmenuFrameAction[3];

static GtkWidget *gPopupMenuWidget;

#ifdef HAVE_GMENU

/* The submenus of 'goto', 'program' and 'loglevel' */
static GMenu *gSubmenuMbar[3] = { NULL };

/* The submenu of 'goto' */
static GMenu *gSubmenuPopup = NULL;

#endif /* HAVE_GMENU */


/*
	------------
	 MENU TYPES
	------------
*/

enum
{
	kMTypeEndOfList,
	kMTypeItem,
	kMTypeCheckItem,
	kMTypeSeparator,
	kMTypeSubmenuItem
};


/*
	------------
	 MENU ITEMS
	------------
*/

static const gchar mn_file_open_str[]	= gettext_nop("Open ...");
static const gchar mn_file_save_str[]	= gettext_nop("Save ...");
static const gchar mn_edit_undo_str[]	= gettext_nop("Undo");
static const gchar mn_edit_redo_str[]	= gettext_nop("Redo");
static const gchar mn_edit_load_str[]	= gettext_nop("Load list ...");
static const gchar mn_edit_save_str[]	= gettext_nop("Save list ...");
static const gchar mn_edit_clear_str[]	= gettext_nop("Clear list ...");
static const gchar mn_edit_goto_str[]	= gettext_nop("Goto");
static const gchar mn_opt_keyframe_str[]= gettext_nop("Key-Frame");
static const gchar mn_opt_iframe_str[]	= gettext_nop("I-Frame");
static const gchar mn_opt_pframe_str[]	= gettext_nop("P-Frame");

typedef struct
{
	gint		type;		/* See above */
	const gchar	*label;		/* Name to display */
	gint		accel_id;	/* Accel key */
	GuiMenuAction   **storage;	/* gMenuAction */
	GCallback	callback;	/* Function pointer */
	gpointer	data;		/* Data for it */

} cff_guiMenuItem_t;

static const cff_guiMenuItem_t cff_gui_menuFile[] =
{
	{kMTypeItem, mn_file_open_str,
		kAccelFileOpen, NULL,
		(GCallback) cff_gui_file_open, &gMenuBarTitle[kMenuBarFile]},
	{kMTypeItem, mn_file_save_str,
		kAccelFileSave, &gMenuAction[kMenuFileSave],
		(GCallback) cff_gui_file_save, &gMenuBarTitle[kMenuBarFile]},
	{kMTypeItem, gettext_nop("Save image ..."),
		kAccelFileImage, &gMenuAction[kMenuFileImage],
		(GCallback) cff_gui_file_image, &gMenuBarTitle[kMenuBarFile]},
	{kMTypeSeparator, NULL, kAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Close"),
		kAccelFileClose, &gMenuAction[kMenuFileClose],
		(GCallback) cff_gui_file_close, &gMenuBarTitle[kMenuBarFile]},
	{kMTypeItem, gettext_nop("Info ..."),
		kAccelFileInfo, &gMenuAction[kMenuFileInfo],
		(GCallback) cff_gui_file_info, &gMenuBarTitle[kMenuBarFile]},
	{kMTypeSeparator, NULL, kAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Save rc file"),
		kAccelFileSaveRC, NULL,
		(GCallback) cff_gui_file_save_rc, NULL},
	{kMTypeSeparator, NULL, kAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Quit"),
		kAccelFileQuit, NULL,
		(GCallback) cff_gui_file_quit, &gMenuBarTitle[kMenuBarFile]},
	{kMTypeEndOfList, NULL, kAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuFile[] */

static const cff_guiMenuItem_t cff_gui_menuEdit[] =
{
	{kMTypeItem, mn_edit_undo_str,
		kAccelEditUndo, &gMenuAction[kMenuEditUndo],
		(GCallback) cff_gui_edit_undo, &gVideo},
	{kMTypeItem, mn_edit_redo_str,
		kAccelEditRedo, &gMenuAction[kMenuEditRedo],
		(GCallback) cff_gui_edit_redo, &gVideo},
	{kMTypeSeparator, NULL, kAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_edit_load_str,
		kAccelEditLoad, &gMenuAction[kMenuEditLoad],
		(GCallback) cff_gui_edit_load, &gMenuBarTitle[kMenuBarEdit]},
	{kMTypeItem, mn_edit_save_str,
		kAccelEditSave, &gMenuAction[kMenuEditSave],
		(GCallback) cff_gui_edit_save, &gMenuBarTitle[kMenuBarEdit]},
	{kMTypeItem, mn_edit_clear_str,
		kAccelEditClear, &gMenuAction[kMenuEditClear],
		(GCallback) cff_gui_edit_clear, &gMenuBarTitle[kMenuBarEdit]},
	{kMTypeSeparator, NULL, kAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("Add"),
		kAccelEditAdd, &gMenuAction[kMenuEditAdd],
		(GCallback) cff_gui_edit_add, &gVideo},
	{kMTypeItem, gettext_nop("Remove"),
		kAccelEditRemove, &gMenuAction[kMenuEditRemove],
		(GCallback) cff_gui_edit_remove, &gVideo},
	{kMTypeSeparator, NULL, kAccelKeyNone, NULL, NULL, NULL},
	{kMTypeSubmenuItem, mn_edit_goto_str,
		kAccelKeyNone, &gMenuAction[kMenuEditGoto],
		(GCallback) cff_gui_build_frame_submenu, &gVideo},
	{kMTypeSeparator, NULL, kAccelKeyNone, NULL, NULL, NULL},
	{kMTypeSubmenuItem, gettext_nop("Program"),
		kAccelKeyNone, &gMenuAction[kMenuEditProgram],
		(GCallback) cff_gui_build_program_submenu, &gVideo},
	{kMTypeEndOfList, NULL, kAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuEdit[] */

static const cff_guiMenuItem_t cff_gui_menuOptions[] =
{
	{kMTypeItem, gettext_nop("Keyboard ..."),
		kAccelOptAccel, NULL,
		(GCallback) cff_gui_opt_accel_dialog, &gMenuBarTitle[kMenuBarOption]},
	{kMTypeSeparator, NULL, kAccelKeyNone, NULL, NULL, NULL},
	{kMTypeCheckItem, mn_opt_keyframe_str,
		kAccelOptKeyFrame, &gCheckmenuFrameAction[0],
		(GCallback) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptKeyFrame)},
	{kMTypeCheckItem, mn_opt_iframe_str,
		kAccelOptIFrame, &gCheckmenuFrameAction[1],
		(GCallback) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptIFrame)},
	{kMTypeCheckItem, mn_opt_pframe_str,
		kAccelOptPFrame, &gCheckmenuFrameAction[2],
		(GCallback) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptPFrame)},
	{kMTypeCheckItem, gettext_nop("Cursor"),
		kAccelOptCursor, NULL,
		(GCallback) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptCursor)},
	{kMTypeCheckItem, gettext_nop("Overwrite"),
		kAccelOptOverwrite, NULL,
		(GCallback) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptOverwrite)},
	{kMTypeCheckItem, gettext_nop("Delete copy on error"),
		kAccelOptDelete, NULL,
		(GCallback) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptDelete)},
	{kMTypeCheckItem, gettext_nop("Beep"),
		kAccelOptBeep, NULL,
		(GCallback) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptBeep)},
	{kMTypeCheckItem, gettext_nop("Save rc on exit"),
		kAccelOptSave, NULL,
		(GCallback) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptSaveConfig)},
	{kMTypeSubmenuItem, gettext_nop("Log ..."),
		kAccelKeyNone, NULL,
		(GCallback) cff_gui_build_loglevel_submenu, NULL},
	{kMTypeEndOfList, NULL, kAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuOptions[] */

static const cff_guiMenuItem_t cff_gui_menuHelp[] =
{
	{kMTypeItem, gettext_nop("FFmpeg ..."),
		kAccelHelpFFmpeg, NULL,
		(GCallback) cff_gui_help_ffmpeg, &gMenuBarTitle[kMenuBarHelp]},
	{kMTypeSeparator, NULL, kAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, gettext_nop("CutterFF ..."),
		kAccelHelpCutterff, NULL,
		(GCallback) cff_gui_help_cutterff, &gMenuBarTitle[kMenuBarHelp]},
	{kMTypeItem, gettext_nop("About ..."),
		kAccelHelpAbout, NULL,
		(GCallback) cff_gui_help_about, &gMenuBarTitle[kMenuBarHelp]},
	{kMTypeEndOfList, NULL, kAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuHelp[] */

static const cff_guiMenuItem_t cff_gui_menuPopup[] =
{
	{kMTypeItem, mn_edit_undo_str,
		kAccelKeyNone,  &gPopupAction[kPopupEditUndo],
		(GCallback) cff_gui_edit_undo, &gVideo},
	{kMTypeItem, mn_edit_redo_str,
		kAccelKeyNone,  &gPopupAction[kPopupEditRedo],
		(GCallback) cff_gui_edit_redo, &gVideo},
	{kMTypeSeparator, NULL, kAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_file_open_str,
		kAccelKeyNone, NULL,
		(GCallback) cff_gui_file_open, &gMenuBarTitle[kMenuBarFile]},
	{kMTypeItem, mn_file_save_str,
		kAccelKeyNone, &gPopupAction[kPopupFileSave],
		(GCallback) cff_gui_file_save, &gMenuBarTitle[kMenuBarFile]},
	{kMTypeSeparator, NULL, kAccelKeyNone, NULL, NULL, NULL},
	{kMTypeItem, mn_edit_load_str,
		kAccelKeyNone, &gPopupAction[kPopupEditLoad],
		(GCallback) cff_gui_edit_load, &gMenuBarTitle[kMenuBarEdit]},
	{kMTypeItem, mn_edit_save_str,
		kAccelKeyNone, &gPopupAction[kPopupEditSave],
		(GCallback) cff_gui_edit_save, &gMenuBarTitle[kMenuBarEdit]},
	{kMTypeItem, mn_edit_clear_str,
		kAccelKeyNone, &gPopupAction[kPopupEditClear],
		(GCallback) cff_gui_edit_clear, &gMenuBarTitle[kMenuBarEdit]},
	{kMTypeSeparator, NULL, kAccelKeyNone, NULL, NULL, NULL},
	{kMTypeSubmenuItem, mn_edit_goto_str,
		kAccelKeyNone, &gPopupAction[kPopupEditGoto],
		(GCallback) cff_gui_build_frame_submenu, &gVideo},
	{kMTypeSeparator, NULL, kAccelKeyNone, NULL, NULL, NULL},
	{kMTypeCheckItem, mn_opt_keyframe_str,
		kAccelKeyNone, &gPopupCheckmenuFrameAction[0],
		(GCallback) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptKeyFrame)},
	{kMTypeCheckItem, mn_opt_iframe_str,
		kAccelKeyNone, &gPopupCheckmenuFrameAction[1],
		(GCallback) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptIFrame)},
	{kMTypeCheckItem, mn_opt_pframe_str,
		kAccelKeyNone, &gPopupCheckmenuFrameAction[2],
		(GCallback) cff_gui_opt_toggle_bool, GINT_TO_POINTER(kPropBoolOptPFrame)},
	{kMTypeEndOfList, NULL, kAccelKeyNone, NULL, NULL, NULL}

}; /* static const cff_guiMenuItem_t cff_gui_menuPopup[] */


/* ------------------- */


/*
	------
	 MISC
	------
*/


static void cff_gui_slider_adjustment_configure(guint upper)
{
	/*
		Using the same upper and lower bounds for
		the GtkScale will hide the slider itself
	*/

	if (upper == 0)
	{
		upper = 1;
	}

	gtk_adjustment_configure(gSlider.adjust,
				0.0,		/* Value */
				0.0,		/* Lower */
				(gdouble) upper,
				0.0,		/* Step_increment */
				0.0,		/* Page_increment */
				0.0);		/* Page_size */

} /* static void cff_gui_slider_adjustment_configure(guint) */

static gboolean cff_gui_slider_adjustment_set_value(cff_slider_t *slider)
{
	gtk_adjustment_set_value(slider->adjust,
			(gdouble) gVideo.vframe.frame_current);

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_slider_adjustment_set_value(cff_slider_t*) */

static GtkWidget *cff_gui_window_get_topwin(void)
{
	if (gWindata.topwin == gLogWindow.window)
	{
		return gWindata.mainwin;
	}

	return gWindata.topwin;

} /* static GtkWidget *cff_gui_window_get_topwin(void) */


#ifdef HAVE_GMENU

static gboolean cff_gui_contoller_key_pressed(GtkEventControllerKey *cntrl,
		guint keyval, guint keycode, GdkModifierType state, gint *isBusy)
{
	/* Supress key press event if not the main level */

	if (*isBusy != 0 || GuiMainLevelGetBusy())
	{
		return TRUE;
	}

	return FALSE;

} /* static gboolean cff_gui_contoller_key_pressed(GtkEventControllerKey*,
					guint, guint, GdkModifierType, gint*) */

static void cff_gui_contoller_mouse_button_pressed(GtkGesture *gesture,
				GdkEventSequence *sequence, gint *isBusy)
{
	/* Supress mouse button event if not the main level */

	if (*isBusy != 0 || GuiMainLevelGetBusy())
	{
		gtk_gesture_set_state(gesture, GTK_EVENT_SEQUENCE_CLAIMED);
	}

} /* static void cff_gui_contoller_mouse_button_pressed(GtkGesture*,
						GdkEventSequence*, gint*) */

static void cff_gui_signal_connect_key_pressed(GtkWidget *widget,
					GCallback func, gpointer data)
{
	GtkEventController *controller;

#if GTK_MAJOR_VERSION == 3
	controller = gtk_event_controller_key_new(widget);
#else
	controller = gtk_event_controller_key_new();
	gtk_widget_add_controller(widget, controller);
#endif

	/* Ignore this event if busy */
	GuiSignalConnect(controller, "key-pressed",
			cff_gui_contoller_key_pressed, &gIsBusy);

	if (func != NULL)
	{
		GuiSignalConnect(controller, "key-pressed", func, data);
	}

} /* static void cff_gui_signal_connect_key_pressed(GtkWidget*, GCallback, gpointer) */

static void cff_gui_signal_connect_mouse_button(GtkWidget *widget,
			gint button, gint type, GCallback func, gpointer data)
{
	GtkGesture *gesture;

#if GTK_MAJOR_VERSION == 3
	gesture = gtk_gesture_multi_press_new(widget);
#else
	gesture = gtk_gesture_click_new();
	gtk_widget_add_controller(widget, GTK_EVENT_CONTROLLER(gesture));
#endif

	/* Ignore this event if busy */
	gtk_event_controller_set_propagation_phase(
		GTK_EVENT_CONTROLLER(gesture), GTK_PHASE_CAPTURE);

	GuiSignalConnect(gesture, "begin",
		cff_gui_contoller_mouse_button_pressed, &gIsBusy);

	gtk_gesture_single_set_button(
			GTK_GESTURE_SINGLE(gesture), button);

	if (func != NULL)
	{
		GuiSignalConnect(gesture, "pressed", func, data);

		if (type & kMouseButtonRelease)
		{
			/* GTK-4: no signal emitted if widget is GtkScale */
			GuiSignalConnect(gesture, "released", func, data);
		}
	}

} /* static void cff_gui_signal_connect_mouse_button(GtkWidget*,
					gint, gint, GCallback, gpointer) */

#else /* ! HAVE_GMENU */

static gboolean cff_gui_key_or_mouse_button_pressed(GtkWidget *widget,
						GdkEvent *event, gint *isBusy)
{
	/* Supress key press or mouse button event if not the main level */

	if (*isBusy != 0 || GuiMainLevelGetBusy())
	{
		return TRUE;
	}

	return FALSE;

} /* static gboolean cff_gui_key_or_mouse_button_pressed(GtkWidget*,
							GdkEvent*, gint*) */

static void cff_gui_signal_connect_key_pressed(GtkWidget *widget,
					GCallback func, gpointer data)
{
	/* Ignore this event if busy */
	GuiSignalConnect(widget, "key-press-event",
		cff_gui_key_or_mouse_button_pressed, &gIsBusy);

	if (func != NULL)
	{
		GuiSignalConnect(widget, "key-press-event", func, data);
	}

} /* static void cff_gui_signal_connect_key_pressed(GtkWidget*,
						GCallback, gpointer) */

static void cff_gui_signal_connect_mouse_button(GtkWidget *widget,
				gint type, GCallback func, gpointer data)
{
	/* Ignore this event if busy */
	GuiSignalConnect(widget, "button-press-event",
		cff_gui_key_or_mouse_button_pressed, &gIsBusy);

	if (func != NULL)
	{
		GuiSignalConnect(widget, "button-press-event", func, data);

		if (type & kMouseButtonRelease)
		{
			GuiSignalConnect(widget, "button-release-event",
								func, data);
		}
	}

} /* static void cff_gui_signal_connect_mouse_button(GtkWidget*,
					gint, GCallback, gpointer) */

#endif /* #ifdef HAVE_GMENU .. #else */


static void cff_gui_signal_connect_key_ignore(GtkWidget *widget)
{
	/* Ignore this events if busy */
	cff_gui_signal_connect_key_pressed(widget, NULL, NULL);

} /* static void cff_gui_signal_connect_key_ignore(GtkWidget*) */

static void cff_gui_signal_connect_mouse_button_ignore(GtkWidget *widget)
{
	/* Ignore this events if busy */
	GuiSignalConnectMouseButton(widget, 0, 0, NULL, NULL);

} /* static void cff_gui_signal_connect_mouse_button_ignore(GtkWidget*) */

static gboolean cff_gui_overwrite(GFile *file)
{
	int overwrite;

	if (cff_gprop_get_bool(kPropBoolOptOverwrite, &overwrite) && overwrite)
	{
		return TRUE;
	}

	if (g_file_query_exists(file, NULL) == FALSE)
	{
		return TRUE;
	}

	/* TRUE == GTK_RESPONSE_ACCEPT */
	return cff_guitl_dialog_warning_2btn(
			gettext_nop("File already exists! Overwrite?"),
					cff_gui_window_get_topwin(), NULL);

} /* static gboolean cff_gui_overwrite(GFile*) */

static gboolean cff_gui_file_is_equal(GFile *file_dst, const gchar *fname_src, gboolean stop)
{
	/* Test whether input and output are equal */

	static const gchar error_equal_fname[] =
		gettext_nop("The file- and/or foldername must differ!");

	static const gchar error_equal_file[] =
		gettext_nop("The input and output file is the same!");

	gchar		*fname_dst;
	const gchar	*error_str;

	if (file_dst == NULL || fname_src == NULL)
	{
		return FALSE;
	}

	fname_dst = g_file_get_parse_name(file_dst);
	error_str = NULL;

	if (fname_dst != NULL && strcmp(fname_dst, fname_src) == 0)
	{
		error_str = error_equal_fname;
	}

	else
	{
		GFile *file_src = g_file_parse_name(fname_src);

		if (file_src != NULL)
		{
			if (cff_tl_file_is_equal(file_src, file_dst))
			{
				error_str = error_equal_file;
			}

			g_object_unref(G_OBJECT(file_src));
		}

	} /* if (fname_dst != NULL && strcmp(fname_dst, fname_src) == 0) .. else */

	g_free(fname_dst);

	if (error_str != NULL)
	{
		if (stop == FALSE)
		{
			cff_guitl_dialog_warning(error_str, NULL);
		}

		else
		{
			cff_guitl_dialog_error(error_str, NULL);
		}

		cff_guitl_beep();
		return TRUE;

	} /* if (error_str != NULL) */

	return FALSE;

} /* static gboolean cff_gui_file_is_equal(GFile*, const gchar*, gboolean) */

static gboolean cff_gui_file_discard_changes(GtkWidget **mbar)
{
	if (gVideo.dirty)
	{
		/* TRUE == GTK_RESPONSE_ACCEPT */
		return cff_guitl_dialog_warning_2btn(
				gettext_nop("Discard changes?"),
					cff_gui_window_get_topwin(), mbar);
	}

	return TRUE;

} /* static gboolean cff_gui_file_discard_changes(GtkWidget**) */

static gboolean cff_gui_thread_remove_me(gpointer data)
{
	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_thread_remove_me(gpointer) */

static gpointer cff_gui_thread_wait(cff_thread_t *thread)
{
	for (;;)
	{
		GuiMutexLock(&(thread->mutex));

		GuiMutexUnlock(&(thread->mutex));
		thread->step = 1;

		if (thread->func == NULL)
		{
			/* Quit */
			break;
		}

		gLogWindow.nThreads++;
		thread->result = (*thread->func)(thread->data);
		gLogWindow.nThreads--;

		while (thread->step != 2) {;}

		/*
			Calling 'gtk_main_quit' directly may
			freeze the program, if the thread
			function returns before 'gtk_main' in the
			main thread has its job successfully done.
		*/

		g_idle_add_full(G_PRIORITY_HIGH_IDLE,
			(GSourceFunc) cff_gui_thread_remove_me,
				NULL, (GDestroyNotify) GuiMainLevelQuitFunc);
	} /* for (;;) */

	return NULL;

} /* gpointer cff_gui_thread_wait(cff_thread_t*) */

static gpointer cff_gui_thread_invoke(GThreadFunc func, gpointer data)
{
	cff_thread_t *thread = &gThread;

	if (thread->thread == NULL)
	{
		return (*func)(data);
	}

	thread->func = func;
	thread->data = data;

	thread->step = 0;
	GuiMutexUnlock(&(thread->mutex));

	while (thread->step != 1) {;}

	GuiMutexLock(&(thread->mutex));
	thread->step = 2;

	GuiMainLevelStart();

	return thread->result;

} /* static gpointer cff_gui_thread_invoke(GThreadFunc, gpointer) */

static gpointer cff_gui_thread_run(gpointer data)
{
	gpointer result;

	struct
	{
		GThreadFunc	func;
		gpointer	data;

	} *s = data;

	gLogWindow.nThreads++;
	result = (*s->func)(s->data);
	gLogWindow.nThreads--;

	GuiMainLevelQuit();

	return result;

} /* static gpointer cff_gui_thread_run(gpointer) */

static gboolean cff_gui_thread_new(GThreadFunc func, gpointer data, gpointer *retv)
{
	cff_thread_t *thread = &gThread;

	if (thread->max > thread->num)
	{
		int nThreads;

#ifdef HAVE_GNUMPROCESSORS
		nThreads = g_get_num_processors();
#else

#ifdef HAVE_SYS_NPROCESSORS

#ifdef G_OS_WIN32

		SYSTEM_INFO sysinfo;

		GetSystemInfo(&sysinfo);
		nThreads = sysinfo.dwNumberOfProcessors;
#elif _MACOS
		size_t		len;
		uint32_t	count;
		int		nm[2];

		len   = sizeof(uint32_t);
		nm[0] = CTL_HW;
		nm[1] = HW_AVAILCPU;

		sysctl(nm, 2, &count, &len, NULL, 0);

		if (count < 1)
		{
			nm[1] = HW_NCPU;
			sysctl(nm, 2, &count, &len, NULL, 0);
		}

		nThreads = count;

#else /* UNIX */

		nThreads = sysconf(_SC_NPROCESSORS_ONLN);
#endif

#else
		nThreads = 0;

#endif /* #ifdef HAVE_SYS_NPROCESSORS .. #else */
#endif /* #ifdef HAVE_GNUMPROCESSORS .. #else */


		if (nThreads > thread->num)
		{
			if (func == NULL)
			{
				GuiMutexLock(&(thread->mutex));

				thread->thread =
					GuiThreadNew(cff_gui_thread_wait, thread);

				if (thread->thread != NULL)
				{
					thread->num++;
					return TRUE;
				}

				thread->max = 0;
				GuiMutexUnlock(&(thread->mutex));
			}

			else
			{
				GThread *new_thread;

				struct
				{
					GThreadFunc	func;
					gpointer	data;
				} s;

				s.func = func;
				s.data = data;

				new_thread =
					GuiThreadNew(cff_gui_thread_run, &s);

				if (new_thread != NULL)
				{
					gpointer result;

					GuiMainLevelStart();

					thread->num++;
					result = g_thread_join(new_thread);
					thread->num--;

					if (retv != NULL)
					{
						*retv = result;
					}

					return TRUE;
				}

			} /* if (func == NULL) .. else */

		} /* if (nThreads > thread->num) */

	} /* if (thread->max > thread->num) */

	return FALSE;

} /* static gboolean cff_gui_thread_new(GThreadFunc, gpointer, gpointer*) */


/*
	--------------
	 PROGRESS-BAR
	--------------
*/

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
static gboolean cff_gui_dialog_progress_halt(GtkWidget *w, GdkEvent *e, cff_progressbar_t *data)
#else
static gboolean cff_gui_dialog_progress_halt(GtkWidget *w, cff_progressbar_t *data)
#endif
{
	/* 'delete-event' or 'button-press-event*/

#if GTK_MAJOR_VERSION == 4
	if (data->progress_bar == NULL)
	{
		/* Not the close button */
		return FALSE;
	}
#endif
	/* Stop */
	data->progress_bar = NULL;

	return TRUE;

} /* static gboolean cff_gui_dialog_progress_halt(GtkWidget*, GdkEvent*, cff_progressbar_t*) */

static gboolean cff_gui_dialog_progress_set_text(cff_progressbar_callback_t *cb)
{
	if (cb != NULL)
	{
		if (GuiMainLevelGetBusy())
		{
			GtkProgressBar *pbar;

			if (cb->mutex != NULL)
			{
				GuiMutexLock(cb->mutex);
			}

			pbar = (GtkProgressBar *) cb->widget;

			if (pbar != NULL)
			{
				if (cb->u.softcut.decoded < 0 || cb->u.softcut.encoded < 0)
				{
						cb->u.softcut.show_text = FALSE;
						GuiProgressbarTextHide(pbar);
				}

				else
				{
					gchar str[256];

					if (cb->u.softcut.show_text == FALSE)
					{
						cb->u.softcut.show_text = TRUE;
#if GTK_MAJOR_VERSION > 2
						gtk_progress_bar_set_show_text(pbar, 1);
#endif
					}

					g_snprintf(str, sizeof(str), "d: %i / e: %i",
							cb->u.softcut.decoded, cb->u.softcut.encoded);
					gtk_progress_bar_set_text(pbar, str);
				}

			} /* if (pbar != NULL) */

			cb->widget = NULL;

			if (cb->mutex == NULL)
			{
				while (g_main_context_iteration(NULL, FALSE)) {;}
			}

			else
			{
				GuiMutexUnlock(cb->mutex);
			}
		}

		else
		{
			cb->widget = NULL;

		} /* if (GuiMainLevelGetBusy()) .. else */

	} /* if (cb != NULL) */

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_dialog_progress_set_text(cff_progressbar_callback_t*) */

static int cff_gui_dialog_progress_softcut_cb(cff_softcut_pbar_t *data)
{
	/*
		Called from 'cff_ffmpeg_copy_softcut' to
		display the number of decoded and encoded
		frames.
	*/

	cff_progressbar_t *pbar, **pbarhdl;

	if (data == NULL)
	{
		return 0;
	}

	pbarhdl = data->func_data;

	if (pbarhdl == NULL)
	{
		return 0;
	}

	pbar = *pbarhdl;

	if (pbar == NULL)
	{
		return 0;
	}

	if (pbar->progress_bar == NULL)
	{
		/* Stop */
		return -1;
	}

	if (pbar->cb.mutex != NULL)
	{
		GuiMutexLock(pbar->cb.mutex);
	}

	if (pbar->cb.widget == NULL)
	{
		pbar->cb.widget = pbar->progress_bar;

		pbar->cb.u.softcut.decoded = data->decoded;
		pbar->cb.u.softcut.encoded = data->encoded;

		g_main_context_invoke_full(NULL, Gui_PRIORITY_REDRAW,
			(GSourceFunc) cff_gui_dialog_progress_set_text, &(pbar->cb), NULL);
	}

	if (pbar->cb.mutex != NULL)
	{
		GuiMutexUnlock(pbar->cb.mutex);
	}

	return 0;

} /* static int cff_gui_dialog_progress_softcut_cb(cff_softcut_pbar_t*) */

static gboolean cff_gui_dialog_progress_set_fraction(cff_progressbar_callback_t *cb)
{
	if (cb != NULL)
	{
		if (GuiMainLevelGetBusy())
		{
			GtkProgressBar *pbar;

			if (cb->mutex != NULL)
			{
				GuiMutexLock(cb->mutex);
			}

			pbar = (GtkProgressBar *) cb->widget;

			if (pbar != NULL)
			{
				if (cb->u.softcut.show_text)
				{
					cb->u.softcut.show_text = FALSE;
					GuiProgressbarTextHide(pbar);
				}

				gtk_progress_bar_set_fraction(pbar, cb->u.progress);
			}

			cb->widget = NULL;

			if (cb->mutex != NULL)
			{
				GuiMutexUnlock(cb->mutex);
			}
		}

		else
		{
			cb->widget = NULL;

		} /* if (GuiMainLevelGetBusy()) .. else */

	} /* if (cb != NULL) */

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_dialog_progress_set_fraction(cff_progressbar_callback_t*) */

static gboolean cff_gui_dialog_progress_callback(cff_progressbar_t *pbar)
{
	double progress;

	if (pbar->cb.mutex != NULL)
	{
		GuiMutexInit(pbar->cb.mutex);
	}

	while ((pbar->retv = (*pbar->func)(pbar->video,
					&progress, &(pbar->error))) == 0)
	{
		if (pbar->error != NULL || pbar->progress_bar == NULL)
		{
			break;
		}

		if (pbar->cb.mutex != NULL)
		{
			GuiMutexLock(pbar->cb.mutex);
		}

		if (pbar->cb.widget == NULL)
		{
			pbar->cb.widget     = pbar->progress_bar;
			pbar->cb.u.progress = (gdouble) progress;

			g_main_context_invoke_full(NULL, G_PRIORITY_HIGH_IDLE,
				(GSourceFunc) cff_gui_dialog_progress_set_fraction, &(pbar->cb), NULL);
		}

		if (pbar->cb.mutex == NULL)
		{
			return Gui_SOURCE_CONTINUE;
		}

		GuiMutexUnlock(pbar->cb.mutex);
	}

	/* Wait */

	while (pbar->cb.widget != NULL)
	{
		g_main_context_wakeup(NULL);
		g_main_context_iteration(NULL, FALSE);
	}

	if (pbar->cb.mutex != NULL)
	{
		GuiMutexFree(pbar->cb.mutex);
	}

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_dialog_progress_callback(cff_progressbar_t*) */

static gboolean cff_gui_dialog_progress_bar(cff_progressbar_t *data)
{
	GtkWidget *widget, *window, *button, *frame, *parent, *hbox, *vbox;
	GuiMutex  mutex;

	if (data->mutex)
	{
		/* 'cff_gui_open_callback' */
		GuiMutexLock(data->mutex);
		data->is_locked = TRUE;
	}

	vbox = GuiVBoxNew(FALSE, 4);
	GuiContainerSetBorderWidth(vbox, 4);

	parent = cff_gui_window_get_topwin();
	window = cff_guitl_window_new_gettext(data->title, vbox, 8);
	cff_guitl_dialog_set_modal(window, parent, data->mbar);
	gtk_window_set_resizable(GTK_WINDOW(window), FALSE);

	hbox = GuiHBoxNew(FALSE, 4);
	GuiContainerSetBorderWidth(hbox, 4);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, 2, hbox);
	GuiBoxAddChild(vbox, frame);

	widget = gtk_progress_bar_new();
	data->progress_bar = GTK_PROGRESS_BAR(widget);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 4);

#if GTK_MAJOR_VERSION == 4
	gtk_widget_show(window);
	gtk_widget_set_size_request(widget, 200, -1);
#endif

	button = gtk_button_new_with_label(gettext_nop("Halt"));
	GuiHBoxPackStart(hbox, button, FALSE, FALSE, 1);

	GuiWidgetCanDefault(button);

	GuiSignalConnect(window, gDeleteEventStr,
			cff_gui_dialog_progress_halt, data);
	GuiSignalConnect(button,  gButtonPressEventStr,
			cff_gui_dialog_progress_halt, data);

	GuiWidgetShowAll(window);

	data->cb.u.softcut.show_text = FALSE;

	/* Callback */
	data->cb.widget = NULL;
	data->cb.mutex  = &mutex;

	/* Halt-button */
	data->retv = 0;

	if (cff_gui_thread_new((GThreadFunc)
		cff_gui_dialog_progress_callback, data, NULL) == FALSE)
	{
		data->cb.mutex = NULL;

		g_idle_add_full(Gui_PRIORITY_REDRAW,
			(GSourceFunc) cff_gui_dialog_progress_callback,
				data, (GDestroyNotify) GuiMainLevelQuitFunc);

		GuiMainLevelStart();
	}

#if GTK_MAJOR_VERSION == 4
	/* Inform handler to return 'FALSE' */
	data->progress_bar = NULL;
#endif
	GuiWindowClose(window);

	if (data->retv == 0 && data->error == NULL)
	{
		data->error = gettext_nop("User canceled");
	}

	if (data->mutex)
	{
		GuiMutexUnlock(data->mutex);
	}

	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_dialog_progress_bar(cff_progressbar_t*) */


/*
	----------------
	 FILE-SELECTION
	----------------
*/

static GFile *cff_gui_gprop_get_directory(gint prop)
{
	gchar *path;
	GFile *file;

	path = cff_gprop_get_str(prop);

	if (path == NULL)
	{
		return NULL;
	}

	cff_tl_file_path_unix_to_local(path);
	file = cff_tl_file_get_directory(path);
	g_free(path);

	return file;

} /* static GFile *cff_gui_gprop_get_directory(gint prop) */

static gchar *cff_gui_file_get_directory(GFile *file)
{
	gchar *path;

	if (file == NULL)
	{
		return NULL;
	}

	if (cff_tl_file_is_directory(file) == FALSE)
	{
		file = g_file_get_parent(file);
		path = cff_gui_file_get_directory(file);
		g_object_unref(G_OBJECT(file));
	}

	else
	{
		path = g_file_get_parse_name(file);

		if (path != NULL)
		{
			if (path[0] == '\0')
			{
				g_free(path);
				path = NULL;
			}

		} /* if (path != NULL) */

	} /* if (cff_tl_file_is_directory(file) == FALSE .. else) */

	return path;

} /* static gchar *cff_gui_file_get_directory(GFile*) */

static gchar *cff_gui_gprop_get_filename(const gchar *fname, gint prop)
{
	gchar *path, *str;
	GFile *file;

	if (fname == NULL)
	{
		return NULL;
	}

	file = cff_gui_gprop_get_directory(prop);

	if (file == NULL)
	{
		path = NULL;
	}

	else
	{
		path = cff_gui_file_get_directory(file);
		g_object_unref(G_OBJECT(file));
	}

	if (path == NULL)
	{
		str = NULL;

		if (fname != NULL)
		{
			str = g_strdup(fname);
			cff_tl_file_path_unix_to_local(str);
		}
	}

	else
	{
		const gchar *basename;

		basename = cff_tl_file_get_basename(fname);
		str      = g_strconcat(path, G_DIR_SEPARATOR_S, basename, NULL);

		g_free(path);

	} /* if (path == NULL) .. else */

	return str;

} /* static gchar *cff_gui_gprop_get_filename(const gchar*, gint) */

static void cff_gui_gprop_set_directory(GFile *file, gint prop)
{
	gchar *path = cff_gui_file_get_directory(file);

	if (path != NULL)
	{
		if (path[0] != '\0')
		{
			cff_tl_file_path_local_to_unix(path);
			cff_gprop_set_str(prop, path);
		}

		g_free(path);

	} /* if (path != NULL) */

} /* static void cff_gui_gprop_set_directory(GFile*, gint) */

static gchar *cff_gui_file_filter_validate_string(const gchar *text, gint length, guint skip)
{
	gboolean	changed;
	gint		i, n;
	gchar		*str, *s;
	const gchar	*txt;

	n = length;

	if (n < 0)
	{
		n = strlen(text);
	}

	changed	= FALSE;

	while (n > 0 && *(text) == ' ')
	{
		n--;
		text++;
		changed = TRUE;
	}

	if (n == 0)
	{
		return NULL;
	}

	txt = text;
	i   = n;

	while (--i >= 0)
	{
		/* Skip space if string contains multiple extensions */

		guint chr = *(txt++);

		if (chr != skip)
		{
			if (g_ascii_isalnum(chr) == FALSE)
			{
				/* Not alphanumeric */
				return NULL;
			}

			if (g_ascii_isupper(chr))
			{
				/* Convert to lower */
				changed = TRUE;
			}
		}

	} /* while (--i >= 0) */

	if (changed == FALSE)
	{
		return (gchar *) text;
	}

	str = g_malloc(n + 1);
	s   = str;

	while (--n >= 0)
	{
		*(s++) = g_ascii_tolower(*(text++));
	}

	*s = '\0';

	return str;

} /* static gchar *cff_gui_file_filter_validate_string(const gchar*, gint, guint) */

static gchar *cff_gui_file_filter_extension_next(gchar *str, gchar **nextptr)
{
	guint chr;
	gchar *end, *next;

	end = strchr(str, ' ');

	if (end == NULL)
	{
		end = strchr(str, '\0');
	}

	chr  = *end;
	*end = '\0';
	next = end;

	while (chr == ' ')
	{
		chr = *(++next);
	}

	*nextptr = next;

	return end;

} /* static gchar *cff_gui_file_filter_extension_next(gchar *beg, gchar**) */

static gboolean cff_gui_file_filter_extension_compare(const gchar *str, const gchar *pattern)
{
	if (str != NULL)
	{
		const gchar *match;

		while ((match = strstr(str, pattern)) != NULL)
		{
			/* (match == pattern) => found itself */

			if (match != pattern)
			{
				guint  chr;
				size_t len;

				len = strlen(pattern);
				chr = match[len];

				if (chr == ' ' || chr == '\0')
				{
					/* Last characters are equal */

					chr = ' ';

					if (match > str)
					{
						chr = *(--match);
						match++;
					}

					if (chr == ' ')
					{
						/* First characters are equal */
						return TRUE;
					}
				}

			} /* if (match != pattern) */

			str = &match[1];

		} /* while ((match = strstr(str, pattern)) != NULL) */

	} /* if (str != NULL) */

	return FALSE;

} /* static gboolean cff_gui_file_filter_extension_compare(const gchar*, const gchar*) */

static gchar *cff_gui_file_filter_extension_remove(gchar *beg, gchar *end, gchar *start)
{
	if (*end == '\0')
	{
		/* Last extension in the string */

		while (beg > start)
		{
			/* Not the first entry */

			if (*(--beg) != ' ')
			{
				beg++;
				break;
			}
		}

		/* Remove last extension */
		*beg = '\0';
	}

	else
	{
		/* Move following extension(s) down */
		beg = memmove(beg, end, strlen(end) + 1);
	}

	return beg;

} /* static gchar *cff_gui_file_filter_extension_remove(gchar*, gchar*, gchar*) */

static void cff_gui_file_filter_validate_prop(gint prop)
{
	gchar *ext2 = cff_gprop_get_str(prop);

	if (ext2 != NULL)
	{
		gboolean changed;
		gchar    *str;

		changed	= FALSE;
		str	= cff_gui_file_filter_validate_string(ext2, -1, ' ');

		if (str != ext2)
		{
			/* 'ext2' was changed or is invalid */
			g_free(ext2);
			ext2	= str;
			changed	= TRUE;
		}

		if (ext2 != NULL)
		{
			gint		n;
			const gchar	*ext;

			if (prop == kPropStrFileVideoFilter)
			{
				ext = cff_video_ext;
			}

			else /* kPropStrFileImageFilter */
			{
				ext = cff_image_ext;
			}

			n = 0;

			do
			{
				gchar *end, *next;

				end = cff_gui_file_filter_extension_next(str, &next);

				if ((next - end) > 1)
				{
					/* More than one space */

					if (*next != '\0')
					{
						/* No following extension(s) */
						next--;
					}

					/* Move following spaces down */
					cff_gui_file_filter_extension_remove(end, next, ext2);
					changed = TRUE;
				}

				else if ((end - str) >= kFileFilterCharsMax ||
					cff_gui_file_filter_extension_compare(ext, str) ||
						cff_gui_file_filter_extension_compare(ext2, str))
				{
					/* Move following extensions down */
					str = cff_gui_file_filter_extension_remove(str, next, ext2);
					changed = TRUE;
				}

				else
				{
					if (++n >= kFileFilterEntriesMax)
					{
						/* String was terminated (end) */
						changed = TRUE;
						break;
					}

					if (*next != '\0')
					{
						/* Not the last extension */
						*end = ' ';
					}

					str = next;
				}
			}
			while (*str != '\0');

		} /* if (ext2 != NULL) */

		if (changed)
		{
			cff_gprop_set_str(prop, ext2);
		}

		g_free(ext2);

	} /* if (ext2 != NULL) */

} /* static void cff_gui_file_filter_validate_prop(gint) */

static const gchar *cff_gui_file_filter_add(GtkFileChooser *chooser, const gchar *ext, GtkFileFilter **filter)
{
	static const gchar all_files[] = gettext_nop("All files");

	guint chr;
	gchar *str, *name;
	gchar filter_glob[256], filter_name[256];

	do
	{
		chr = *(ext++);
	}
	while (chr == ' ');

	if (chr == '\0')
	{
		ext--;
		return ext;
	}

	str = filter_glob;

	if (chr == '*')
	{
		*(str++) = chr;
		chr = *(ext++);

		name = cff_charconv_utf8_gettext(all_files);
		strcpy(filter_name, name);
		CFF_CHARCONV_UTF8_FREE(name, all_files);
	}

	else
	{
		name = filter_name;

		*(str++) = '*';
		*(str++) = '.';

		do
		{
			/* Extension in lower and uppercase */
			*(str++) = '[';
			*(str++) = chr;

			/* Name in uppercase */
			chr = g_ascii_toupper(chr);
			*(str++)  = chr;
			*(name++) = chr;

			*(str++) = ']';

			chr = *(ext++);

			if (chr == '\0')
			{
				/* Done */
				ext--;
				break;
			}
		}
		while (chr != ' ');

		*name = '\0';

	} /* if (chr == '*') .. else */

	*str = '\0';

	*filter = gtk_file_filter_new();
	gtk_file_filter_set_name(*filter, filter_name);
	gtk_file_filter_add_pattern(*filter, filter_glob);
	gtk_file_chooser_add_filter(chooser, *filter);

	return ext;

} /* static const gchar *cff_gui_file_filter_add(GtkFileChooser*, const gchar*, GtkFileFilter**) */

static gint cff_gui_file_filter_init(GtkFileChooser *chooser, gint prop, GtkFileFilter *filterbuf[kFileFilterEntriesMax])
{
	static const gchar cutterff_ext[] = "cff";

	gint		n, select;
	guint		num;
	gchar		*ext2;
	const gchar	*ext;

	if (prop == kPropStrEditCutListPath)
	{
		ext = cutterff_ext;
		select = -1;
	}

	else if (prop == kPropStrFileImagePath)
	{
		ext	= cff_image_ext;
		select	= 1;
		prop	= kPropStrFileImageFilter;
	}

	else
	{
		if (cff_gprop_get_int(kPropIntFileFilterIndex, &select) == FALSE)
		{
			select = -1;
		}

		ext  = cff_video_ext;
		prop = kPropStrFileVideoFilter;

	} /* if (prop == kPropStrEditCutListPath) .. else if .. else */


	n = -1;

	do
	{
		GtkFileFilter *filter;

		ext = cff_gui_file_filter_add(chooser, ext, &filter);

		if (++n == select)
		{
			gtk_file_chooser_set_filter(chooser, filter);
		}
	}
	while (*ext != '\0');


	num = 0;

	if (prop != kPropStrEditCutListPath &&
			(ext2 = cff_gprop_get_str(prop)) != NULL)
	{
		ext = ext2;

		do
		{
			ext = cff_gui_file_filter_add(chooser, ext, &filterbuf[num]);

			if (++n == select)
			{
				gtk_file_chooser_set_filter(chooser, filterbuf[num]);
			}
		}
		while (++num < kFileFilterEntriesMax && *ext != '\0');

		g_free(ext2);

	} /* if (prop != kPropStrEditCutListPath &&
				(ext2 = cff_gprop_get_str(prop)) != NULL) */
	return num;

} /* static gint cff_gui_file_filter_init(GtkFileChooser*, gint, GtkFileFilter*[kFileFilterEntriesMax]) */

static void cff_gui_file_selection_add_filter_entry_cb(GtkEditable *editable, const gchar *text, gint length, gint *pos, gint *sigid)
{
	gchar *str;

	GuiSignalHandlerBlock(editable, *sigid);

	str = cff_gui_file_filter_validate_string(text, length, '\0');

	if (str != NULL)
	{
		gtk_editable_insert_text(editable, str, length, pos);

		if (str != text)
		{
			g_free(str);
		}
	}

	GuiSignalEmitStopByName(editable, "insert-text");
	GuiSignalHandlerUnblock(editable, *sigid);

} /* static void cff_gui_file_selection_add_filter_entry_cb(GtkEditable *editable, const gchar*, gint, gint*, gint*) */

static void cff_gui_file_selection_add_filter_cb(GtkButton *b, cff_filefilter_t *filter)
{
	guint		sigid;
	GtkResponseType	response;
	gchar		*ext, *ext2;
	const gchar	*space;
	GtkWidget	*dialog, *entry, *frame, *label, *vbox;
	GtkEditable	*editable;
#if GTK_MAJOR_VERSION == 2
	GtkWidget	*widget;
#endif

	if (filter->num >= kFileFilterEntriesMax)
	{
		return;
	}

	vbox  = GuiVBoxNew(FALSE, 2);
	entry = gtk_entry_new();
	gtk_entry_set_max_length(GTK_ENTRY(entry), kFileFilterCharsMax);

#if GTK_MAJOR_VERSION == 2
	widget = GuiHBoxNew(FALSE, 2);
	GuiHBoxPackStart(widget, entry, TRUE,  TRUE, 8);
	GuiVBoxPackStart(vbox, widget, FALSE, FALSE, 4);
#else
	GuiMiscSetPadding(entry, 8);
	GuiVBoxPackStart(vbox, entry, FALSE, FALSE, 4);
#endif

	editable = GuiEntryGetEditable(entry);
	sigid    = GuiSignalConnect(editable, "insert-text",
			cff_gui_file_selection_add_filter_entry_cb, &sigid);

	ext2  = cff_gprop_get_str(filter->prop);
	space = NULL;

	if (ext2 != NULL)
	{
		space = " ";
	}

	/* Skip "* " */
	ext = g_strconcat(&filter->ext[2], space, ext2, NULL);

	label = gtk_label_new(ext);
	GuiLabelSetLineWrap(label, TRUE);
	GuiMiscSetPadding(label, 8);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);

	g_free(ext);

	frame = cff_guitl_frame_new_add_child_gettext(gettext_nop("Filters"),
									8, label);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 2);

	dialog = cff_guitl_dialog_new_cancel_ok(gettext_nop("Add filter"),
								filter->dialog);
#if GTK_MAJOR_VERSION == 4
	gtk_widget_set_size_request(dialog, 300, -1);
	gtk_widget_show(dialog);
#endif
	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, 8, vbox);
	cff_guitl_dialog_set_child(dialog, frame);
	gtk_widget_grab_focus(entry);


	while ((response = GuiDialogRun(dialog)) == GTK_RESPONSE_ACCEPT)
	{
		const gchar *str = GuiEntryGetText(GTK_ENTRY(entry));

		if (str != NULL && str[0] != '\0')
		{
			gint		slen;
			gchar		*s;
			GtkFileFilter	**f;

			if (cff_gui_file_filter_extension_compare(filter->ext, str) ||
					cff_gui_file_filter_extension_compare(ext2, str))
			{
				/* Filter already exists */
				cff_guitl_beep();
				continue;
			}

			f = &filter->buf[filter->num];

			cff_gui_file_filter_add(filter->chooser, str, f);
			gtk_file_chooser_set_filter(filter->chooser, *f);
			filter->num++;

			slen = (ext2 == NULL) ? 0 : strlen(ext2);
			/* Plus '\0' or ' ' */
			ext2 = g_realloc(ext2, strlen(str) + slen + 1);
			s    = &ext2[slen];

			if (slen > 0)
			{
				*(s++) = ' ';
			}

			strcpy(s, str);
			cff_gprop_set_str(filter->prop, ext2);

			gtk_widget_set_sensitive(filter->btn_remove, TRUE);

			if (filter->num >= kFileFilterEntriesMax)
			{
				gtk_widget_set_sensitive(filter->btn_add, FALSE);
			}

			break;

		} /* if (str != NULL && str[0] != '\0') */

	} /* while ((response = GuiDialogRun(dialog)) == GTK_RESPONSE_ACCEPT) */

	g_free(ext2);

	GuiDialogClose(dialog, response);

} /* static void cff_gui_file_selection_add_filter_cb(GtkButton*, cff_filefilter_t*) */

static void cff_gui_file_selection_remove_filter_cb(GtkButton *b, cff_filefilter_t *filter)
{
	guint			num;
	GtkResponseType		response;
	gchar			*ext2, *str;
	GtkWidget		*dialog, *frame, *swin;
	GtkListStore		*store;
	GtkTreeView		*view;
	GtkTreeViewColumn	*column;
	GtkTreeSelection	*selection;
	GtkCellRenderer		*renderer;
	GtkTreeIter		iter;

	if (filter->num == 0)
	{
		return;
	}

	ext2 = cff_gprop_get_str(filter->prop);

	if (ext2 == NULL)
	{
		return;
	}

	store = gtk_list_store_new(4, G_TYPE_STRING, G_TYPE_POINTER,
						G_TYPE_POINTER, G_TYPE_INT);
	view = (GtkTreeView *)
		gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));
	gtk_tree_view_set_headers_visible(view, FALSE);

	g_object_unref(G_OBJECT(store));

	selection = gtk_tree_view_get_selection(view);
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);
	GuiWidgetSetCanFocus(GTK_WIDGET(view), FALSE);

	column   = gtk_tree_view_column_new();
	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
	gtk_tree_view_column_set_attributes(column, renderer, "text", 0, NULL);

	gtk_tree_view_append_column(view, column);


	num = 0;
	str = ext2;

	do
	{
		gchar *end, *next;

		end = cff_gui_file_filter_extension_next(str, &next);

		gtk_list_store_append(store, &iter);
		gtk_list_store_set(store, &iter,
				0, str, 1, str, 2, next, 3, num, -1);

		if (end != next)
		{
			*end = ' ';
		}

		if (num == 0)
		{
			gtk_tree_selection_select_iter(selection, &iter);
		}

		str = next;
	}
	while (*str != '\0' && ++num < filter->num);


	dialog = cff_guitl_dialog_new_cancel_ok(gettext_nop("Remove filter"),
								filter->dialog);
#if GTK_MAJOR_VERSION == 4
	gtk_widget_show(dialog);
#endif
	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);

	swin  = cff_guitl_scrolled_window_new(-1, 150, GTK_WIDGET(view));
	frame = cff_guitl_frame_new_add_child_gettext(NULL, 8, swin);
	cff_guitl_dialog_set_child(dialog, frame);


	while ((response = GuiDialogRun(dialog)) == GTK_RESPONSE_ACCEPT)
	{
		guint		n;
		gchar		*s, *e;
		GtkTreeModel	*model;

		if (gtk_tree_selection_get_selected(selection,
						&model, &iter) == FALSE)
		{
			/* Nothing selected */
			cff_guitl_beep();
			continue;
		}

		gtk_tree_model_get(model, &iter, 1, &s, 2, &e, 3, &n, -1);

		cff_gui_file_filter_extension_remove(s, e, ext2);
		gtk_file_chooser_remove_filter(filter->chooser, filter->buf[n]);

		num = filter->num - (n + 1);

		if (num != 0)
		{
			/* Move following entries one position down */
			memmove(&filter->buf[n], &filter->buf[n + 1],
						num * sizeof(GtkFileFilter*));
		}

		filter->num--;

		if (filter->num == 0)
		{
			gtk_widget_set_sensitive(filter->btn_remove, FALSE);
		}

		if (filter->num < kFileFilterEntriesMax)
		{
			gtk_widget_set_sensitive(filter->btn_add, TRUE);
		}

		cff_gprop_set_str(filter->prop, ext2);

		break;

	} /* while ((response = GuiDialogRun(dialog)) == GTK_RESPONSE_ACCEPT) */

	g_free(ext2);

	GuiDialogClose(dialog, response);

} /* static void cff_gui_file_selection_remove_filter_cb(GtkButton*, cff_filefilter_t*) */


#if GTK_MAJOR_VERSION == 4

static void cff_gui_file_selection_popover(GtkWidget *button, gpointer data)
{
	/* Display help text */

	static const gchar bug_str[] = gettext_nop(
		"Typing a character in the name entry opens the search entry"
		" and put the character into it.\n"
		"If this bug is still present, then try the following:\n"
		"One character: Press the key, click into the position"
		" in the name entry and press the key again.\n"
		"More characters: Type them in the search entry, select"
		" them, press 'CNTRL-C', click into the position"
		" in the name entry and press 'CNTRL-V'"
		); /* gettext_nop */

	GtkWidget *label, *popover;

	label = cff_guitl_label_new_gettext(bug_str);
	gtk_label_set_wrap(GTK_LABEL(label), TRUE);

	popover = gtk_popover_new();

	/* Connect popover */
	gtk_popover_set_child(GTK_POPOVER(popover), label);
	gtk_widget_set_parent(popover, button);

	/* Show popover */
	gtk_popover_popup(GTK_POPOVER(popover));

	/* Destroy popover */
	GuiSignalConnect(popover, "hide", gtk_widget_unparent, NULL);

} /* static void cff_gui_file_selection_popover(GtkWidget*, gpointer) */

#endif /* GTK_MAJOR_VERSION == 4 */


static void cff_gui_file_selection_cancel(GtkWidget *button, GtkResponseType *response)
{
	/* Cancel button */
	*response = GTK_RESPONSE_CANCEL;
	GuiMainLevelQuit();

} /* static void cff_gui_file_selection_cancel(GtkWidget*, GtkResponseType*) */

static void cff_gui_file_selection_accept(GtkWidget *button, GtkResponseType *response)
{
	/* Open, save or load */
	*response = GTK_RESPONSE_ACCEPT;
	GuiMainLevelQuit();

} /* static void cff_gui_file_selection_accept(GtkWidget*, GtkResponseType*) */

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
static gboolean cff_gui_file_selection_delevnt(GtkWidget *w, GdkEvent *e, GtkResponseType *response)
#else
static gboolean cff_gui_file_selection_delevnt(GtkWidget *w, GtkResponseType *response)
#endif
{
	/* The windows close button */
	*response = GTK_RESPONSE_DELETE_EVENT;
	GuiMainLevelQuit();
	return FALSE;

} /* static gboolean cff_gui_file_selection_delevnt(GtkWidget*, GdkEvent*, GtkResponseType*) */

static gboolean cff_gui_file_selection_key_pressed(GtkWidget *button)
{
	/* Return or cancel button */
	GuiWidgetUnsetStateSelected(button);
	GuiButtonClicked(button);
	return Gui_SOURCE_REMOVE;

} /* static gboolean cff_gui_file_selection_key_pressed(GtkWidget*) */

#ifdef HAVE_GMENU
static gboolean cff_gui_file_selection_escape_return_key(GtkEventControllerKey *c,
			guint keyval, guint keycode, GdkModifierType state,
							gpointer data)
#else
static gboolean cff_gui_file_selection_escape_return_key(GtkWidget *w, GdkEvent *e,
							gpointer data)
#endif
{
	/* Test whether the Escape- or Return-key was pressed */

#ifndef HAVE_GMENU
	guint		keyval;
	GdkModifierType	state;
	GdkEventKey	*kevent;

	kevent	= (GdkEventKey *) e;
	keyval	= kevent->keyval;
	state	= kevent->state;
#endif
	if ((state & (GDK_CONTROL_MASK | GDK_SHIFT_MASK)) == 0)
	{
		GtkWidget *button, **btn_done;

		btn_done = data;
		
		if (keyval == Gui_KEY_Escape)
		{
			button = btn_done[0];
		}

		else if (keyval == Gui_KEY_Return)
		{
			button = btn_done[1];
		}

		else
		{
			return FALSE;
		}

		GuiWidgetSetStateSelected(button);

		/* 175ms */
		g_timeout_add(175,
			(GSourceFunc) cff_gui_file_selection_key_pressed,
			button);

		return TRUE;

	} /* if ((state & (GDK_CONTROL_MASK | GDK_SHIFT_MASK)) == 0) */

	return FALSE;

} /* static gboolean cff_gui_file_selection_escape_return_key(GtkWidget*, GdkEvent*, gpointer) */

static GFile *cff_gui_file_selection_new_gettext(const gchar *title, const gchar *fname_dst, const gchar *fname_src, GtkWidget **mbar, gint prop)
{
	gboolean			path_prop;
	guint				sigid;
	GtkFileChooserAction		action;
	volatile GtkResponseType	response;
	const gchar			*btn_ok;
	GFile				*file;
	GtkWidget			*chooser, *hwoption, *frame, *bbox, *hbox, *vbox, *parent, *window, *done_btn[2];
	cff_filefilter_t		filter_s;
	cff_hwacclist_t			*hwacclist;
#ifdef HAVE_GMENU
	GtkEventController		*controller;
#endif

	if (fname_dst == NULL)
	{
		if (prop == kPropStrFileOpenPath)
		{
			btn_ok = gettext_nop("Open");
		}

		else
		{
			btn_ok = gettext_nop("Load");
		}

		action = GTK_FILE_CHOOSER_ACTION_OPEN;
	}

	else
	{
		if (prop == kPropStrFileSavePath ||
				prop == kPropStrFileImagePath)
		{
			btn_ok = gettext_nop("OK");
		}

		else
		{
			btn_ok = gettext_nop("Save");
		}

		action = GTK_FILE_CHOOSER_ACTION_SAVE;

	} /* if (fname_dst == NULL) .. else */


	vbox = GuiVBoxNew(FALSE, 0);

	/* File chooser window */

	parent = cff_gui_window_get_topwin();
	window = cff_guitl_window_new_gettext(title, vbox, 4);
	filter_s.dialog = window;
	cff_guitl_window_setup(window, kPropIntFileChooserWinWidth);
	cff_guitl_dialog_set_modal(window, parent, mbar);

	chooser = gtk_file_chooser_widget_new(action);
	GuiContainerSetBorderWidth(chooser, 4);
	filter_s.chooser = GTK_FILE_CHOOSER(chooser);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, 4, chooser);
	GuiVBoxPackStart(vbox, frame, TRUE, TRUE, 4);

	filter_s.num = cff_gui_file_filter_init(filter_s.chooser,
						prop, filter_s.buf);

	hbox = GuiHBoxNew(FALSE, 0);
	GuiVBoxPackStart(vbox, hbox, FALSE, FALSE, 0);

	if (prop != kPropStrEditCutListPath)
	{
		/* Extra widget */

		if (prop == kPropStrFileImagePath)
		{
			filter_s.ext  = cff_image_ext;
			filter_s.prop = kPropStrFileImageFilter;
		}

		else
		{
			filter_s.ext  = cff_video_ext;
			filter_s.prop = kPropStrFileVideoFilter;
		}

		bbox = GuiHButtonBoxNew();
		GuiContainerSetBorderWidth(bbox, 2);

#if GTK_MAJOR_VERSION == 2
		gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), GTK_BUTTONBOX_START);
#endif
		frame = cff_guitl_frame_new_add_child_gettext(gettext_nop("Filter"), 0, bbox);
		GuiHBoxPackStart(hbox, frame, FALSE, FALSE, 4);

		filter_s.btn_add = cff_guitl_button_box_add_button(bbox,
				gettext_nop("Add ..."),
				(GCallback) cff_gui_file_selection_add_filter_cb,
				&filter_s);

		cff_guitl_widget_set_tooltip_gettext(filter_s.btn_add,
				gettext_nop("Add a file type to the list"));

		if (filter_s.num >= kFileFilterEntriesMax)
		{
			gtk_widget_set_sensitive(filter_s.btn_add, FALSE);
		}

		filter_s.btn_remove = cff_guitl_button_box_add_button(bbox,
				gettext_nop("Remove ..."),
				(GCallback) cff_gui_file_selection_remove_filter_cb,
				&filter_s);

		cff_guitl_widget_set_tooltip_gettext(filter_s.btn_remove,
				gettext_nop("Remove a file type from the list"));

		if (filter_s.num == 0)
		{
			gtk_widget_set_sensitive(filter_s.btn_remove, FALSE);
		}

	} /* if (prop != kPropStrEditCutListPath) */


	hwacclist = NULL;
	hwoption  = NULL;

	if (prop == kPropStrFileOpenPath)
	{
		gchar		*str;
		const gchar	*no_str;
		cff_hwacclist_t	 *hwaccinfo;

#if GTK_MAJOR_VERSION == 2
		GtkWidget	*menu;
#endif
		hwoption = GuiOptionMenuTextNew();
		GuiContainerSetBorderWidth(hwoption, 4);

		vbox = GuiVBoxNew(TRUE, 4);

		GuiVBoxPackStart(vbox, hwoption, FALSE, FALSE, 4);
		frame = cff_guitl_frame_new_add_child_gettext(
				gettext_nop("HW-acceleration"), 0, vbox);
		GuiHBoxPackStart(hbox, frame, FALSE, FALSE, 12);

#if GTK_MAJOR_VERSION == 2
		menu = gtk_menu_new();
		gtk_widget_realize(menu);
		gtk_option_menu_set_menu(GTK_OPTION_MENU(hwoption), menu);
#endif

		no_str = gettext_nop("No");
		str = cff_charconv_utf8_gettext(no_str);
		GuiOptionMenuTextAppend(hwoption, menu, 0, str);
		CFF_CHARCONV_UTF8_FREE(str, no_str);

		hwacclist = cff_ffmpeg_get_hwaccel_names();

		if ((hwaccinfo = hwacclist) == NULL)
		{
			GuiOptionMenuSetActive(hwoption, 0);
			gtk_widget_set_sensitive(hwoption, FALSE);
		}

		else
		{
			gint  i, select;
			gchar *prv_str, *buf[32];

			i	= 0;
			select	= 0;
			prv_str	= cff_gprop_get_str(kPropStrFileHWAccel);

			while (hwaccinfo->name != NULL)
			{
				if (hwaccinfo->codecs == NULL)
				{
					continue;
				}

				if (i > 30)
				{
					break;
				}

				buf[i] = g_strconcat(hwaccinfo->name, " ",
							hwaccinfo->codecs, NULL);
				/* malloc'ed */
				free(hwaccinfo->codecs);

				i++;

				if (select == 0 && prv_str != NULL &&
					strcmp(prv_str, hwaccinfo->name) == 0)
				{
					select = i;
				}

				GuiOptionMenuTextAppend(hwoption, menu, i, hwaccinfo->name);
				hwaccinfo++;

			} /* while ((hwaccinfo->name) != NULL) */

			g_free(prv_str);

			GuiOptionMenuSetActive(hwoption, select);

			/* Tooltip shows available hardware codecs */
			buf[i]	= NULL;
			str	= g_strjoinv("\n", buf);

			while (--i >= 0)
			{
				g_free(buf[i]);
			}

			gtk_widget_set_tooltip_text(frame, str);
			g_free(str);

		} /* if ((hwaccinfo = hwacclist) == NULL) .. else */

	} /* if (prop == kPropStrFileOpenPath) */


#if GTK_MAJOR_VERSION == 4

	if (action == GTK_FILE_CHOOSER_ACTION_SAVE)
	{
		/* Display help text */

		GtkWidget *button;

		button = gtk_button_new_from_icon_name("dialog-question");
		gtk_widget_set_valign(button, GTK_ALIGN_CENTER);
		GuiHBoxPackStart(hbox, button, FALSE, FALSE, 0);

		GuiSignalConnect(button, "clicked",
				cff_gui_file_selection_popover, NULL);

	} /* if (action == GTK_FILE_CHOOSER_ACTION_SAVE) */

#endif /* GTK_MAJOR_VERSION == 4 */


	/* Padding */
	bbox = GuiVBoxNew(TRUE, 0);
	GuiHBoxPackStart(hbox, bbox, TRUE, TRUE, 0);
	GuiVBoxPackStart(bbox, GuiHBoxNew(FALSE, 0), TRUE, TRUE, 0);

	/* Button box*/
	bbox = cff_guitl_hbutton_box_new(bbox, 2);

	done_btn[0] = cff_guitl_button_box_add_button(bbox,
			gettext_nop("Cancel"),
			(GCallback) cff_gui_file_selection_cancel,
			(gpointer) &response);

	done_btn[1] = cff_guitl_button_box_add_button(bbox,
			btn_ok,
			(GCallback) cff_gui_file_selection_accept,
			(gpointer) &response);


	/* Set the directory */

	file = NULL;

	if (prop == kPropStrFileSavePath || prop == kPropStrFileImagePath)
	{
		/* Get path from filename (may changed in the entry) */
		file      = cff_tl_file_get_directory(fname_dst);
		path_prop = FALSE;
	}

	else
	{
		path_prop = TRUE;
	}

	if (path_prop || file == NULL)
	{
		/* Get path stored in gprop */
		file = cff_gui_gprop_get_directory(prop);

		if (file == NULL && path_prop)
		{
			if (fname_dst == NULL)
			{
				if (prop == kPropStrEditCutListPath)
				{
					/* Not stored yet, take the video directory */
					file = cff_gui_gprop_get_directory(kPropStrFileOpenPath);
				}
			}

			else
			{
				/* Get path from filename */
				file = cff_tl_file_get_directory(fname_dst);
			}

		} /* if (file == NULL && path_prop) */

	} /* if (path_prop || file == NULL) */

#if GTK_MAJOR_VERSION == 4
	if (action == GTK_FILE_CHOOSER_ACTION_SAVE && file != NULL)
	{
		/* GTK-4 needs a filename (should not exist) */
		file = cff_tl_file_append_fname(file, " x ");
	}
#endif

	if (file != NULL)
	{
		GuiFileChooserSetCurrentFolder(filter_s.chooser, file);
		g_object_unref(G_OBJECT(file));
		file = NULL;

	} /* if (file != NULL) */


	/* Set the file name */

	if ((prop == kPropStrFileSavePath ||
		prop == kPropStrFileImagePath) &&
			(fname_dst == NULL || fname_dst[0] == '\0'))
	{
		fname_dst = "noname";
	}

	if (fname_dst != NULL)
	{
		const gchar *basename;

		basename = cff_tl_file_get_basename(fname_dst);

		/* The name passed in here is a UTF-8 string */
		gtk_file_chooser_set_current_name(filter_s.chooser, basename);

	} /* if (fname_dst != NULL) */


	/* Set the signals */

#ifdef HAVE_GMENU

#if GTK_MAJOR_VERSION == 3
	controller = gtk_event_controller_key_new(window);
#else
	controller = gtk_event_controller_key_new();
	gtk_widget_add_controller(window, controller);
#endif
	gtk_event_controller_set_propagation_phase(controller,
							GTK_PHASE_CAPTURE);
	GuiSignalConnect(controller, "key-pressed",
		cff_gui_file_selection_escape_return_key, (gpointer) done_btn);
#else
	GuiSignalConnect(window, "key-press-event",
		cff_gui_file_selection_escape_return_key, (gpointer) done_btn);

#endif /* #ifdef HAVE_GMENU .. #else */

	sigid = GuiSignalConnect(window, gDeleteEventStr,
			cff_gui_file_selection_delevnt, (gpointer) &response);

	GuiWidgetShowAll(window);


	while (GuiMainLevelStart(), response == GTK_RESPONSE_ACCEPT)
	{

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
		GSList		*list;
#else
		GListModel	*list;
#endif
		file = gtk_file_chooser_get_file(filter_s.chooser);

		if (file == NULL)
		{
			/* Nothing selected */

#if GTK_MAJOR_VERSION == 4

			gchar *fname =
				gtk_file_chooser_get_current_name(filter_s.chooser);

			file = GuiFileChooserGetCurrentFolder(filter_s.chooser);
			file = cff_tl_file_append_fname(file, fname);
			g_free(fname);

			if (file == NULL)

#endif /* GTK_MAJOR_VERSION == 4 */

			{
				cff_guitl_beep();
				continue;
			}
		}

		else if (cff_tl_file_is_directory(file))
		{
			GuiFileChooserSetCurrentFolder(filter_s.chooser, file);
			g_object_unref(G_OBJECT(file));
			file = NULL;
			continue;

		} /* if (file == NULL) .. else */

		if (prop == kPropStrFileSavePath ||
					prop == kPropStrFileImagePath)
		{
			/* Warning icon if equal */
			if (cff_gui_file_is_equal(file, fname_src, FALSE))
			{
				g_object_unref(G_OBJECT(file));
				file = NULL;
				continue;
			}
		}

		if (action == GTK_FILE_CHOOSER_ACTION_SAVE &&
					cff_gui_overwrite(file) == FALSE)
		{
			g_object_unref(G_OBJECT(file));
			file = NULL;
			continue;
		}

		if (prop == kPropStrFileOpenPath ||
					prop == kPropStrFileSavePath)
		{
			gint		index;
			GtkFileFilter	*filter;

			filter = gtk_file_chooser_get_filter(filter_s.chooser);
			list   = GuiFileChooserListFilters(filter_s.chooser);

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

			index  = g_slist_index(list, filter);
			g_slist_free(list);
#else
			index = 0;

			for (;;)
			{
				gpointer item;

				item = g_list_model_get_item(list, index);

				if (item == (gpointer) filter)
				{
					break;
				}

				if (item == NULL)
				{
					break;
				}

				index++;

			} /* for (;;) */

			g_object_unref(G_OBJECT(list));

#endif /* #if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 .. #else */

			cff_gprop_set_int(kPropIntFileFilterIndex, index);

		} /* if (prop == kPropStrFileOpenPath ||
					prop == kPropStrFileSavePath) */

		/* The save dialogs store the path itself */

		if (prop == kPropStrFileOpenPath ||
					prop == kPropStrEditCutListPath)
		{
			cff_gui_gprop_set_directory(file, prop);
		}

		if (hwoption != NULL)
		{
			/* prop == kPropStrFileOpenPath */

			gint		selected;
			const gchar	*hwname;

			selected = GuiOptionMenuGetActive(hwoption);

			hwname = (hwacclist == NULL || selected == 0) ?
					NULL : hwacclist[selected - 1].name;

			cff_gprop_set_str(kPropStrFileHWAccel, hwname);

		} /* if (hwoption != NULL) */

		break;

	} /* while (GuiMainLevelStart(), response == GTK_RESPONSE_ACCEPT) */


	if (response != GTK_RESPONSE_DELETE_EVENT)
	{
		GuiSignalHandlerDisconnect(window, sigid);
	}

	if (hwacclist != NULL)
	{
		/* malloc'ed */
		free(hwacclist);
	}

	GuiDialogClose(window, response);

	return file;

} /* static GFile *cff_gui_file_selection_new_gettext(const gchar*, const gchar*, const gchar*, GtkWidget**, gint) */


/*
	-------------
	 SAVE COMMON
	-------------
*/

static void cff_gui_save_okay_button_set_sensitive(cff_save_common_t *vscommon, guint bits)
{
	if (bits != vscommon->okay_bits)
	{
		gboolean active;

		vscommon->okay_bits = bits;

		active = TRUE;

		if (vscommon->okay_bits != 0)
		{
			/* Invalid entry, or no stream in the output list */
			active = FALSE;
		}

		gtk_dialog_set_response_sensitive(GTK_DIALOG(vscommon->dialog),
						GTK_RESPONSE_ACCEPT, active);

	} /* if (bits != vscommon->okay_bits) */

} /* static void cff_gui_save_okay_button_set_sensitive(cff_save_common_t*, guint) */

static void cff_gui_save_entry_changed(cff_save_common_t *vscommon)
{
	/* Disable the okay-button if the entry is empty or invalid */

	const gchar	*str;
	guint		bits;

	str  = GuiEntryGetText(vscommon->entry);
	bits = vscommon->okay_bits;

	if (str == NULL)
	{
		/* Empty */
		bits |= 0x01;
	}

	else
	{
		guint    chr, last_chr;
		gboolean first, ext;

		first    = TRUE;
		ext      = FALSE;
		last_chr = '\0';

		while ((chr = *(str++)) != '\0')
		{
			if (isspace(chr))
			{
				if (first)
				{
					break;
				}

				if (last_chr == '.')
				{
					ext = TRUE;
				}
			}

			last_chr = chr;
			first    = CFF_IS_DIR_SEPARATOR(chr);

			if (first)
			{
				ext = FALSE;
			}

		} /* while ((chr = *(str++)) != '\0') */

		if (chr != '\0' || first == TRUE ||
					ext == TRUE || isspace(last_chr))
		{
			/* Empty, or contains leading spaces */
			bits |= 0x01;
		}

		else
		{
			/* Entry is okay */
			bits &= ~0x01;
		}

	} /* if (str == NULL .. else */

	cff_gui_save_okay_button_set_sensitive(vscommon, bits);

} /* static void cff_gui_save_entry_changed(cff_save_common_t*) */

static void cff_gui_save_entry_delete_callback(GtkEntryBuffer *buffer,
		guint position, guint n_chars, cff_save_common_t *vscommon)
{
	cff_gui_save_entry_changed(vscommon);

} /* static void cff_gui_save_entry_delete_callback(GtkEntryBuffer*,
					guint, guint, cff_save_common_t*) */

static void cff_gui_save_entry_insert_callback(GtkEntryBuffer *buffer,
		guint position, gchar *chars, guint n_chars, cff_save_common_t *vscommon)
{
	cff_gui_save_entry_changed(vscommon);

} /* static void cff_gui_save_entry_insert_callback(GtkEntryBuffer *buffer,
					guint, gchar*, guint, cff_save_common_t*) */

static void cff_gui_save_selection_callback(GtkWidget *w, cff_save_common_t *vscommon)
{
	gchar		*fname, *fname_dst;
	const gchar	*ename;
	GFile		*file;

	/* Free previous 'fname_dst' */
	g_free(vscommon->filechooser_utf8str);
	vscommon->filechooser_utf8str = NULL;

	ename = GuiEntryGetText(vscommon->entry);

	if (ename == NULL || ename[0] == '\0')
	{
		fname     = NULL;
		fname_dst = "";
	}

	else
	{
		fname = g_strdup(ename);
		cff_tl_file_path_unix_to_local(fname);
		fname_dst = fname;
	}

	file = cff_gui_file_selection_new_gettext(gettext_nop("Select ..."),
		fname_dst, (vscommon->video)->fname.utf8, NULL, vscommon->path_prop);

	g_free(fname);

	if (file != NULL)
	{
		fname_dst = g_file_get_parse_name(file);
		g_object_unref(G_OBJECT(file));

		if (fname_dst != NULL)
		{
			if (fname_dst[0] == '\0')
			{
				g_free(fname_dst);
			}

			else if (vscommon->filechooser_utf8str == NULL ||
					strcmp(vscommon->filechooser_utf8str, fname_dst) != 0)
			{
				/* Do not free 'fname_dst' */
				vscommon->filechooser_utf8str = fname_dst;
				GuiEntrySetText(vscommon->entry, fname_dst);

			} /* if (fname_dst[0] == '\0') .. else */

		} /* if (fname_dst != NULL) */

	} /* if (file != NULL) */

} /* static void cff_gui_save_selection_callback(GtkWidget*, cff_save_common_t*) */

static void cff_gui_save_optionmenu_callback(GuiOptionMenu *option, cff_save_common_t *vscommon)
{
	gint		i, pos, slen;
	gchar		*extension;
	const gchar	*str, *end;
	GtkEditable	*editable;

	i = GuiOptionMenuGetActive(option);

	extension = NULL;

	if (i > 0)
	{
		const gchar *format_name = NULL;

		if (vscommon->path_prop == kPropStrFileSavePath)
		{
			if (vscommon->format.u.list != NULL)
			{
				/* Array without "Auto" */
				format_name =
					vscommon->format.u.list[i - 1].name;
			}
		}

		else /* kPropStrFileImagePath */
		{
			if (vscommon->format.u.names != NULL)
			{
				/* Array without "Auto" */
				format_name = vscommon->format.u.names[i - 1];
			}
		}

		if (format_name != NULL)
		{
			extension = cff_ffmpeg_get_outformat_extension(format_name, &slen);
		}

	} /* if (i > 0) */

	if (extension == NULL)
	{
		if ((extension = vscommon->format.extension_orig) == NULL)
		{
			return;
		}

		slen = -1;
	}

	str = GuiEntryGetText(vscommon->entry);
	end = strrchr(str, '.');

	editable = GTK_EDITABLE(vscommon->entry);

	if (end != NULL)
	{
		guint txtlen, extlen;

		if (strcmp(end, extension) == 0)
		{
			/* Extensions are equal => no replace */

			if (extension != vscommon->format.extension_orig)
			{
				/* malloc'ed */
				free(extension);
			}

			return;

		} /* if (strcmp(end, extension) == 0) */

		txtlen	= gtk_entry_get_text_length(vscommon->entry);
		extlen	= g_utf8_strlen(end, -1);
		pos	= txtlen - extlen;

		gtk_editable_delete_text(editable, pos, -1);

	} /* if (end != NULL) */

	/* Append new extension */
	pos = -1;
	gtk_editable_insert_text(editable, extension, slen, &pos);
	gtk_editable_set_position(editable, -1);

	g_free(vscommon->filechooser_utf8str);
	vscommon->filechooser_utf8str = NULL;

	if (extension != vscommon->format.extension_orig)
	{
		/* malloc'ed */
		free(extension);
	}

} /* static void cff_gui_save_optionmenu_callback(GuiOptionMenu*, cff_save_common_t*) */

static void cff_gui_save_softcut_option_changed_cb(GuiOptionMenu *option, GtkWidget *both_gops)
{
	/*If softcut is "No", disable 'both gops' */

	gboolean sensitive = FALSE;

	if (GuiOptionMenuGetActive(option) > 0)
	{
		sensitive = TRUE;
	}

	gtk_widget_set_sensitive(both_gops, sensitive);

} /* static void cff_gui_save_softcut_option_changed_cb(GuiOptionMenu*, GtkWidget*) */

static GtkWidget *cff_gui_save_build_common(cff_save_common_t *vscommon,
				const gchar *title, gint path_prop, gint width_prop)
{
	gchar		*fname, *str, *end;
	const gchar	*dot, *ext;
	GtkWidget	*entry, *hbox, *vbox, *frame;
	GtkEntryBuffer	*buffer;
	gchar		picture_num[128];

	vscommon->path_prop = path_prop;
	vscommon->okay_bits = 0;

	vscommon->dialog = cff_guitl_dialog_new_cancel_ok(title,
						cff_gui_window_get_topwin());

	cff_guitl_window_setup(vscommon->dialog, width_prop);
	cff_guitl_window_set_topwin(vscommon->dialog);

	frame = gtk_frame_new(NULL);
	GuiContainerSetBorderWidth(frame, 8);
	cff_guitl_dialog_set_child(vscommon->dialog, frame);

#if GTK_MAJOR_VERSION > 2
	gtk_widget_set_vexpand(frame, TRUE);
#endif
	vbox = GuiVBoxNew(FALSE, 4);
	GuiContainerSetBorderWidth(vbox, 4);
	GuiFrameAddChild(frame, vbox);

	hbox = GuiHBoxNew(FALSE, 4);
	GuiContainerSetBorderWidth(hbox, 4);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, 2, hbox);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 4);

	entry = gtk_entry_new();
	vscommon->entry = GTK_ENTRY(entry);
	GuiHBoxPackStart(hbox, entry, TRUE, TRUE, 1);

#if GTK_MAJOR_VERSION > 2
	gtk_widget_set_hexpand(entry, TRUE);
#endif
	cff_guitl_hbox_add_button(hbox, "..",
		(GCallback) cff_gui_save_selection_callback, vscommon);

	vscommon->filechooser_utf8str = NULL;
	fname = cff_gui_gprop_get_filename((vscommon->video)->fname.utf8,
							vscommon->path_prop);
	end = strrchr(fname, '.');
	dot = end;
	ext = NULL;

	if (end != NULL)
	{
		*end = '\0';
		dot  = ".";
	}

	if (vscommon->path_prop == kPropStrFileSavePath)
	{
		if (end != NULL)
		{
			ext = &end[1];
		}

		picture_num[0] = '\0';
	}

	else
	{
		/* kPropStrFileImagePath */
		dot  = ".";
		ext  = "jpg";
		snprintf(picture_num, sizeof(picture_num), "_%u",
				(vscommon->video)->vframe.frame_current);
	}

	str = g_strconcat(fname, picture_num, "_cff", dot, ext, NULL);
	g_free(fname);

	ext = strrchr(str, '.');

	if (ext == NULL)
	{
		vscommon->format.extension_orig = NULL;
	}

	else
	{
		vscommon->format.extension_orig = g_strdup(ext);
	}

	GuiEntrySetText(vscommon->entry, str);
	g_free(str);

	buffer = gtk_entry_get_buffer(vscommon->entry);

	GuiSignalConnect(buffer, "deleted-text",
				cff_gui_save_entry_delete_callback, vscommon);
	GuiSignalConnect(buffer, "inserted-text",
				cff_gui_save_entry_insert_callback, vscommon);

	return vbox;

} /* static GtkWidget *cff_gui_save_build_common(cff_save_common_t *vscommon,
							const gchar*, gint, gint) */

static gchar *cff_gui_save_get_fname(cff_save_common_t *vscommon)
{
	gchar		*fname;
	const gchar	*ename;
	gboolean	do_stop;
	GFile		*file_dst;

	if ((vscommon->video)->fname.utf8 == NULL)
	{
		return NULL;
	}

	ename = GuiEntryGetText(vscommon->entry);

	if (ename == NULL || ename[0] == '\0')
	{
		cff_guitl_beep();
		return NULL;
	}

	fname = g_strdup(ename);
	cff_tl_file_path_unix_to_local(fname);

	file_dst = g_file_parse_name(fname);
	do_stop  = FALSE;

	if (vscommon->filechooser_utf8str == NULL ||
			strcmp(fname, vscommon->filechooser_utf8str) != 0)
	{
		/* Stop icon if equal */
		if (cff_gui_file_is_equal(file_dst,
			(vscommon->video)->fname.utf8, TRUE))
		{
			do_stop = TRUE;
		}

		else if (cff_gui_overwrite(file_dst) == FALSE)
		{
			do_stop = TRUE;
		}

	} /* if (vscommon->filechooser_utf8str == NULL ||
			strcmp(fname, vscommon->filechooser_utf8str) != 0) */

	g_free(fname);

	if (do_stop)
	{
		fname = NULL;
	}

	else
	{
		cff_gui_gprop_set_directory(file_dst, vscommon->path_prop);
		fname = g_file_get_path(file_dst);
	}

	g_object_unref(G_OBJECT(file_dst));

	return fname;

} /* static gchar *cff_gui_save_get_fname(cff_save_common_t*) */


/*
	------------
	 SAVE VIDEO
	------------
*/

static void cff_gui_save_video_prev_settings_free(void)
{
	cff_save_video_prev_options_t *save_prev = gSaveVideoPrevOptionsPtr;

	if (save_prev != NULL)
	{
		g_free(save_prev->stream_idx);
		g_free(save_prev);

		gSaveVideoPrevOptionsPtr = NULL;
	}

} /* static void cff_gui_save_video_prev_settings_free(void) */

static int cff_gui_save_copy_callback(gpointer data)
{
	cff_progressbar_t user_data;

	struct
	{
		const gchar		*str;
		cff_progressbar_t	*pbar;

	} *s = data;

	if (s != NULL)
	{
		s->pbar = &user_data;
	}

	user_data.func  = cff_ffmpeg_copy_cb;
	user_data.video = &gVideo;
	user_data.title = gettext_nop("Copy Video File");
	user_data.mbar  = NULL;
	user_data.mutex = NULL;

	cff_gui_dialog_progress_bar(&user_data);

	if (s != NULL)
	{
		s->str = user_data.error;
	}

	return user_data.retv;

} /* static int cff_gui_save_copy_callback(gpointer) */

static gboolean cff_gui_save_video(cff_save_video_t *vsave)
{
	gboolean	retv;
	gint		i, n, rows, nb_streams, *stream_idx;
	gint		encoder_idx, format_idx, bsfname_idx[kCFFMediaTotal];
	gchar		*fname;
	gchar		*info;
	const gchar	*result;
	GtkTreeModel	*model;
	GtkTreeIter	iter;
	cff_copy_param_t param;

	struct
	{
		const gchar		*error;
		cff_progressbar_t	*pbar;

	} s;


	model = GTK_TREE_MODEL(vsave->out.list.store);
	rows  = gtk_tree_model_iter_n_children(model, NULL);

	if (rows <= 0)
	{
		cff_guitl_beep();
		return FALSE;
	}

	fname = cff_gui_save_get_fname(&(vsave->common));

	if (fname == NULL)
	{
		return FALSE;
	}

	stream_idx = g_malloc(rows * sizeof(gint));

	if (stream_idx == NULL)
	{
		g_free(fname);
		cff_guitl_beep();
		return FALSE;
	}

	g_free(vsave->common.format.extension_orig);
	g_free(vsave->common.filechooser_utf8str);


	/* Save video ... */

	retv = gtk_tree_model_get_iter_first(model, &iter);

	nb_streams = 0;

	while (retv == TRUE)
	{
		gtk_tree_model_get(model, &iter,
				1, &stream_idx[nb_streams], -1);
		nb_streams++;
		retv = gtk_tree_model_iter_next(model, &iter);
	}

	memset(&param, '\0', sizeof(cff_copy_param_t));

	format_idx = 0;

	if (vsave->common.format.u.list != NULL)
	{
		format_idx = GuiOptionMenuGetActive(vsave->common.format.option);

		if (format_idx > 0)
		{
			/* Array without "Auto" */
			param.format_name =
				vsave->common.format.u.list[format_idx - 1].name;
		}

		/* malloc'ed */
		free(vsave->common.format.u.list);
	}

	n = kCFFMediaVideo;

	do
	{
		cff_mediaopts_t *mediaopts = &(vsave->media[n]);

		bsfname_idx[n] = 0;

		if (mediaopts->bsf_list != NULL)
		{
			i = GuiOptionMenuGetActive(mediaopts->bsf_option);
			bsfname_idx[n] = i;

			if (i > 0)
			{
				/* Array without "None" */
				param.media[n].bsf_name = (mediaopts->bsf_list[i - 1]).name;
			}

			/* malloc'ed */
			free(mediaopts->bsf_list);
		}

		if (GuiCheckButtonGetActive(mediaopts->ts_from_duration) == TRUE)
		{
			param.media[n].ts_from_duration = 1;
		}

		else if (GuiCheckButtonGetActive(mediaopts->ts_adjust) == TRUE)
		{
			param.media[n].ts_adjust = 1;
		}
	}
	while (++n < kCFFMediaTotal);

	info		= NULL;
	param.fname	= fname;
	param.info	= &info;
	param.func	= cff_gui_save_copy_callback;
	param.user_data	= &s;

	param.stream_idx = stream_idx;
	param.nb_streams = nb_streams;

	if (GuiCheckButtonGetActive(vsave->shortest) == TRUE)
	{
		param.shortest = 1;
	}

	if (gtk_widget_get_sensitive(vsave->stream_sync) == TRUE &&
		GuiCheckButtonGetActive(vsave->stream_sync) == TRUE)
	{
		param.stream_sync = 1;
	}

	if (GuiCheckButtonGetActive(vsave->ts_keep_start) == TRUE)
	{
		param.ts_keep_start = 1;
	}

	encoder_idx = 0;

	if (vsave->softcut.encoder_names != NULL)
	{
		if (gtk_widget_get_sensitive(vsave->softcut.option) == TRUE &&
			(encoder_idx = GuiOptionMenuGetActive(vsave->softcut.option)) > 0)
		{
			/* Array without "No" */
			param.encoder_name = vsave->softcut.encoder_names[encoder_idx - 1];
			/* Callback */
			param.softcut.func = cff_gui_dialog_progress_softcut_cb;
			param.softcut.data = &(s.pbar);

			if (GuiCheckButtonGetActive(vsave->softcut.both_gops) == TRUE)
			{
				param.softcut.both_gops = 1;
			}
		}

		/* malloc'ed */
		free(vsave->softcut.encoder_names);

	} /* if (vsave->softcut.encoder_names != NULL) */


	gIsBusy |= kIsBusyFile;
	result   = cff_ffmpeg_copy_open(vsave->common.video, &param);
	gIsBusy &= ~kIsBusyFile;

	if (result == NULL)
	{
		result = s.error;
	}

	if (result == NULL)
	{	
		cff_save_video_prev_options_t *save_prev = gSaveVideoPrevOptionsPtr;

		if (save_prev != NULL)
		{
			g_free(save_prev->stream_idx);
			save_prev->stream_idx = param.stream_idx;
			save_prev->nb_streams = param.nb_streams;

			save_prev->format_idx  = format_idx;
			save_prev->encoder_idx = encoder_idx;

			save_prev->shortest      = param.shortest;
			save_prev->stream_sync   = param.stream_sync;
			save_prev->ts_keep_start = param.ts_keep_start;
			save_prev->both_gops     = param.softcut.both_gops;

			n = kCFFMediaVideo;

			do
			{
				save_prev->media[n].bsfname_idx =
						bsfname_idx[n];
				save_prev->media[n].ts_from_duration =
						param.media[n].ts_from_duration;
				save_prev->media[n].ts_adjust =
						param.media[n].ts_adjust;
			}
			while (++n < kCFFMediaTotal);

		} /* if (save_prev != NULL) */
	}

	else
	{
		gint delete;

		if (cff_gprop_get_bool(kPropBoolOptDelete, &delete) == FALSE || delete)
		{
			/* Error: remove the output file */
			g_unlink(fname);
		}

		/* Was not stored in 'save_prev' - free memory */
		g_free(stream_idx);

	} /* if (result == NULL) .. else */


	g_free(fname);

	GuiWindowClose(vsave->common.dialog);

	if (result != NULL)
	{
		cff_guitl_dialog_error(result, NULL);
	}

	else
	{
		cff_guitl_dialog_info((info == NULL) ?
			gettext_nop("No info available!") : info, NULL);

		if (info != NULL)
		{
			/* malloc'ed */
			free(info);
		}

	} /* if (result != NULL) .. else */

	return TRUE;

} /* static gboolean cff_gui_save_video(cff_save_video_t*) */

static void cff_gui_save_ts_duration_toggled(GtkWidget *ts_duration, GtkWidget *ts_adjust)
{
	/* Disable ts_adjust if ts_duration is enabled */

	gboolean active;

	active = ! GuiCheckButtonGetActive(ts_duration);
	gtk_widget_set_sensitive(ts_adjust, active);

} /* static void cff_gui_save_ts_duration_toggled(GtkWidget*, GtkWidget*) */

static void cff_gui_save_list_changed(GtkTreeSelection *selection, cff_save_video_t *vsave)
{
	gint			rows, num;
	guint			bits;
	gboolean		button, up, down, first_row, last_row, stream_sync, softcut;
	GtkTreeModel		*model;
	cff_streamlist_t	*stream;

	if (selection == vsave->avail.list.selection)
	{
		stream = &(vsave->avail);
	}

	else
	{
		stream = &(vsave->out);
	}

	num = gtk_tree_selection_count_selected_rows(selection);

	button = FALSE;

	if (num > 0)
	{
		button = TRUE;
	}

	gtk_widget_set_sensitive(stream->some, button);

	rows = gtk_tree_model_iter_n_children(GTK_TREE_MODEL(stream->list.store), NULL);

	button = FALSE;

	if (rows > 0)
	{
		button = TRUE;
	}

	gtk_widget_set_sensitive(stream->all, button);

	if (selection == vsave->avail.list.selection)
	{
		return;
	}


	bits = vsave->common.okay_bits;

	if (button == FALSE)
	{
		bits |= 0x02;
	}

	else
	{
		bits &= ~0x02;
	}

	cff_gui_save_okay_button_set_sensitive(&(vsave->common), bits);

	first_row = FALSE;
	last_row  = FALSE;

	if (num > 0 && rows > 1)
	{
		GList *list, *ptr;

		up   = TRUE;
		down = TRUE;

		list = gtk_tree_selection_get_selected_rows(selection, &model);
		ptr  = list;

		do
		{
			/*
				This is an array of integers, each representing
				a node in a tree. This value should not be freed.
			*/

			gint *i = gtk_tree_path_get_indices(ptr->data);

			if (*i == 0)
			{
				first_row = TRUE;
			}

			else if (*i == rows - 1)
			{
				last_row = TRUE;
			}

			gtk_tree_path_free(ptr->data);
		}
		while ((ptr = ptr->next) != NULL);

		g_list_free(list);
	}

	else
	{
		up   = FALSE;
		down = FALSE;

	} /* if (num > 0 && rows > 1) .. else */

	gtk_widget_set_sensitive(vsave->up,
			(first_row == FALSE) ? up : FALSE);
	gtk_widget_set_sensitive(vsave->down,
			(last_row == FALSE) ? down : FALSE);


	stream_sync = FALSE;
	softcut     = FALSE;

	if (rows > 0)
	{
		if (cff_ffmpeg_video_cut_list_distance(&gVideo) != 0)
		{
			stream_sync = TRUE;
		}

		if (vsave->softcut.encoder_names != NULL)
		{
			softcut = TRUE;
		}

		if (stream_sync || softcut)
		{
			gint        retv, video, other;
			GtkTreeIter iter;

			video = 0;
			other = 0;

			model = GTK_TREE_MODEL(vsave->out.list.store);
			retv  = gtk_tree_model_get_iter_first(model, &iter);

			while (retv == TRUE)
			{
				gint media, stream_index;

				gtk_tree_model_get(model, &iter,
					1, &stream_index, 2, &media, -1);

				if (media == kCFFMediaVideo &&
					stream_index == gVideo.vframe.stream_index)
				{
					video++;
				}

				else
				{
					other++;
				}

				retv = gtk_tree_model_iter_next(model, &iter);

			} /* while (retv == TRUE) */

			if (video == 0)
			{
				softcut     = FALSE;
				stream_sync = FALSE;
			}

			else if (other == 0)
			{
				stream_sync = FALSE;
			}

		} /* if (stream_sync || softcut) */

	} /* if (rows > 0) */

	gtk_widget_set_sensitive(vsave->stream_sync, stream_sync);
	gtk_widget_set_sensitive(vsave->softcut.option, softcut);
	
	if (softcut && GuiOptionMenuGetActive(vsave->softcut.option) == 0)
	{
		softcut = FALSE;
	}

	gtk_widget_set_sensitive(vsave->softcut.both_gops, softcut);

} /* static void cff_gui_save_list_changed(GtkTreeSelection*, cff_save_video_t*) */

static GList *cff_gui_save_list_get_all(GtkTreeModel *model)
{
	gboolean	retv;
	GList		*list;
	GtkTreeIter	iter;

	list = NULL;
	retv = gtk_tree_model_get_iter_first(model, &iter);

	while (retv == TRUE)
	{
		list = g_list_append(list, GuiMemDup(&iter, sizeof(GtkTreeIter)));
		retv = gtk_tree_model_iter_next(model, &iter);
	}

	return list;

} /* static GList *cff_gui_save_list_get_all(GtkTreeModel*) */

static GList *cff_gui_save_list_get_selected(GtkTreeSelection *selection)
{
	GtkTreeModel	*model;
	GList		*list, *ptr;

	list = gtk_tree_selection_get_selected_rows(selection, &model);

	if ((ptr = list) == NULL)
	{
		return list;
	}

	do
	{
		GtkTreeIter *iter;

		iter = g_malloc(sizeof(GtkTreeIter));
		gtk_tree_model_get_iter(model, iter, ptr->data);
		gtk_tree_path_free(ptr->data);

		ptr->data = iter;
	}
	while ((ptr = ptr->next) != NULL);

	return list;

} /* static GList *cff_gui_save_list_get_selected(GtkTreeSelection*) */

static void cff_gui_save_list_updown(GtkWidget *w, cff_save_video_t *vsave)
{
	gint		add, i, rows, stop, *order;
	GtkTreeModel	*model;
	GtkTreePath	*path;
	GList		*list, *selected;

	add = 1;

	if (w == vsave->up)
	{
		add = -1;
	}

	list = gtk_tree_selection_get_selected_rows(vsave->out.list.selection, &model);

	if (list == NULL)
	{
		return;
	}

	rows = gtk_tree_model_iter_n_children(model, NULL);

	if (rows <= 0)
	{
		return;
	}

	order = malloc(rows * sizeof(gint));

	i = 0;

	do
	{
		order[i] = i;
	}
	while (++i < rows);

	if (add == -1)
	{
		/* Move selected rows up */

		selected = list;
		i = 0;
		stop = rows;
	}

	else
	{
		/* Move selected rows down */

		selected = g_list_last(list);
		i = rows - 1;
		stop = -1;
	}

	path = gtk_tree_path_new_from_indices(i, -1);

	do
	{
		if (gtk_tree_path_compare(path, selected->data) == 0)
		{
			/* Exchange selected row with next (previous) row */

			gint prev = order[i + add];

			order[i + add] = i;
			order[i] = prev;

			gtk_tree_path_free(selected->data);

			if (add == -1)
			{
				selected = selected->next;
			}

			else
			{
				selected = selected->prev;
			}

			if (selected == NULL)
			{
				break;
			}

		} /* if (gtk_tree_path_compare(path, selected->data) == 0) */

		if (add == -1)
		{
			gtk_tree_path_next(path);
		}

		else
		{
			gtk_tree_path_prev(path);
		}
	}
	while ((i -= add) != stop);

	gtk_tree_path_free(path);
	g_list_free(list);

	gtk_list_store_reorder(vsave->out.list.store, order);
	g_free(order);

	/* Set buttons (up, down) sensitive or insensitive */
	cff_gui_save_list_changed(vsave->out.list.selection, vsave);

} /* static void cff_gui_save_list_updown(GtkWidget*, cff_save_video_t*) */

static void cff_gui_save_list_leftright(GtkWidget *w, cff_save_video_t *vsave)
{
	GList		*list, *ptr;
	cff_ListStore_t	*src, *dst;

	if (w == vsave->avail.some || w == vsave->avail.all)
	{
		src = &(vsave->avail.list);
		dst = &(vsave->out.list);
	}

	else
	{
		src = &(vsave->out.list);
		dst = &(vsave->avail.list);
	}

	if (w == vsave->avail.all || w == vsave->out.all)
	{
		list = cff_gui_save_list_get_all(GTK_TREE_MODEL(src->store));
	}

	else
	{
		list = cff_gui_save_list_get_selected(src->selection);
	}

	if ((ptr = list) == NULL)
	{
		return;
	}

	do
	{
		gint		media, stream_idx;
		gchar		*str;
		GtkTreeIter	*iter;

		iter = ptr->data;

		gtk_tree_model_get(GTK_TREE_MODEL(src->store), iter,
				0, &str, 1, &stream_idx, 2, &media, -1);

		gtk_list_store_remove(src->store, iter);

		gtk_list_store_append(dst->store, iter);
		gtk_list_store_set(dst->store, iter,
				0, str, 1, stream_idx, 2, media, -1);

		g_free(str);
		g_free(iter);
	}
	while ((ptr = ptr->next) != NULL);

	g_list_free(list);

	/*
		Set buttons (up, down, right, right-all,
		left, left-all and okay) sensitive or insensitive
	*/

	cff_gui_save_list_changed(src->selection, vsave);
	cff_gui_save_list_changed(dst->selection, vsave);

} /* static void cff_gui_save_list_leftright(GtkWidget*, cff_save_video_t*) */

static void cff_gui_save_list_selection_init(cff_ListStore_t *list,
							cff_save_video_t *vsave)
{
	GtkTreeIter iter;

	if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(list->store), &iter))
	{
		gtk_tree_selection_select_iter(list->selection, &iter);
	}

	GuiSignalConnect(list->selection, "changed",
					cff_gui_save_list_changed, vsave);

	/* Set buttons (up, down, left, left-all and okay) insensitive */
	cff_gui_save_list_changed(list->selection, vsave);

} /* static void cff_gui_save_list_selection_init(cff_ListStore_t*,
							cff_save_video_t*) */

static void cff_gui_build_list(const gchar *title, cff_ListStore_t *list, GtkWidget *hbox)
{
	gchar			*str;
	GtkWidget		*frame, *swin;
	GtkTreeViewColumn	*column;
	GtkCellRenderer		*renderer;

	/* Visible stream descripton, stream-index, media-type (kCFFMedia...) */
	list->store = gtk_list_store_new(3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

	list->view = (GtkTreeView *)
		gtk_tree_view_new_with_model(GTK_TREE_MODEL(list->store));

	gtk_widget_set_can_focus(GTK_WIDGET(list->view), TRUE);

	list->selection = gtk_tree_view_get_selection(list->view);
	gtk_tree_selection_set_mode(list->selection, GTK_SELECTION_MULTIPLE);

	column = gtk_tree_view_column_new();

	str = cff_charconv_utf8_gettext(title);
	gtk_tree_view_column_set_title(column, str);
	CFF_CHARCONV_UTF8_FREE(str, title);

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start(column, renderer, FALSE);
	gtk_tree_view_column_set_attributes(column, renderer, "text", 0, NULL);

	gtk_tree_view_append_column(list->view, column);

	g_object_set(list->view,
			"headers-visible", TRUE,
			"headers-clickable", FALSE,
			NULL);

	g_object_unref(G_OBJECT(list->store));

	swin  = cff_guitl_scrolled_window_new(-1, 150, GTK_WIDGET(list->view));
	frame = cff_guitl_frame_new_add_child_gettext(NULL, 0, swin);
	GuiHBoxPackStart(hbox, frame, TRUE, TRUE, 4);

} /* static void cff_gui_build_list(const gchar*, cff_ListStore_t*, GtkWidget*) */

static GtkWidget *cff_gui_build_option_menu(GtkWidget *parent, GtkWidget *vbox,
		const gchar *title, const gchar *entry, const gchar **name_list)
{
	gchar		*str;
	const gchar	*name;
	GtkWidget	*option, *frame, *hbox;

#if GTK_MAJOR_VERSION == 2
	gint		i;
	GtkWidget	*menu;
#endif

	option = GuiOptionMenuTextNew();
	GuiContainerSetBorderWidth(option, 4);

	if (vbox != NULL)
	{
		GuiVBoxPackStart(vbox, option, FALSE, FALSE, 2);
		frame = cff_guitl_frame_new_add_child_gettext(title, 0, vbox);
		GuiHBoxPackStart(parent, frame, TRUE, FALSE, 2);
	}

	else
	{
		hbox = GuiHBoxNew(FALSE, 2);
		GuiVBoxPackStart(parent, hbox, FALSE, FALSE, 2);
		frame = cff_guitl_frame_new_add_child_gettext(title, 0, option);
		GuiHBoxPackStart(hbox, frame, TRUE, FALSE, 2);
	}


#if GTK_MAJOR_VERSION == 2
	menu = gtk_menu_new();
	gtk_widget_realize(menu);
	gtk_option_menu_set_menu(GTK_OPTION_MENU(option), menu);
	i = 0;
#endif

	if (name_list == NULL || name_list[0] == NULL)
	{
		name_list = NULL;

		if (entry == NULL)
		{
			entry = gettext_nop("Empty");
		}
	}

	if (entry != NULL)
	{
		str = cff_charconv_utf8_gettext(entry);
		GuiOptionMenuTextAppend(option, menu, 0, str);
		CFF_CHARCONV_UTF8_FREE(str, entry);
	}

	if (name_list == NULL)
	{
		gtk_widget_set_sensitive(option, FALSE);
	}

	else
	{
		while ((name = *(name_list++)) != NULL)
		{
#if GTK_MAJOR_VERSION == 2
			i++;
#endif
			GuiOptionMenuTextAppend(option, menu, i, name);
		}

	} /* if (name_list == NULL) .. else */

	GuiOptionMenuSetActive(option, 0);

	return option;

} /* static GtkWidget *cff_gui_build_option_menu(GtkWidget*, GtkWidget*,
					const gchar*, const gchar*, const gchar**) */

static GtkWidget *cff_gui_notebook_append_page(GtkWidget *notebook, const gchar *title)
{
	GtkWidget *label, *vbox;

	vbox = GuiVBoxNew(TRUE, 2);
	GuiContainerSetBorderWidth(vbox, 2);

	label = cff_guitl_label_new_gettext(title);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), vbox, label);

	return vbox;

} /* static GtkWidget *cff_gui_notebook_append_page(GtkWidget*, const gchar*) */

static void cff_gui_build_media_option(GtkWidget *notebook, cff_save_video_t *vsave, gint media)
{
	gchar		*str;
	gchar		**ptr;
	const gchar	*title;
	GtkWidget	*vbox, *hbox;
	cff_mediaopts_t *mediaopts;
	cff_bsflist_t	*bsf_list, *bsf_info;
	gchar		*buf[512];

	if (media == kCFFMediaVideo)
	{
		title = gettext_nop("Video");
	}

	else if (media == kCFFMediaAudio)
	{
		title = gettext_nop("Audio");
	}

	else
	{
		title = gettext_nop("Other");
	}

	mediaopts = &(vsave->media[media]);

	/* Check buttons */

	vbox = GuiVBoxNew(FALSE, 4);

	mediaopts->ts_from_duration =
		cff_guitl_check_button_new_gettext(
			gettext_nop("Create timestamps from packet duration"));
	GuiVBoxPackStart(vbox, mediaopts->ts_from_duration, FALSE, FALSE, 2);

	mediaopts->ts_adjust =
		cff_guitl_check_button_new_gettext(
			gettext_nop("Adjust broken timestamps"));
	GuiVBoxPackStart(vbox, mediaopts->ts_adjust, FALSE, FALSE, 2);

	GuiSignalConnect(mediaopts->ts_from_duration,
		"toggled", cff_gui_save_ts_duration_toggled, mediaopts->ts_adjust);

	/* Notebook */

	hbox = GuiHBoxNew(TRUE, 4);
	GuiHBoxPackStart(hbox, vbox, FALSE, FALSE, 4);

	vbox = cff_gui_notebook_append_page(notebook, title);
	GuiVBoxPackStart(vbox, hbox, FALSE, FALSE, 4);

	/* Option menu */

	bsf_list = cff_ffmpeg_get_bitstream_filter_names(vsave->common.video, media);
	mediaopts->bsf_list = bsf_list;

	ptr = buf;

	if ((bsf_info = bsf_list) != NULL)
	{
		while (bsf_info->name != NULL)
		{
			if (ptr - buf < (sizeof(buf) / sizeof(gchar*)) - 1)
			{
				*(ptr++) = g_strconcat(bsf_info->name, " ",
							bsf_info->codecs, NULL);
			}

			if (bsf_info->codecs != NULL)
			{
				/* malloc'ed */
				free(bsf_info->codecs);
				bsf_info->codecs = NULL;
			}

			bsf_info++;
		}

	} /* if ((bsf_info = bsf_list) != NULL) */

	*ptr = NULL;

	mediaopts->bsf_option = cff_gui_build_option_menu(vbox, NULL,
					gettext_nop("Bitstream Filter"),
					gettext_nop("None"),
					(const gchar **) buf);
	/* Free strings in array */

	ptr = buf;

	while ((str = *(ptr++)) != NULL)
	{
		g_free(str);
	}

} /* static void cff_gui_build_media_option(GtkWidget*, cff_save_video_t*, gint) */


/*
	------------
	 SAVE IMAGE
	------------
*/

static void cff_gui_image_keep_dimension_toggled(GtkWidget *button, cff_save_image_t *vimage)
{
	gint			n;
	gboolean		active;
	cff_spinbutton_t	*spinbutton;

	active = ! GuiCheckButtonGetActive(button);

	spinbutton = &(vimage->width);

	n = 2;

	do
	{
		gint value;

		value = gtk_spin_button_get_value_as_int(spinbutton->widget);
		gtk_spin_button_set_value(spinbutton->widget,
						(gdouble) spinbutton->prev_value);
		spinbutton->prev_value = value;

		gtk_widget_set_sensitive(GTK_WIDGET(spinbutton->widget), active);
		spinbutton = &(vimage->height);
	}
	while (--n > 0);

} /* static void cff_gui_image_keep_dimension_toggled(GtkWidget*, cff_save_image_t*) */

static GtkSpinButton *cff_gui_image_spinbutton(GtkWidget *hbox, gint value, const gchar *title)
{
	GtkAdjustment	*adjustment;
	GtkWidget	*widget, *bbox, *frame;
	GtkSpinButton	*spinbutton;

	bbox = GuiHBoxNew(FALSE, 4);
	GuiContainerSetBorderWidth(bbox, 4);

	frame = cff_guitl_frame_new_add_child_gettext(title, 4, bbox);
	GuiHBoxPackStart(hbox, frame, TRUE, FALSE, 4);

	adjustment = (GtkAdjustment *)
			gtk_adjustment_new((gdouble) value,
						1.0,
						99999.0,
						1.0,
						10.0,
						0.0);

	widget = gtk_spin_button_new(adjustment, 1.0, 0);
	GuiHBoxPackStart(bbox, widget, FALSE, FALSE, 4);

	spinbutton = GTK_SPIN_BUTTON(widget);
	gtk_spin_button_set_numeric(spinbutton, TRUE);
	gtk_spin_button_set_update_policy(spinbutton, GTK_UPDATE_IF_VALID);

	return spinbutton;

} /* static GtkSpinButton *cff_gui_image_spinbutton(GtkWidget*, gint, const gchar*) */

static gboolean cff_gui_save_image(cff_save_image_t *vimage)
{
	gchar		*fname, *info;
	const gchar	*error;
	cff_image_param_t param;

	fname = cff_gui_save_get_fname(&(vimage->common));

	if (fname == NULL)
	{
		return FALSE;
	}

	g_free(vimage->common.format.extension_orig);
	g_free(vimage->common.filechooser_utf8str);


	/* Save image ... */

	memset(&param, '\0', sizeof(cff_image_param_t));

	info		= NULL;
	param.fname	= fname;
	param.info	= &info;

	if (GuiCheckButtonGetActive(vimage->keep_dimension) == TRUE)
	{
		param.keep_dimension = 1;
	}

	if (GuiCheckButtonGetActive(vimage->keep_ratio) == TRUE)
	{
		param.keep_ratio = 1;
	}

	if (GuiCheckButtonGetActive(vimage->gray) == TRUE)
	{
		param.gray = 1;
	}

	if (param.keep_dimension == FALSE)
	{
		param.width  = gtk_spin_button_get_value_as_int(vimage->width.widget);
		param.height = gtk_spin_button_get_value_as_int(vimage->height.widget);
	}

	if (vimage->common.format.u.names != NULL)
	{
		gint i = GuiOptionMenuGetActive(vimage->common.format.option);

		if (i > 0)
		{
			/* Array without "Auto" */
			param.format_name = vimage->common.format.u.names[i - 1];
		}

		/* malloc'ed */
		free(vimage->common.format.u.names);

	} /* if (vimage->common.format.u.names != NULL) */


	gIsBusy |= kIsBusyFile;
	error    = cff_ffmpeg_save_image(vimage->common.video, &param);
	gIsBusy &= ~kIsBusyFile;

	if (error != NULL)
	{
		gint delete;

		if (cff_gprop_get_bool(kPropBoolOptDelete, &delete) == FALSE || delete)
		{
			/* Error: remove the image */
			g_unlink(fname);
		}

	} /* if (error != NULL) */


	g_free(fname);

	GuiWindowClose(vimage->common.dialog);

	if (error != NULL)
	{
		cff_guitl_dialog_error(error, NULL);
	}

	else
	{
		cff_guitl_dialog_info((info == NULL) ?
			gettext_nop("No info available!") : info, NULL);

		if (info != NULL)
		{
			/* malloc'ed */
			free(info);
		}

	} /* if (error != NULL) .. else */

	return TRUE;

} /* static gboolean cff_gui_save_image(cff_save_image_t*) */


/*
	--------------
	 INFO WINDOWS
	--------------
*/

static void cff_gui_scrolled_info_dialog(const char *title, GtkWidget *child, GtkWidget **mbar, int prop)
{
	GtkWidget *frame, *swin;

	GuiContainerSetBorderWidth(child, 8);

	swin  = cff_guitl_scrolled_window_new(320, 180, NULL);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, 4, swin);

#if GTK_MAJOR_VERSION == 2
	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(swin), child);
#else
	/* Above is deprecated since version 3.8 */
	GuiScrolledWindowAddChild(swin, child);
#endif
	cff_guitl_dialog_okay_run(title, frame, mbar, prop);

} /* static void cff_gui_scrolled_info_dialog(const char*, GtkWidget*, GtkWidget**, int) */

static void cff_gui_ffmpeg_info_label(const char *title, char *str, GtkWidget *parent)
{
	GtkWidget *frame, *label;

	if (str == NULL || str[0] == '\0')
	{
		label = cff_guitl_label_new_gettext(gettext_nop("No info available!"));
	}

	else
	{
		PangoFontDescription	*desc;
#if GTK_MAJOR_VERSION > 2
		PangoAttribute		*attr;
		PangoAttrList		*attrlist;
#endif
		str   = cff_tl_replace_non_utf8_chars_in_str(str);
		label = gtk_label_new(str); 
		desc  = pango_font_description_from_string("monospace");

#if GTK_MAJOR_VERSION == 2
		gtk_widget_modify_font(label, desc);
#else
		attr     = pango_attr_font_desc_new(desc);
		attrlist = pango_attr_list_new();
		pango_attr_list_insert(attrlist, attr);
		gtk_label_set_attributes(GTK_LABEL(label), attrlist);
		pango_attr_list_unref(attrlist);
#endif
		pango_font_description_free(desc);

	} /* if (str == NULL || str[0] == '\0') .. else */

	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
	GuiMiscSetAlignment(label, Gui_ALIGN_START, Gui_ALIGN_CENTER);
	GuiMiscSetPadding(label, 12);

	frame = cff_guitl_frame_new_add_child_gettext(title, 4, label);
	GuiVBoxPackStart(parent, frame, FALSE, FALSE, 4);

} /* static void cff_gui_ffmpeg_info_label(const char*, char*, GtkWidget*) */


/*
	-----
	 CSS
	-----
*/

#if GTK_MAJOR_VERSION > 2

static gboolean cff_gui_css_load_from_data(GtkWidget *widget, const gchar *str, guint priority)
{
	GtkCssProvider	*provider;
	GError		*error;

	provider = gtk_css_provider_new();

#if GTK_MAJOR_VERSION == 4

	if (widget == NULL)
	{
		/* Global style provider (for display) */

		GdkDisplay *display;

		/* The data is owned by the instance */
		display = gtk_widget_get_display(gWindata.mainwin);
		gtk_style_context_add_provider_for_display(display,
					GTK_STYLE_PROVIDER(provider),
					priority);
	}

	else

#endif /* GTK_MAJOR_VERSION == 4 */

	{
		/* Widget style provider */

		GtkStyleContext	*context;

 		/* This memory is owned by widget and must not be freed */
		context  = gtk_widget_get_style_context(widget);
		gtk_style_context_add_provider(context,
					GTK_STYLE_PROVIDER(provider),
					priority);

	} /* if (widget == NULL) .. else */

	error = NULL;
	GuiCssLoadFromData(provider, str, &error);

	g_object_unref(provider);

	if (error != NULL)
	{
		g_error_free(error);
		return FALSE;
	}

	return TRUE;

} /* static gboolean cff_gui_css_load_from_data(GtkWidget*, const gchar*, guint) */

#endif /* GTK_MAJOR_VERSION > 2 */


/*
	------------
	 LOG-WINDOW
	------------
*/

static void cff_gui_logwindow_color_parse(gint prop, gpointer color)
{
	gchar *colorstr, *defcolor;

#if GTK_MAJOR_VERSION == 2 && defined(HAVE_SSCANF)
	gchar buf[16];
#endif

	if (prop == kPropStrLogWinText)
	{
		defcolor = (gchar *) gLogWinTextColor;
	}

	else
	{
		defcolor = (gchar *) gLogWinBackgroundColor;
	}

	colorstr = cff_gprop_get_str(prop);

	if (colorstr == NULL)
	{
		colorstr = defcolor;
	}

#if GTK_MAJOR_VERSION == 2 && defined(HAVE_SSCANF)

	if (colorstr[0] == 'r')
	{
		/* If 'cutterffrc' was saved with GTK > 2 */

		gint n, r, g, b;

		n = sscanf(colorstr, "rgb(%i,%i,%i)", &r, &g, &b);

		if (n == 3)
		{
			n = g_snprintf(buf, sizeof(buf),
					"#%02X%02X%02X", r, g, b);
			if (n == 7)
			{
				/* colorstr != defcolor */
				g_free(colorstr);
				colorstr = buf;
			}
		}

	} /* if (colorstr[0] == 'r') */

#endif /* GTK_MAJOR_VERSION == 2 && defined(HAVE_SSCANF) */

	while (GuiColorParse(colorstr, color) == FALSE)
	{
		if (colorstr == defcolor)
		{
			return;
		}

		g_free(colorstr);
		colorstr = defcolor;
	}

	if (colorstr != defcolor)
	{

#if GTK_MAJOR_VERSION == 2 && defined(HAVE_SSCANF)
		if (colorstr != buf)
#endif
		g_free(colorstr);
	}

	return;

} /* static void cff_gui_logwindow_color_parse(gint, gpointer) */

static void cff_gui_logwindow_color_set(cff_logwindow_t *log)
{

#if GTK_MAJOR_VERSION == 2

	GdkColor color;

	cff_gui_logwindow_color_parse(kPropStrLogWinText, &color);
	gtk_widget_modify_text(log->textview, GTK_STATE_NORMAL, &color);

	cff_gui_logwindow_color_parse(kPropStrLogWinBackground, &color);
	gtk_widget_modify_base(log->textview, GTK_STATE_NORMAL, &color);

	gtk_widget_modify_bg(log->window, GTK_STATE_NORMAL, &color);
#else
	guint i;
	gchar *background, *text, buf_wind[256], buf_view[256];

	text = cff_gprop_get_str(kPropStrLogWinText);

	if (text == NULL)
	{
		text = (gchar *) gLogWinTextColor;
	}

	background = cff_gprop_get_str(kPropStrLogWinBackground);

	if (background == NULL)
	{
		background = (gchar *) gLogWinBackgroundColor;
	}

	i = 2;

	do
	{
		g_snprintf(buf_view, sizeof(buf_view),
#if GTK_MAJOR_VERSION == 4
				"GtkTextView,"
				"textview {"
				"  color: %s;"
				"}"
#endif
				"GtkTextView text,"
				"textview text {"
#if GTK_MAJOR_VERSION == 3
				"  color: %s;"
#endif
				"  background-color: %s;"
				"  background-image: none;"
				"}",
				text,
				background);

		g_snprintf(buf_wind, sizeof(buf_wind),
				"GtkWindow,"
				"window {"
				"  background-color: %s;"
				"  background-image: none;"
				"}",
				background);

		if (text != gLogWinTextColor)
		{
			g_free(text);
			text = (gchar *) gLogWinTextColor;
		}

		if (background != gLogWinBackgroundColor)
		{
			g_free(background);
			background = (gchar *) gLogWinBackgroundColor;
		}

		if (cff_gui_css_load_from_data(log->textview, buf_view,
				GTK_STYLE_PROVIDER_PRIORITY_USER) == TRUE)
		{
			cff_gui_css_load_from_data(log->window, buf_wind,
					GTK_STYLE_PROVIDER_PRIORITY_USER);
			break;
		}
	}
	while (--i > 0);

#endif /* #if GTK_MAJOR_VERSION == 2 .. #else */

} /* static void cff_gui_logwindow_color_set(cff_logwindow_t*) */

static GtkWidget *cff_gui_logwindow_colorbutton_new(gint prop, const gchar *title, gpointer color)
{
	gchar		*str;
	GtkWidget	*button, *label, *hbox, *vbox;
	GCallback	func;

	vbox = GuiVBoxNew(FALSE, 5);
	hbox = GuiHBoxNew(FALSE, 5);
	GuiVBoxPackStart(vbox, hbox, FALSE, FALSE, 3);

	cff_gui_logwindow_color_parse(prop, color);
	button = GuiColorButtonNewWithColor(color);
	GuiHBoxPackStart(hbox, button, TRUE, FALSE, 5);

	str = cff_charconv_utf8_gettext(title);
	gtk_color_button_set_title(GTK_COLOR_BUTTON(button), str);

#if GTK_MAJOR_VERSION == 2
	func = (GCallback) gtk_color_button_get_color;
#else
	func = (GCallback) gtk_color_chooser_get_rgba;
#endif
	GuiSignalConnect(button, "color-set", func, color);

	label = gtk_label_new(str);
	GuiVBoxPackStart(vbox, label, FALSE, FALSE, 3);

	CFF_CHARCONV_UTF8_FREE(str, title);

	return vbox;

} /* static GtkWidget *cff_gui_logwindow_colorbutton_new(gint, const gchar*, gpointer) */


#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
static void cff_gui_logwindow_color(GtkWidget *w, cff_logwindow_t *log)
#else
static void cff_gui_logwindow_color(GSimpleAction *a, GVariant *p, cff_logwindow_t *log)
#endif
{
	GtkResponseType	response;
	gchar		*text, *background;
	GtkWidget	*dialog, *frame, *hbox, *widget;

#if GTK_MAJOR_VERSION == 2
	GdkColor	fg, bg;
#else
	GdkRGBA		fg, bg;
#endif

	hbox  = GuiHBoxNew(TRUE, 5);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, 5, hbox);

	widget = cff_gui_logwindow_colorbutton_new(kPropStrLogWinText,
						gettext_nop("Text"), &fg);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 1);

	widget = cff_gui_logwindow_colorbutton_new(kPropStrLogWinBackground,
						gettext_nop("Background"), &bg);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 1);

	dialog = cff_guitl_dialog_new_cancel_ok(gettext_nop("CutterFF: Color"),
									log->window);
#if GTK_MAJOR_VERSION == 4
	gtk_widget_show(dialog);
#endif
	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);
	cff_guitl_dialog_set_child(dialog, frame);

	response = GuiDialogRun(dialog);

	GuiDialogClose(dialog, response);

	if (response == GTK_RESPONSE_ACCEPT)
	{
		text = GuiColorToString(&fg);
		cff_gprop_set_str(kPropStrLogWinText, text);
		g_free(text);

		background = GuiColorToString(&bg);
		cff_gprop_set_str(kPropStrLogWinBackground, background);
		g_free(background);

		cff_gui_logwindow_color_set(log);

	} /* if (response == GTK_RESPONSE_ACCEPT) */

} /* static void cff_gui_logwindow_color(GtkWidget*, cff_logwindow_t*) */


#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

static void cff_gui_logwindow_popup(GtkTextView *textview, GtkMenu *menu, cff_logwindow_t *log)
{
	/* Append to the GtkTextView popup */

	GtkWidget	*mi;
#ifdef HAVE_GMENU
	const gchar	*label;
	gchar		*str;
#endif

	mi = GuiMenuSeparatorNew();
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
	gtk_widget_show(mi);

#ifdef HAVE_GMENU
	label	= gettext_nop("Color ...");
	str	= cff_charconv_utf8_gettext(label);
	mi	= gtk_menu_item_new_with_label(str);
	CFF_CHARCONV_UTF8_FREE(str, label);
#else
	mi = cff_guitl_menu_item_new_gettext(gettext_nop("Color ..."));
#endif

	gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
	gtk_widget_show(mi);

	GuiSignalConnect(mi, "activate", cff_gui_logwindow_color, log);

} /* static void cff_gui_logwindow_popup(GtkTextView*, GtkMenu*, cff_logwindow_t*) */

#endif /* GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 */


#if defined(G_OS_WIN32) && GTK_MAJOR_VERSION == 2

static gboolean cff_gui_logwindow_copy_clipboard_cb(gpointer data)
{
	struct
	{
		gchar	*str;
		gint	len;
		gint	lines;

	} *s = data;

	if (s == NULL)
	{
		return FALSE;
	}

	if (s->str == NULL)
	{
		g_free(s);
		return FALSE;
	}

	if (s->lines >= 0 && s->len > 0 && s->str[0] != '\0')
	{
		if (OpenClipboard(NULL))
		{
			HGLOBAL	clipbuffer;

			EmptyClipboard();
			clipbuffer = GlobalAlloc(GMEM_MOVEABLE, s->len + s->lines + 1);

			if (clipbuffer != NULL)
			{
				char *dst = (char *) GlobalLock(clipbuffer);

				if (dst != NULL)
				{
					/* UNIX to DOS : LF to CR-LF */

					unsigned int	chr;
					char		*src;

					src = s->str;

					do
					{
						chr = *(src++);

						if (chr == 0x0A)
						{
							*(dst++) = 0x0D;
						}

						*(dst++) = chr;
					}
					while (chr != '\0');

					GlobalUnlock(clipbuffer);
					SetClipboardData(CF_TEXT, clipbuffer);
				}

			} /* if (clipbuffer != NULL) */

			CloseClipboard();

		} /* if (OpenClipboard(NULL)) */

	} /* if (s->lines >= 0 && s->len > 0 && s->str[0] != '\0') */

	g_free(s->str);
	g_free(s);

	return FALSE;

} /* static gboolean cff_gui_logwindow_copy_clipboard_cb(gpointer) */

static void cff_gui_logwindow_copy_clipboard(GtkTextView *textview, gpointer data)
{
	/* Copying to the global clipwork do not work with GTK-2 and Windows. */

	struct
	{
		gchar	*str;
		gint	len;
		gint	lines;

	} s;

	GtkTextBuffer	*buffer;
	GtkTextIter	start, end;

	buffer = gtk_text_view_get_buffer(textview);

	if (gtk_text_buffer_get_selection_bounds(buffer, &start, &end) == FALSE)
	{
		return;
	}

	s.str = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);

	if (s.str == NULL)
	{
		return;
	}

	s.len = strlen(s.str);

	if (s.len == 0)
	{
		return;
	}

	s.lines  = gtk_text_iter_get_line(&end);
	s.lines -= gtk_text_iter_get_line(&start);

	if (s.lines < 0)
	{
		return;
	}

	/* Wait a litle bit, else copying does not work */

	g_idle_add_full(G_PRIORITY_LOW,
		(GSourceFunc) cff_gui_logwindow_copy_clipboard_cb,
					g_memdup(&s, sizeof(s)), NULL);

} /* static void cff_gui_logwindow_copy_clipboard(GtkTextView*, gpointer) */

#endif /* #if defined(G_OS_WIN32) && GTK_MAJOR_VERSION == 2 */


static void cff_gui_logwindow_adjustment(GtkAdjustment *vadj, cff_logwindow_t *log)
{
	/* Move the sliders to bottom left */

	if (log != NULL && log->window != NULL)
	{
		gdouble upper, page_size;

		upper     = GuiAdjustmentGetUpper(vadj);
		page_size = GuiAdjustmentGetPageSize(vadj);

		if (upper != log->scrollbar.v_upper ||
				page_size != log->scrollbar.v_pagesize)
		{
			gdouble new_value;

			log->scrollbar.v_upper    = upper;
			log->scrollbar.v_pagesize = page_size;

			new_value = upper - page_size;

			if (new_value > 0.0 &&
				new_value != GuiAdjustmentGetValue(vadj))
			{
				gtk_adjustment_set_value(vadj, new_value);
			}

			if (GuiAdjustmentGetValue(log->scrollbar.h_widget) != 0.0)
			{
				gtk_adjustment_set_value(log->scrollbar.h_widget, 0.0);
			}

		} /* if (upper != log->scrollbar.v_upper ||
					page_size != log->scrollbar.v_pagesize) */

	} /* if (log != NULL && log->window != NULL) */

} /* static void cff_gui_logwindow_adjustment(GtkAdjustment*, cff_logwindow_t*) */

static gboolean cff_gui_logwindow_mainthread(cff_logwindow_t *log)
{
	gint		lines;
	guint		lock;
	GtkTextBuffer	*buffer;
	GtkTextIter	start, end;

	if (log == NULL || log->window == NULL)
	{
		return Gui_SOURCE_REMOVE;
	}

	if (log->utf8.len == 0)
	{
		return Gui_SOURCE_CONTINUE;
	}

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(log->textview));
	lines  = gtk_text_buffer_get_line_count(buffer);

	lock = log->nThreads;

	if (lock)
	{
		GuiMutexLock(&(log->mutex));
	}

	if (log->utf8.overwrite)
	{
		log->utf8.overwrite = FALSE;

		/* Overwrite last line */
		gtk_text_buffer_get_iter_at_line(buffer, &start, lines - 1);
		gtk_text_buffer_get_end_iter(buffer, &end);
		gtk_text_buffer_delete(buffer, &start, &end);
	}

	else if (lines > 1000)
	{
		/* Delete some lines */
		gtk_text_buffer_get_start_iter(buffer, &start);
		gtk_text_buffer_get_iter_at_line(buffer, &end, lines - 250);
		gtk_text_buffer_delete(buffer, &start, &end);
	}

	gtk_text_buffer_get_end_iter(buffer, &end);
	gtk_text_buffer_insert(buffer, &end, log->utf8.str, log->utf8.len);

	if (log->utf8.max > 4096)
	{
		log->utf8.max = 4096;
		log->utf8.str = g_realloc(log->utf8.str, log->utf8.max);
	}

	log->utf8.len = 0;

	if (lock)
	{
		GuiMutexUnlock(&(log->mutex));
	}

	return Gui_SOURCE_CONTINUE;

} /* static gboolean cff_gui_logwindow_mainthread(cff_logwindow_t*) */

static int cff_gui_logwindow_cb(const char *str, int len, void *data)
{
	gint		slen;
	guint		lock, max, firstchar;
	gchar		*utf8, *ptr;
	cff_logwindow_t	*log;

	if (str == NULL || len == 0)
	{
		return 0;
	}

	log = data;

	if (log == NULL || log->window == NULL || log->timeout_id == 0)
	{
		/* Restore previous level */
		return -1;
	}

	firstchar = str[0];

	if (firstchar == '\r')
	{
		str++;
		len--;
	}
		
	slen = len;
	utf8 = cff_charconv_utf8_from_locale(str, &slen);
	lock = log->nThreads;

	if (lock)
	{
		if (GuiMutexTryLock(&(log->mutex)) == FALSE)
		{
			if (log->window == NULL)
			{
				CFF_CHARCONV_UTF8_FREE(utf8, str);
				return -1;
			}

			GuiMutexLock(&(log->mutex));
		}
	}

	len = log->utf8.len + slen;
	max = log->utf8.max;

	/* 2 => '\n' + '\0' */

	while (len + 2 > max)
	{
		max += 4096;
	}

	if (max > log->utf8.max)
	{
		log->utf8.max = max;
		log->utf8.str = g_realloc(log->utf8.str, max);
	}

	ptr = &log->utf8.str[log->utf8.len];

	if (firstchar == '\r')
	{
		if (log->utf8.cr_len != 0)
		{
			if (log->utf8.len == 0)
			{
				/* Repeat - delete last line */
				log->utf8.overwrite = TRUE;
			}

			else if (log->utf8.len >= log->utf8.cr_len)
			{
				/* Overwrite last characters */
				ptr -= log->utf8.cr_len;
				len -= log->utf8.cr_len;
			}

			/* Repeat - avoid double newline */
			log->utf8.lastchar = '\r';

		} /* if (log->utf8.cr_len != 0) */

		log->utf8.cr_len = slen;

		if (utf8[slen - 1] == '\n')
		{
			log->utf8.cr_len--;
		}
	}

	else
	{
		log->utf8.cr_len = 0;

	} /* if (firstchar == '\r') .. else */


	/*
		If the newline character from the
		last line was removed, append it now.
	*/

	if (log->utf8.lastchar == '\n')
	{
		*(ptr++) = '\n';
		len++;
	}

	ptr = strcpy(ptr, utf8);
	CFF_CHARCONV_UTF8_FREE(utf8, str);

	/* If the last character is newline, remove it. */

	log->utf8.lastchar = ptr[slen - 1];

	if (log->utf8.lastchar == '\n')
	{
		len--;
	}

	log->utf8.len = len;

	if (lock)
	{
		GuiMutexUnlock(&(log->mutex));
	}

	return 0;

} /* static int cff_gui_logwindow_cb(const char*, int, void*) */

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
static gboolean cff_gui_logwindow_free(GtkWidget *widget, GdkEvent *e, gpointer data)
#else
static gboolean cff_gui_logwindow_free(GtkWidget *widget, gpointer data)
#endif
{
	const gchar	*logname;
	cff_logwindow_t *log;

	if (widget == GINT_TO_POINTER(-1))
	{
		log = &gLogWindow;
	}

	else
	{
		log = data;
	}

	/*
		If the log-window is currently busy, and
		'log->nThreads' is not zero, ffmpeg runs
		in another thread.
		If 'widget' is NULL, this function was called
		from 'cff_gui_opt_logwindow_menu_selected'
	*/

	if (log->nThreads == 0)
	{
		if (widget != NULL)
		{
			/* Destroy the window */
			log->window = NULL;
		}

		if (log->timeout_id != 0)
		{
			g_source_remove(log->timeout_id);
		}
	}

	else /* widget != NULL => delete-event */
	{

#ifndef HAVE_SEMAPHORE
		/* Keep the window open */
		return TRUE;
#else
		/* Close the window */
		log->window = NULL;

		if (log->timeout_id != 0)
		{
			if (log->busy_loop != 0)
			{
				/* Try again */
				return Gui_SOURCE_CONTINUE;
			}

			g_idle_add_full(G_PRIORITY_LOW,
				(GSourceFunc) cff_gui_logwindow_free,
						GINT_TO_POINTER(-1), NULL);
			log->busy_loop = 10;

			/* Destroy the window */
			return FALSE;

		} /* if (log->timeout_id != 0) */

		if (--log->busy_loop > 0 &&
				cff_ffmpeg_get_current_loglevel_index() > 0)
		{
			/* While previous level was not restored */
			return Gui_SOURCE_CONTINUE;
		}

#endif /* #ifndef HAVE_SEMAPHORE .. #else */

	} /* if (log->nThreads == 0) .. else */


	g_free(log->utf8.str);
	log->utf8.str  = NULL;
	log->busy_loop = 0;

	/* If set to 'quiet', prepend newline in the next output */
	log->utf8.lastchar = '\n';

	if (cff_ffmpeg_get_current_loglevel_index() > 0)
	{
		/* Restore previous level */
		cff_ffmpeg_set_loglevel(NULL, log->nThreads, NULL, NULL);
	}

	logname = cff_ffmpeg_get_loglevel_name(0);
	cff_gprop_set_str(kPropStrOptLogLevel, logname);

	return FALSE;

} /* static gboolean cff_gui_logwindow_free(GtkWidget*, GdkEvent*, gpointer) */

static void cff_gui_logwindow_set_title(GtkWidget *window, const gchar *logname)
{
	static const gchar logwin_title[] = "Cutterff: FFmpeg-Log (%s)";

	if (logname != NULL)
	{
		gchar buf[256];

		g_snprintf(buf, sizeof(buf), logwin_title, logname);
		gtk_window_set_title(GTK_WINDOW(window), buf);
	}

} /* static void cff_gui_logwindow_set_title(GtkWidget*, const gchar*) */

static void cff_gui_logwindow_source_destroyed(cff_logwindow_t *log)
{
	log->timeout_id = 0;

} /* static void cff_gui_logwindow_source_destroyed(cff_logwindow_t*) */

static void cff_gui_logwindow_add_timeout(cff_logwindow_t *log)
{
	if (log->timeout_id == 0)
	{
		log->utf8.str		= NULL;
		log->utf8.len		= 0;
		log->utf8.max		= 0;
		log->utf8.cr_len	= 0;
		log->utf8.overwrite	= FALSE;
		log->busy_loop		= 0;

		/* 5ms */
		log->timeout_id =
			g_timeout_add_full(CFF_TEXT_VIEW_PRIORITY_VALIDATE, 5,
				(GSourceFunc) cff_gui_logwindow_mainthread,
					log, (GDestroyNotify)
						cff_gui_logwindow_source_destroyed);
	} /* if (log->timeout_id == 0) */

} /* static void cff_gui_logwindow_add_timeout(cff_logwindow_t*) */

static gboolean cff_gui_logwindow_open(const gchar *logname)
{
	GtkWidget		*widget;
	GtkScrolledWindow	*swin;
	GtkAdjustment		*vadj;
	cff_logwindow_t		*log;
#if GTK_MAJOR_VERSION == 2
	PangoFontDescription	*desc;
#elif GTK_MAJOR_VERSION == 4
	GMenu			*menu, *section;
	GMenuItem		*item;
	GSimpleAction		*action;
	GSimpleActionGroup	*group;
#endif

	log = &gLogWindow;

	if (cff_ffmpeg_set_loglevel(logname, 0, cff_gui_logwindow_cb, log) < 0)
	{
		return FALSE;
	}

	if (log->window != NULL)
	{
		cff_gui_logwindow_add_timeout(log);
		cff_gui_logwindow_set_title(log->window, logname);
		return TRUE;
	}

	log->textview = gtk_text_view_new();

#if GTK_MAJOR_VERSION == 2
	desc = pango_font_description_from_string("monospace");
	gtk_widget_modify_font(log->textview, desc);
	pango_font_description_free(desc);
#else
	cff_gui_css_load_from_data(log->textview,
				"GtkTextView,"
				"textview {"
				"  font-family: monospace;"
				"}",
				GTK_STYLE_PROVIDER_PRIORITY_USER);
#endif

	g_object_set(G_OBJECT(log->textview), "cursor-visible", FALSE,
						"editable", FALSE,
						"wrap-mode", GTK_WRAP_NONE,
						NULL);

	widget	= cff_guitl_scrolled_window_new(320, 180, log->textview);
	swin	= GTK_SCROLLED_WINDOW(widget);

#if GTK_MAJOR_VERSION == 3
	gtk_scrolled_window_set_shadow_type(swin, GTK_SHADOW_NONE);
#endif
	log->scrollbar.v_upper  = 0.0;
	log->scrollbar.h_widget = gtk_scrolled_window_get_hadjustment(swin);

	vadj = gtk_scrolled_window_get_vadjustment(swin);
	GuiSignalConnect(vadj, "changed", cff_gui_logwindow_adjustment, log);

	log->window = cff_guitl_window_new(GTK_WIDGET(swin), 8);
	cff_gui_logwindow_set_title(log->window, logname);
	cff_guitl_window_setup(log->window, kPropIntLogWinWidth);

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
	GuiSignalConnect(log->textview, "populate-popup",
					cff_gui_logwindow_popup, log);
#else
	menu	= g_menu_new();
	section	= g_menu_new();

	item = g_menu_item_new_section(NULL, G_MENU_MODEL(section));
	g_menu_append_item(menu, item);
	g_object_unref(item);

	/*
		Keep this order:
			1) g_menu_item_set_detailed_action
			2) g_menu_append_item
	*/

	item = cff_guitl_menu_item_new_gettext(gettext_nop("Color ..."));
	g_menu_item_set_detailed_action(item, "txt.setcolor");
	g_menu_append_item(section, item);
	g_object_unref(item);

	group  = g_simple_action_group_new();
	action = g_simple_action_new("setcolor", NULL);
	g_action_map_add_action(G_ACTION_MAP(group), G_ACTION(action));

	GuiSignalConnect(action, "activate", cff_gui_logwindow_color, log);

	gtk_widget_insert_action_group(log->textview,
					"txt", G_ACTION_GROUP(group));
	gtk_text_view_set_extra_menu(GTK_TEXT_VIEW(log->textview),
							G_MENU_MODEL(menu));
	g_object_unref(section);
	g_object_unref(menu);

#endif /* #if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 .. #else */

#if defined(G_OS_WIN32) && GTK_MAJOR_VERSION == 2
	GuiSignalConnect(log->textview, "copy-clipboard",
					cff_gui_logwindow_copy_clipboard, NULL);
#endif
	GuiSignalConnect(log->window, gDeleteEventStr,
					cff_gui_logwindow_free, log);

	/* Ignore these events if busy */
	cff_gui_signal_connect_key_ignore(log->window);
	cff_gui_signal_connect_mouse_button_ignore(log->window);
	cff_gui_signal_connect_key_ignore(log->textview);
	cff_gui_signal_connect_mouse_button_ignore(log->textview);

	log->utf8.lastchar = '\0';
	log->timeout_id    = 0;

	cff_gui_logwindow_add_timeout(log);

	cff_gui_logwindow_color_set(log);
	GuiWidgetShowAll(log->window);

	return TRUE;

} /* static gboolean cff_gui_logwindow_open(const gchar*) */


/*
	-----------
	 RECTANGLE
	-----------
*/

#ifdef HAVE_CAIRO

static void cff_gui_cairo_draw_rectangle(cairo_t *cr, gint color,
					gint fill, const GdkRectangle *rect)
{
	/*
		The signal handler will get a 'cr' with a clip region
		already set to the widget's dirty region, i.e. to the
		area that needs repainting.
	*/

#if GTK_MAJOR_VERSION == 2
	gdk_cairo_set_source_color(cr, &gColor[color]);
#else
	gdk_cairo_set_source_rgba(cr, &gRGBA[color]);
#endif
	gdk_cairo_rectangle(cr, rect);

	if (fill)
	{
		cairo_fill(cr);
	}

	else
	{
		cairo_set_line_width(cr, 1.0);
		cairo_stroke(cr);
	}

} /* static void cff_gui_cairo_draw_rectangle(cairo_t*, gint, gint,
						const GdkRectangle*) */
#if GTK_MAJOR_VERSION == 2

static void cff_gui_cairo_draw_rect_gtk2(cairo_t *cr, const GdkEventExpose *e,
					gint color, gint fill, const GdkRectangle *rect)
{
	/* Calculates the intersection of two rectangles */

	GdkRectangle dst;

	if (gdk_rectangle_intersect(&(e->area), rect, &dst))
	{
		cff_gui_cairo_draw_rectangle(cr, color, fill,
					(fill == FALSE) ? rect : &dst);
	}

} /* static void cff_gui_cairo_draw_rect_gtk2(cairo_t*, const GdkEventExpose*, gint, gint, const GdkRectangle*) */

#endif /* GTK_MAJOR_VERSION == 2 */

#else /* ! HAVE_CAIRO */

static void cff_gui_gdk_draw_rectangle(GdkEventExpose *e,
			gint color, gint fill, const GdkRectangle *rect)
{
	/* Calculates the intersection of two rectangles */

	GdkRectangle dst;

	if (gdk_rectangle_intersect(&(e->area), rect, &dst))
	{
		const GdkRectangle *ptr = &dst;

		if (fill == FALSE)
		{
			ptr = rect;
		}	

		gdk_draw_rectangle(e->window,
				gContext[color],
				fill,
				ptr->x,
				ptr->y,
				ptr->width,
				ptr->height);
	}

} /*static void cff_gui_gdk_draw_rectangle(GdkEventExpose*, gint, gint, const GdkRectangle*) */

#endif /* #ifdef HAVE_CAIRO .. #else */


static void cff_gui_calculate_frame_rects(cff_video_frame_t *current_frame,
							const cff_video_t *video)
{
	gint x, y, w, h, width, height;

	width  = GuiWidgetGetAllocatedWidth(current_frame->drawingArea);
	height = GuiWidgetGetAllocatedHeight(current_frame->drawingArea);

	/* Current video frame */

	current_frame->rectRGB24.x = (width - video->vframe.rgb24.width) / 2;
	current_frame->rectRGB24.y = (height - video->vframe.rgb24.height) / 2;
	current_frame->rectRGB24.width  = video->vframe.rgb24.width;
	current_frame->rectRGB24.height = video->vframe.rgb24.height;

	/* Left or top bar */

	w = width;

	if (width > video->vframe.rgb24.width)
	{
		w = current_frame->rectRGB24.x;
	}

	h = height;

	if (height > video->vframe.rgb24.height)
	{
		h = current_frame->rectRGB24.y;
	}

	current_frame->rectLeftTop.x = 0;
	current_frame->rectLeftTop.y = 0;
	current_frame->rectLeftTop.width  = w;
	current_frame->rectLeftTop.height = h;

	/* Right or bottom bar */

	x = 0;
	y = 0;

	if (current_frame->rectRGB24.x != 0)
	{
		x = current_frame->rectRGB24.x + current_frame->rectRGB24.width;
	}

	if (current_frame->rectRGB24.y != 0)
	{
		y = current_frame->rectRGB24.y + current_frame->rectRGB24.height;
	}

	current_frame->rectRightBottom.x = x;
	current_frame->rectRightBottom.y = y;
	current_frame->rectRightBottom.width  = width - x;
	current_frame->rectRightBottom.height = height - y;

	current_frame->dataRGB24 = (const guchar **) &(video->vframe.rgb24.data);

#if ! defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2
	current_frame->rowstrideRGB24 = video->vframe.rgb24.width * 3;
#endif

	/* Clear the whole area */
	GuiWidgetQueueDrawArea(current_frame->drawingArea);

} /* static void cff_gui_calculate_frame_rects(cff_video_frame_t*, const cff_video_t*) */


/*
	-------
	 CAIRO
	-------
*/

#ifdef HAVE_CAIRO

static gboolean cff_gui_cairo_surface_image_create(cff_video_frame_t *current_frame,
								const cff_video_t *video)
{
	cairo_surface_t *ptr;

	cff_gui_calculate_frame_rects(current_frame, video);

	if (current_frame->surfaceRGB24 != NULL)
	{
		cairo_surface_destroy(current_frame->surfaceRGB24);
	}

	ptr = cairo_image_surface_create(CAIRO_FORMAT_RGB24,
					video->vframe.rgb24.width,
					video->vframe.rgb24.height);
	if (ptr != NULL &&
		cairo_surface_status(ptr) != CAIRO_STATUS_SUCCESS)
	{
		cairo_surface_destroy(ptr);
		ptr = NULL;
	}

	current_frame->surfaceRGB24 = ptr;

	return (ptr == NULL) ? FALSE : TRUE;

} /* static gboolean cff_gui_cairo_surface_image_create(cff_video_frame_t*,
								const cff_video_t*) */

static void cff_gui_cairo_surface_image_fill(cff_video_frame_t *current_frame)
{
	int			i;
	unsigned char		z, *dst;
	const unsigned char	*src;
	cairo_surface_t		*surface;

	surface = current_frame->surfaceRGB24;

	if (surface == NULL)
	{
		return;
	}

	dst = cairo_image_surface_get_data(surface);

	if (dst == NULL)
	{
		return;
	}

	if (current_frame->dataRGB24 == NULL)
	{
		return;
	}

	src = *(current_frame->dataRGB24);

	if (src == NULL)
	{
		return;
	}

#if G_BYTE_ORDER != G_BIG_ENDIAN
	src += 3;
#endif

	i = current_frame->rectRGB24.width * current_frame->rectRGB24.height;
	z = 0;

	do
	{

#if G_BYTE_ORDER == G_BIG_ENDIAN
		*(dst++) = z;
		*(dst++) = *(src++);
		*(dst++) = *(src++);
		*(dst++) = *(src++);
#else
		*(dst++) = *(--src);
		*(dst++) = *(--src);
		*(dst++) = *(--src);
		*(dst++) = z;
		src += 6;
#endif

	}
	while (--i > 0);

} /* static void cff_gui_cairo_surface_image_fill(cff_video_frame_t*) */

#endif /* HAVE_CAIRO */


/*
	-------
	 VIDEO
	-------
*/

static void cff_gui_video_button_set_sensitive(gint sensitive)
{
	/*
		Called from
		  'cff_gui_init'
		  'cff_gui_video_open'
		  'cff_gui_file_close'
	*/

	gint		i;
	GtkWidget	**button;

	button = gui_button_video_frame;
	i = 0;

	do
	{
		gtk_widget_set_sensitive(*(button++), sensitive);
	}
	while (++i < kButtonVideoFrameTotal);

	gtk_widget_set_sensitive(gSlider.scale, sensitive);
	gtk_widget_set_sensitive(gButtonVideoCut[kButtonVideoCutAdd], sensitive);
	gtk_widget_set_sensitive(gButtonVideoCut[kButtonVideoCutRemove], FALSE);

} /* static void cff_gui_video_button_set_sensitive(gint) */

static void cff_gui_video_button_cut_set_sensitive(gboolean sensitive)
{
	/* Either add or remove a cut point */

	gtk_widget_set_sensitive(gButtonVideoCut[kButtonVideoCutAdd], sensitive);

	sensitive = (sensitive == FALSE) ? TRUE : FALSE;
	gtk_widget_set_sensitive(gButtonVideoCut[kButtonVideoCutRemove], sensitive);

} /* static void cff_gui_video_button_cut_set_sensitive(gboolean) */

static void cff_gui_video_button_cut_update_sensitive(const cff_video_t *video)
{
	gboolean cut_add;

	if (cff_ffmpeg_video_cut_find_current(video) >= 0)
	{
		/* Found */
		cut_add = FALSE;
	}

	else
	{
		/* Not found */
		cut_add = TRUE;
	}

	cff_gui_video_button_cut_set_sensitive(cut_add);

} /* static void cff_gui_video_button_cut_update_sensitive(const cff_video_t*) */

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

static void cff_gui_widget_queue_draw_area(GtkWidget *widget)
{
	GdkRectangle rect;

	rect.x = 0;
	rect.y = 0;
	rect.width  = GuiWidgetGetAllocatedWidth(widget);
	rect.height = GuiWidgetGetAllocatedHeight(widget);

	gdk_window_invalidate_rect(GuiWidgetGetWindow(widget), &rect, FALSE);

} /* static void cff_gui_widget_queue_draw_area(GtkWidget*) */

#endif /* GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 */


static void cff_gui_frame_info(cff_video_frame_t *current_frame, const cff_video_t *video)
{
	gchar buf[128];

	snprintf(buf, sizeof(buf), "[%ux%u] [%u:%u] %s-Frame : %u / %u",
			video->vframe.frame_width,
			video->vframe.frame_height,
			video->vframe.frame_aspect[0],
			video->vframe.frame_aspect[1],
			video->vframe.pict_type,
			video->vframe.frame_current,
			video->vframe.frame_total);

	gtk_label_set_text(current_frame->info, buf);

	snprintf(buf, sizeof(buf), "Time: %02u:%02u:%02u.%03u / %02u:%02u:%02u.%03u",
			video->vframe.time_current[0],
			video->vframe.time_current[1],
			video->vframe.time_current[2],
			video->vframe.time_current[3],
			video->vframe.time_total[0],
			video->vframe.time_total[1],
			video->vframe.time_total[2],
			video->vframe.time_total[3]);

	gtk_label_set_text(current_frame->time, buf);

} /* static void cff_gui_frame_info(cff_video_frame_t*, const cff_video_t*) */


#if GTK_MAJOR_VERSION == 2
static gboolean cff_gui_draw_frame(GtkWidget *widget, GdkEvent *e, cff_video_frame_t *current_frame)
#elif GTK_MAJOR_VERSION == 3
static gboolean cff_gui_draw_frame(GtkWidget *widget, cairo_t *cr, cff_video_frame_t *current_frame)
#else /* GTK_MAJOR_VERSION == 4 */
static gboolean cff_gui_draw_frame(GtkDrawingArea *widget, cairo_t *cr, gint width, gint height, cff_video_frame_t *current_frame)
#endif
{

#if defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2
	/* Creates a Cairo context for drawing to 'drawable' */
	cairo_t *cr = gdk_cairo_create(widget->window);
#endif

	if (current_frame->dataRGB24 == NULL ||
				*(current_frame->dataRGB24) == NULL)
	{
		/* No frame available, clear the whole area */

		GdkRectangle rect;

		rect.x = 0;
		rect.y = 0;

#if GTK_MAJOR_VERSION == 4
		rect.width  = width;
		rect.height = height;
#else
		rect.width  = GuiWidgetGetAllocatedWidth(widget);
		rect.height = GuiWidgetGetAllocatedHeight(widget);
#endif
		GuiDrawRectangle(cr, e, kColorGrey, TRUE, &rect);
	}

	else
	{
		/*
			Draw rectangles left and right (or top
			and bottom), and the frame in the middle
		*/

		GuiDrawRectangle(cr, e, kColorDarkGrey, TRUE,
					&(current_frame->rectLeftTop));

		GuiDrawRectangle(cr, e, kColorDarkGrey, TRUE,
					&(current_frame->rectRightBottom));

#ifdef HAVE_CAIRO

		if (current_frame->surfaceRGB24 != NULL)
		{
			cairo_set_source_surface(cr,
				current_frame->surfaceRGB24,
					current_frame->rectRGB24.x,
						current_frame->rectRGB24.y);
			cairo_paint(cr);
		}
#else
		gdk_draw_rgb_image(widget->window,
			gContext[kColorBlack],
			current_frame->rectRGB24.x,
			current_frame->rectRGB24.y,
			current_frame->rectRGB24.width,
			current_frame->rectRGB24.height,
			GDK_RGB_DITHER_NONE,
			*(current_frame->dataRGB24),
			current_frame->rowstrideRGB24);

#endif /* #ifdef HAVE_CAIRO .. #else */

	} /* if (current_frame->dataRGB24 == NULL ||
				*(current_frame->dataRGB24) == NULL) .. else */

#if defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2
	cairo_destroy(cr);
#endif

	return TRUE;

} /* static gboolean cff_gui_draw_frame(GtkWidget*, GdkEvent*, cff_video_frame_t*) */


#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
static void cff_gui_draw_frame_sizealloc_cb(GtkWidget *widget,
					GdkRectangle *rect, cff_video_t *video)
#else /* GTK_MAJOR_VERSION == 4 */
static void cff_gui_draw_frame_sizealloc_cb(GtkWidget *widget,
				gint width, gint height, cff_video_t *video)
#endif
{
	/* 'gCurrentVideoFrame.drawingArea' has changed its size */

	if (video->vframe.rgb24.data != NULL)
	{
		cff_video_frame_t *current_frame;

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
		gint width, height;

		/* GdkRectangle == GtkAlloation */
		width  = rect->width;
		height = rect->height;
#endif
		if (cff_ffmpeg_frame2rgb24(video, width, height) < 0)
		{
			return;
		}

		current_frame = &gCurrentVideoFrame;

#ifdef HAVE_CAIRO
		cff_gui_cairo_surface_image_create(current_frame, video);
		cff_gui_cairo_surface_image_fill(current_frame);
#else
		cff_gui_calculate_frame_rects(current_frame, video);
#endif
	}

} /* static void cff_gui_draw_frame_sizealloc_cb(GtkWidget*, GdkRectangle*, cff_video_t*) */


#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

static void cff_gui_mainwindow_cursor_new(GtkWidget *widget, GdkScreen *prev_screen, gpointer data)
{
	GdkDisplay	*display;
	cff_cursor_t	*cursor;

	cursor	= data;
	display = gtk_widget_get_display(widget);

	if (cursor->busy != NULL)
	{
		if (gdk_cursor_get_display(cursor->busy) == display)
		{
			return;
		}

		GuiCursorUnref(cursor->busy);
	}

	cursor->busy   = gdk_cursor_new_for_display(display, GDK_WATCH);
	cursor->window = GuiWidgetGetWindow(widget);

} /* static void cff_gui_mainwindow_cursor_new(GtkWidget*, GdkScreen*, gpointer) */

#endif /* GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 */


static gpointer cff_gui_video_frame_thread(gpointer data)
{
	gint retv;

	struct
	{
		cff_video_t		*video;
		cff_frame_seek_t	*frame_seek;
		gint			width;
		gint			height;
		enum cff_FrameFlags	frame_flags;

	} *s = data;

	switch ((s->frame_seek)->type)
	{
		unsigned int next;

		case kFrameSeekNext:
			retv = cff_ffmpeg_get_frame_next(s->video,
				(s->frame_seek)->value.e_next,
					s->width, s->height, s->frame_flags);
			break;

		case kFrameSeekCutpoint:
			retv = cff_ffmpeg_get_frame_cutpoint(s->video,
				(s->frame_seek)->value.e_next,
					s->width, s->height, s->frame_flags);
			break;

		case kFrameSeekSliderClick:
			retv = cff_ffmpeg_get_frame_slider_click(s->video,
				(s->frame_seek)->value.d_next,
					s->width, s->height, s->frame_flags);
			break;

		case kFrameSeekSliderMove:
			next = (s->frame_seek)->value.i_next;
			retv = cff_ffmpeg_get_frame_slider_move(s->video,
				next, s->width, s->height, s->frame_flags);
			break;

		default:
			retv = -1;
			break;

	} /* switch ((s->frame_seek)->type) */

	return GINT_TO_POINTER(retv);

} /* static gpointer cff_gui_video_frame_thread(gpointer) */

static gint cff_gui_video_frame(cff_frame_seek_t *frame_seek)
{
	gint			cursor, retv;
	gpointer		result;
	cff_slider_t		*slider;
	cff_video_t		*video;
	cff_video_frame_t	*current_frame;

	struct
	{
		cff_video_t		*video;
		cff_frame_seek_t	*frame_seek;
		gint			width;
		gint			height;
		enum cff_FrameFlags	frame_flags;
	} s;

	video = &gVideo;

	if (video->vframe.ic == NULL)
	{
		return -1;
	}

	if ((gIsBusy & kIsBusyVideoFrame) || GuiMainLevelGetBusy())
	{
		return -1;
	}

	gIsBusy |= kIsBusyVideoFrame;

	s.video       = video;
	s.frame_seek  = frame_seek;
	current_frame = &gCurrentVideoFrame;

	s.width  = GuiWidgetGetAllocatedWidth(current_frame->drawingArea);
	s.height = GuiWidgetGetAllocatedHeight(current_frame->drawingArea);

	s.frame_flags = 0;

	slider = &gSlider;

	if (slider->is_slider == FALSE)
	{
		gint flag;

		if (cff_gprop_get_bool(kPropBoolOptKeyFrame, &flag) && flag != 0)
		{
			s.frame_flags |= kKeyFrameFlag;
		}

		if (cff_gprop_get_bool(kPropBoolOptIFrame, &flag) && flag != 0)
		{
			s.frame_flags |= kIFrameFlag;
		}

		if (cff_gprop_get_bool(kPropBoolOptPFrame, &flag) && flag != 0)
		{
			s.frame_flags |= kPFrameFlag;
		}

	} /* if (slider->is_slider == FALSE) */

	if (cff_gprop_get_bool(kPropBoolOptCursor, &cursor) == FALSE)
	{
		cursor = TRUE;
	}

	if (cursor)
	{
		if (gCursor.window != NULL)
		{
			GuiWindowSetCursor(gCursor.window, gCursor.busy);

			if (gThread.thread == NULL)
			{
				/* Draw the cursor and/or the slider */
				while (g_main_context_iteration(NULL, FALSE)) {;}
			}
		}

	} /* if (cursor) */

#if GTK_MAJOR_VERSION == 4
	if (slider->is_slider)
	{
		/*
			Sometimes freeze if slider + multi-thread
			 (includes 'g_main_loop_run') (???)
			Drawing the window contents works still fine,
			but key and mouse events are ignored !
			 Workaround => run single-thread
		*/

		result = cff_gui_video_frame_thread(&s);
	}
	else
#endif
	{
		result = cff_gui_thread_invoke((GThreadFunc)
					cff_gui_video_frame_thread, &s);
	}

	retv = GPOINTER_TO_INT(result);

	if (cursor && gCursor.window != NULL)
	{
		GuiWindowSetCursor(gCursor.window, NULL);
	}

	if (retv == -1)
	{
		if (slider->is_slider == FALSE || slider->button == 0)
		{
			cff_guitl_beep();
		}
	}

	else
	{

#ifdef HAVE_CAIRO

		if (current_frame->surfaceRGB24 == NULL ||
			current_frame->rectRGB24.width !=
					video->vframe.rgb24.width ||
				current_frame->rectRGB24.height !=
						video->vframe.rgb24.height)
		{
			/* Draw the whole area */
			if (cff_gui_cairo_surface_image_create(current_frame,
								video) == FALSE)
			{
				gIsBusy &= ~kIsBusyVideoFrame;
				return -1;
			}
		}
#else
		if (current_frame->rectRGB24.width !=
					video->vframe.rgb24.width ||
				current_frame->rectRGB24.height !=
						video->vframe.rgb24.height)
		{
			/* Draw the whole area */
			cff_gui_calculate_frame_rects(current_frame, video);
		}

#endif /* ifdef HAVE_CAIRO .. #else */

		else
		{

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

			/* Draw the video frame only */

			GdkWindow *wnd =
				GuiWidgetGetWindow(current_frame->drawingArea);

			gdk_window_invalidate_rect(wnd,
				&(current_frame->rectRGB24), FALSE);
#else
			/*
				Above concept is now obsolete and
				GTK-4 always redraw the entire widgets

			*/

			gtk_widget_queue_draw(current_frame->drawingArea);

#endif /* #if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 .. #else */

		} /* if (current_frame->...) */

#ifdef HAVE_CAIRO
		cff_gui_cairo_surface_image_fill(current_frame);
#endif

		if (slider->is_slider == FALSE)
		{
			cff_gui_slider_adjustment_set_value(slider);
		}

		cff_gui_frame_info(current_frame, video);
		cff_gui_video_button_cut_update_sensitive(video);

	} /* if (retv == -1) .. else */

	gIsBusy &= ~kIsBusyVideoFrame;

	return retv;

} /* static gint cff_gui_video_frame(cff_frame_seek_t*) */

#ifdef HAVE_GMENU
static gboolean cff_gui_video_frame_button(GtkGesture *gesture, gint n_press,
						gdouble x, gdouble y, gpointer data)
#else
static gboolean cff_gui_video_frame_button(GtkWidget *btn, GdkEvent *e, gpointer data)
#endif
{
	guint			button;
	GdkModifierType		state;
	enum cff_VideoFrame	next;
	cff_frame_seek_t	frame_seek;

#ifdef HAVE_GMENU

#if GTK_MAJOR_VERSION == 3
	const GdkEvent		*event;
#else
	GdkEvent		*event;
#endif
	GdkEventSequence	*sequence;

	sequence = gtk_gesture_get_last_updated_sequence(gesture);
	event    = gtk_gesture_get_last_event(gesture, sequence);
	gtk_gesture_set_state(gesture, GTK_EVENT_SEQUENCE_CLAIMED);

#if GTK_MAJOR_VERSION == 3
	gdk_event_get_state(event, &state);
#else
	state = gdk_event_get_modifier_state(event);
#endif

	button = gtk_gesture_single_get_current_button(GTK_GESTURE_SINGLE(gesture));

#else /* ! HAVE_GMENU */

	const GdkEventButton *bevent;

	bevent = (GdkEventButton *) e;
	button = bevent->button;
	state  = bevent->state;

#endif /* #ifdef HAVE_GMENU .. #else */

	if ((state & GDK_CONTROL_MASK) != 0)
	{
		return FALSE;
	}

	switch (GPOINTER_TO_INT(data))
	{
		case kButtonVideoFramePrev4:
			next = kVideoFramePrevCut;
			break;
		case kButtonVideoFramePrev3:
			next = kVideoFramePrev05Sec;
			break;
		case kButtonVideoFramePrev2:
			next = kVideoFramePrevHalfSec;
			break;
		case kButtonVideoFramePrev1:
			next = kVideoFramePrevSingle;
			break;
		case kButtonVideoFrameNext1:
			next = kVideoFrameNextSingle;
			break;
		case kButtonVideoFrameNext2:
			next = kVideoFrameNextHalfSec;
			break;
		case kButtonVideoFrameNext3:
			next = kVideoFrameNext05Sec;
			break;
		case kButtonVideoFrameNext4:
			next = kVideoFrameNextCut;
			break;
		default:
			return FALSE;

	} /* switch (GPOINTER_TO_INT(data)) */

	if ((state & GDK_SHIFT_MASK) != 0 || button == 3)
	{
		/* Shift key or right mouse button */
		next += kVideoFrameShiftAdd;
	}

	frame_seek.type		= kFrameSeekNext;
	frame_seek.value.e_next	= next;
	cff_gui_video_frame(&frame_seek);

	return TRUE;

} /* static gboolean cff_gui_video_frame_button(GtkWidget*, GdkEvent*, gpointer) */

#ifdef HAVE_GMENU
static gboolean cff_gui_video_frame_key(GtkEventControllerKey *cntrl,
			guint keyval, guint keycode, GdkModifierType state,
								gpointer data)
#else
static gboolean cff_gui_video_frame_key(GtkWidget *window, GdkEvent *e,
								gpointer data)
#endif
{
	enum cff_VideoFrame	next;
	cff_frame_seek_t	frame_seek;

#ifndef HAVE_GMENU
	guint			keyval;
	GdkModifierType		state;
	GdkEventKey		*kevent;

	kevent	= (GdkEventKey *) e;
	keyval	= kevent->keyval;
	state	= kevent->state;
#endif

	switch (keyval)
	{
		case Gui_KEY_Home:
		case Gui_KEY_KP_Home:
		case Gui_KEY_KP_7:
			next = kVideoFramePrevCut;
			break;
		case Gui_KEY_Page_Down:
		case Gui_KEY_KP_Page_Down:
		case Gui_KEY_KP_3:
			next = kVideoFramePrev05Sec;
			break;
		case Gui_KEY_Down:
		case Gui_KEY_KP_Down:
		case Gui_KEY_KP_2:
			next = kVideoFramePrevHalfSec;
			break;
		case Gui_KEY_Left:
		case Gui_KEY_KP_Left:
		case Gui_KEY_KP_4:
			next = kVideoFramePrevSingle;
			break;
		case Gui_KEY_Right:
		case Gui_KEY_KP_Right:
		case Gui_KEY_KP_6:
			next = kVideoFrameNextSingle;
			break;
		case Gui_KEY_Up:
		case Gui_KEY_KP_Up:
		case Gui_KEY_KP_8:
			next = kVideoFrameNextHalfSec;
			break;
		case Gui_KEY_Page_Up:
		case Gui_KEY_KP_Page_Up:
		case Gui_KEY_KP_9:
			next = kVideoFrameNext05Sec;
			break;
		case Gui_KEY_End:
		case Gui_KEY_KP_End:
		case Gui_KEY_KP_1:
			next = kVideoFrameNextCut;
			break;
		default:
			return FALSE;

	} /* switch (kevent->keyval) */

	if ((state & GDK_CONTROL_MASK) != 0)
	{
		return FALSE;
	}

	if ((state & GDK_SHIFT_MASK) != 0)
	{
		next += kVideoFrameShiftAdd;
	}

	frame_seek.type		= kFrameSeekNext;
	frame_seek.value.e_next	= next;
	cff_gui_video_frame(&frame_seek);

	return TRUE;

} /* static gboolean cff_gui_video_frame_key(GtkWidget*, GdkEvent*, gpointer) */

static gint cff_gui_video_frame_click(GtkWidget *widget, gdouble x)
{
	gint			width;
	gdouble			next;
	cff_frame_seek_t	frame_seek;

	width = GuiWidgetGetAllocatedWidth(widget);
	next  = x / (gdouble) width;

	frame_seek.type		= kFrameSeekSliderClick;
	frame_seek.value.d_next	= next;

	return cff_gui_video_frame(&frame_seek);

} /* static gint cff_gui_video_frame_click(GtkWidget*, gdouble) */

static gboolean cff_gui_video_frame_slider(GtkRange *range, GtkScrollType scroll, gdouble value, cff_slider_t *slider)
{
	unsigned int value_int;
	cff_video_t  *video;

	if (slider->button != 1 || slider->is_slider ||
				gIsBusy || GuiMainLevelGetBusy())
	{
		return TRUE;
	}

	video		= &gVideo;
	slider->value	= value;
	value_int	= (unsigned int) value;

	if (value_int > video->vframe.frame_total)
	{
		value_int = video->vframe.frame_total;
		value     = (gdouble) value_int;
	}

	else if (value < 0.0)
	{
		value_int = 0;
		value     = 0.0;
	}

	if (scroll == GTK_SCROLL_JUMP)
	{
		cff_frame_seek_t frame_seek;

		if (video->vframe.frame_current == value_int)
		{
			return TRUE;
		}

		frame_seek.type   = kFrameSeekSliderMove;
		slider->is_slider = TRUE;

		for (;;)
		{
			frame_seek.value.i_next	= value_int;
			cff_gui_video_frame(&frame_seek);

			if (value == slider->value)
			{
				if (video->vframe.frame_current !=
					(unsigned int) GuiAdjustmentGetValue(slider->adjust))
				{
					/* Move slider to actual position */
					g_idle_add_full(Gui_PRIORITY_REDRAW,
						(GSourceFunc)
							cff_gui_slider_adjustment_set_value,
							slider, NULL);
				}

				slider->is_slider = FALSE;
				return TRUE;
			}

			if (slider->button)
			{
				break;
			}

			value = slider->value;

		} /* for (;;) */

		slider->is_slider = FALSE;
		return FALSE;

	} /* if (scroll == GTK_SCROLL_JUMP) */

	if (slider->button == 1)
	{
		slider->button = 2;
		cff_gui_video_frame_click(slider->scale, slider->x);
	}

	return TRUE;

} /* static gboolean cff_gui_video_frame_slider(GtkRange*, GtkScrollType, gdouble, cff_slider_t*) */


#if GTK_MAJOR_VERSION == 4

static gboolean cff_gui_slider_timeout(cff_slider_t *slider)
{
	if (slider->button != 0)
	{
		/* Emulates "button-release" */

		GtkStateFlags flags =
			gtk_widget_get_state_flags(slider->scale);

		if (flags & GTK_STATE_FLAG_ACTIVE)
		{
			return G_SOURCE_CONTINUE;

		}

		slider->button = 0;

	} /* if (slider->button != 0) */

	return G_SOURCE_REMOVE;

} /* static gboolean cff_gui_slider_timeout(cff_slider_t*) */

#endif /* GTK_MAJOR_VERSION == 4 */


#ifdef HAVE_GMENU
static gboolean cff_gui_slider_button(GtkGesture *gesture, gint n_press,
				gdouble x, gdouble y, cff_slider_t *slider)
#else
static gboolean cff_gui_slider_button(GtkWidget *widget, GdkEvent *e, cff_slider_t *slider)
#endif
{
	GdkEventType type;

#ifdef HAVE_GMENU

	if (gtk_gesture_is_active(gesture) == FALSE)
	{
		type = GDK_BUTTON_RELEASE;
	}

	else
	{
		type = (gtk_gesture_single_get_current_button(
			GTK_GESTURE_SINGLE(gesture)) == 0) ?
				GDK_BUTTON_RELEASE : GDK_BUTTON_PRESS;
	}
#else
	gdouble		x;
	GdkEventButton	*bevent;

	bevent	= (GdkEventButton *) e;
	type	= bevent->type;
	x	= bevent->x;

#endif /* #ifdef HAVE_GMENU .. #else */

	if (type == GDK_BUTTON_RELEASE)
	{
		slider->button = 0;
	}

	else if (slider->button == 0)
	{
		slider->x	= x;
		slider->button	= 1;

#if GTK_MAJOR_VERSION == 4
		/* GTK-4: 'release' signal not emitted if widget is GtkScale */
		g_timeout_add(100,
			(GSourceFunc) cff_gui_slider_timeout, slider);
#endif
	}

	return FALSE;

} /* static gboolean cff_gui_slider_button(GtkWidget*, GdkEvent*, cff_slider_t*) */


#if GTK_MAJOR_VERSION == 2
static gint cff_gui_draw_bar(GtkWidget *widget, GdkEvent *e, cff_video_t *video)
#elif GTK_MAJOR_VERSION == 3
static gint cff_gui_draw_bar(GtkWidget *widget, cairo_t *cr, cff_video_t *video)
#else /* GTK_MAJOR_VERSION == 4 */
static gint cff_gui_draw_bar(GtkWidget *widget, cairo_t *cr, gint width,
						gint height, cff_video_t *video)
#endif
{
	gint		x1, color;
	GdkRectangle	rect;

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

	gint		width, height;

#if defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2
	/* Creates a Cairo context for drawing to 'drawable' */
	cairo_t *cr = gdk_cairo_create(widget->window);
#endif

	width	= GuiWidgetGetAllocatedWidth(widget);
	height	= GuiWidgetGetAllocatedHeight(widget);

#endif /* GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 */

	x1	= 0;
	color	= kColorGrey;

	rect.y      = 0;
	rect.height = height;

	if (video->vframe.frame_total != 0 && video->vframe.stream_duration > 0)
	{
		gint		i;
		cff_dts_t	*dts;

		color	= kColorGreen;
		dts	= video->cut.list;
		i	= video->cut.num;

		if (i > 0 && dts != NULL)
		{
			double ratio =
				(double) width / (double) video->vframe.stream_duration;

			color = kColorRed;

			do
			{
				gint x2 = (double)
					(dts->frame - video->vframe.dts_start) * ratio;

				if (x2 > x1)
				{
					rect.x     = x1;
					rect.width = x2 - x1;

					GuiDrawRectangle(cr, e,
							color, TRUE, &rect);
					x1 = x2;
				}

				if (color == kColorGreen)
				{
					color = kColorRed;
				}

				else
				{
					color = kColorGreen;
				}

				dts++;
			}
			while (--i > 0);

		} /* if (i > 0 && ptr != NULL) */

	} /* if (video->vframe.frame_total != 0 && video->vframe.duration > 0) */


	/* Remaining rectangle */

	rect.x     = x1;
	rect.width = width - x1;

	GuiDrawRectangle(cr, e, color, TRUE, &rect);

	/* Border */

	rect.x     = 0;
	rect.width = width - 1;
	rect.height--;

	GuiDrawRectangle(cr, e, kColorBlack, FALSE, &rect);

#if defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2
	cairo_destroy(cr);
#endif

	return TRUE;

} /* static gint cff_gui_draw_bar(GtkWidget*, GdkEvent*, cff_video_t*) */

#ifdef HAVE_GMENU
static gint cff_gui_draw_bar_button(GtkGesture *gesture,
			gint n_press, gdouble x, gdouble y, cff_video_t *video)
#else
static gboolean cff_gui_draw_bar_button(GtkWidget *widget, GdkEvent *e,
							cff_video_t *video)
#endif
{
	if (video->vframe.frame_total != 0)
	{

#ifdef HAVE_GMENU
		GtkWidget *widget;

		gtk_gesture_set_state(gesture, GTK_EVENT_SEQUENCE_CLAIMED);
		widget = gtk_event_controller_get_widget(GTK_EVENT_CONTROLLER(gesture));
#else
		gdouble		x;
		GdkEventButton	*bevent;

		bevent = (GdkEventButton *) e;

		if (bevent->type != GDK_BUTTON_PRESS || bevent->button != 1)
		{
			return TRUE;
		}

		x = bevent->x;

#endif /* #ifdef HAVE_GMENU .. #else */

		cff_gui_video_frame_click(widget, x);

	} /* if (video->vframe.frame_total != 0) */

	return TRUE;

} /* static gboolean cff_gui_draw_bar_button(GtkWidget*, GdkEvent*, cff_video_t*) */

static void cff_gui_open_result(cff_video_t *video, const gchar *str, guint mbar_item)
{
	gint			sensitive;
	cff_video_frame_t	*current_frame;

	current_frame = &gCurrentVideoFrame;

	if (str == NULL)
	{
		cff_frame_seek_t frame_seek;

#ifdef HAVE_CAIRO

		if (current_frame->surfaceRGB24 != NULL)
		{
			/* Create a new surface */
			cairo_surface_destroy(current_frame->surfaceRGB24);
			current_frame->surfaceRGB24 = NULL;
		}
#endif
		cff_gui_slider_adjustment_configure(video->vframe.frame_total);

		frame_seek.type		= kFrameSeekNext;
		frame_seek.value.e_next	= kVideoFrameStart;
		cff_gui_video_frame(&frame_seek);
		sensitive = TRUE;
	}

	else
	{
		cff_gui_slider_adjustment_configure(0);
		GuiWidgetQueueDrawArea(current_frame->drawingArea);
		sensitive = FALSE;
	}

	cff_gui_video_button_set_sensitive(sensitive);

	cff_gui_frame_info(current_frame, video);
	GuiWidgetQueueDrawArea(gDrawingAreaBar);

	if (str != NULL)
	{
		cff_guitl_beep();
		cff_guitl_dialog_error(str, &gMenuBarTitle[mbar_item]);
	}

} /* static void cff_gui_open_result(cff_video_t*, const gchar*, guint) */

static int cff_gui_open_callback(gpointer data)
{
	const gchar		**str;
	cff_progressbar_t	user_data;

	user_data.func  = cff_ffmpeg_file_open_cb;
	user_data.video = &gVideo;
	user_data.title = gettext_nop("Scanning ...");
	user_data.mbar  = &gMenuBarTitle[kMenuBarFile];

	if (gLogWindow.nThreads == 0)
	{
		user_data.mutex = NULL;
		cff_gui_dialog_progress_bar(&user_data);
	}

	else
	{
		GuiMutex mutex;

		user_data.mutex = &mutex;
		GuiMutexInit(&mutex);

		user_data.is_locked = FALSE;

		g_main_context_invoke_full(NULL, G_PRIORITY_HIGH_IDLE,
			(GSourceFunc) cff_gui_dialog_progress_bar,
							&user_data, NULL);

		while (user_data.is_locked == FALSE)
		{
			g_usleep(250);
		}

		GuiMutexLock(&mutex);
		GuiMutexUnlock(&mutex);
		GuiMutexFree(&mutex);

	} /* if (gLogWindow.nThreads == 0) .. else */

	str = (const gchar **) data;

	if (str != NULL)
	{
		*str = user_data.error;
	}

	return user_data.retv;

} /* static int cff_gui_open_callback(gpointer) */

static gint cff_gui_program_open(gpointer data)
{
	const gchar	*error, *str;
	cff_video_t	*video;

	video = &gVideo;

	gIsBusy |= kIsBusyFile;

	str = cff_ffmpeg_program_open(video, GPOINTER_TO_INT(data),
					cff_gui_open_callback, &error);

	cff_gui_open_result(video, (str == NULL) ? error : str, kMenuBarEdit);

	gIsBusy &= ~kIsBusyFile;

	return Gui_SOURCE_REMOVE;

} /* static gint cff_gui_program_open(gpointer) */

static const gchar *cff_gui_video_open_cb(gpointer data)
{
	struct
	{
		cff_video_t	*video;
		const gchar	*error;
		cff_fname_t	fname;
		gchar		*hwaccel;

	} *s = data;

	return cff_ffmpeg_file_open(s->video, &(s->fname),
			s->hwaccel, cff_gui_open_callback, &(s->error));

} /* static const gchar *cff_gui_video_open_cb(gpointer) */

static gint cff_gui_video_open(GFile *file)
{
	const gchar *str;

	struct
	{
		cff_video_t	*video;
		const gchar	*error;
		cff_fname_t	fname;
		gchar		*hwaccel;

	} s;

	if (file == NULL)
	{
		return Gui_SOURCE_REMOVE;
	}

	s.fname.local   = g_file_get_path(file);
	s.fname.utf8    = g_file_get_parse_name(file);
	s.fname.strfree = (func_strfree_t *) g_free;
	g_object_unref(G_OBJECT(file));

	if (strcmp(s.fname.local, s.fname.utf8) == 0)
	{
		g_free(s.fname.utf8);
		s.fname.utf8 = s.fname.local;
	}

	s.hwaccel = cff_gprop_get_str(kPropStrFileHWAccel);

	s.video  = &gVideo;
	gIsBusy |= kIsBusyFile;

	str = (const gchar *) cff_gui_thread_invoke((GThreadFunc)
						cff_gui_video_open_cb, &s);
	g_free(s.hwaccel);

	cff_gui_open_result(s.video, (str == NULL) ?
					s.error : str, kMenuBarFile);

	if ((s.video)->inFormatCtx != NULL)
	{
		cff_guitl_window_set_title_filename(gWindata.mainwin, s.fname.utf8);
	}

	gIsBusy &= ~kIsBusyFile;

	return Gui_SOURCE_REMOVE;

} /* static gint cff_gui_video_open(GFile*) */


/*
	-------------
	 Menu 'File'
	-------------
*/

#ifdef HAVE_GMENU
static void cff_gui_file_open(GSimpleAction *a, GVariant *p, gpointer *data)
#else
static void cff_gui_file_open(GtkWidget *w, gpointer *data)
#endif
{
	GFile		*file;
	GtkWidget	**mbar;

	mbar = (GtkWidget **) data;

	if (cff_gui_file_discard_changes(mbar) == FALSE)
	{
		return;
	}

	file = cff_gui_file_selection_new_gettext(gettext_nop("Open Video File"),
					NULL, NULL, mbar, kPropStrFileOpenPath);
	if (file == NULL)
	{
		return;
	}

#ifdef HAVE_GMENU
	cff_gui_file_close(NULL, NULL, NULL);
#else
	cff_gui_file_close(NULL, NULL);
#endif

	g_idle_add_full(G_PRIORITY_LOW,
		(GSourceFunc) cff_gui_video_open, file, NULL);

} /* static void cff_gui_file_open(GtkWidget*, gpointer*) */

#ifdef HAVE_GMENU
static void cff_gui_file_save(GSimpleAction *a, GVariant *p, gpointer *data)
#else
static void cff_gui_file_save(GtkWidget *w, gpointer *data)
#endif
{
	static const gchar softcut_0[] =
		gettext_nop("The selected hwaccel decoder disables softcut!");
	static const gchar softcut_1[] =
		gettext_nop("Folling encoders are available:");
	static const gchar softcut_2[] =
		gettext_nop("Open the file with another decoder!");

	gint				i, n_encoders, pos_old, *newpos_ptr, newpos_buf[10];
	gboolean			got_video, got_audio;
	GtkResponseType			response;
	gchar				*str, **ptr;
	GtkWidget			*notebook, *bbox, *hbox, *vbox, *frame, **mbar;
	cff_fmtlist_t			*format_info;
	cff_video_t			*video;
	cff_save_video_prev_options_t	*save_prev;
	gchar				*buf[512];
	cff_ListStore_t			*list;
	GtkTreeIter			iter;
	cff_save_video_t		vsave;

	video = &gVideo;

	if (video->fname.utf8 == NULL)
	{
		return;
	}

	vsave.common.video = video;

	mbar = (GtkWidget **) data;

	if (mbar != NULL && *mbar != NULL)
	{
		GuiWidgetSetStateSelected(*mbar);
	}

	vbox = cff_gui_save_build_common(&(vsave.common),
					gettext_nop("Save Video File"),
					kPropStrFileSavePath,
					kPropIntFileSaveWinWidth);
	hbox = GuiHBoxNew(FALSE, 4);
	GuiContainerSetBorderWidth(hbox, 4);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, 2, hbox);
	GuiVBoxPackStart(vbox, frame, TRUE, TRUE, 4);

	save_prev = gSaveVideoPrevOptionsPtr;

	if (save_prev == NULL)
	{
		save_prev = g_malloc0(sizeof(cff_save_video_prev_options_t));
		gSaveVideoPrevOptionsPtr = save_prev;
	}


	/* List */

	cff_gui_build_list(gettext_nop("Available"), &(vsave.avail.list), hbox);


	/* Button */

	bbox = GuiVBoxNew(FALSE, 4);
	GuiContainerSetBorderWidth(bbox, 4);

	vsave.up =
		cff_guitl_vbox_add_button_gettext(bbox, gettext_nop("Up"),
				(GCallback) cff_gui_save_list_updown, &vsave);
	vsave.avail.all =
		cff_guitl_vbox_add_button(bbox, ">>",
				(GCallback) cff_gui_save_list_leftright, &vsave);
	vsave.avail.some =
		cff_guitl_vbox_add_button(bbox, ">",
				(GCallback) cff_gui_save_list_leftright, &vsave);
	vsave.out.some =
		cff_guitl_vbox_add_button(bbox, "<",
				(GCallback) cff_gui_save_list_leftright, &vsave);
	vsave.out.all =
		cff_guitl_vbox_add_button(bbox, "<<",
				(GCallback) cff_gui_save_list_leftright, &vsave);
	vsave.down =
		cff_guitl_vbox_add_button_gettext(bbox, gettext_nop("Down"),
				(GCallback) cff_gui_save_list_updown, &vsave);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, 0, bbox);

	bbox = GuiVBoxNew(FALSE, 1);
	GuiVBoxPackStart(bbox, frame, FALSE, FALSE, 1);
	GuiHBoxPackStart(hbox, bbox, FALSE, FALSE, 4);


	/* List */

	cff_gui_build_list(gettext_nop("Output"), &(vsave.out.list), hbox);


	/* Check button */

	hbox = GuiHBoxNew(TRUE, 4);
	GuiContainerSetBorderWidth(hbox, 4);
	GuiVBoxPackStart(vbox, hbox, FALSE, FALSE, 1);

	bbox = GuiVBoxNew(FALSE, 1);
	GuiHBoxPackStart(hbox, bbox, TRUE, FALSE, 1);

	vsave.shortest =
		cff_guitl_check_button_new_gettext(gettext_nop("Shortest duration"));
	GuiVBoxPackStart(bbox, vsave.shortest, FALSE, FALSE, 8);

	vsave.stream_sync =
		cff_guitl_check_button_new_gettext(gettext_nop("Stream sync"));
	GuiVBoxPackStart(bbox, vsave.stream_sync, FALSE, FALSE, 8);


	/* Softcut Option Menu */

	n_encoders = cff_ffmpeg_get_encoder_names(video, &(vsave.softcut.encoder_names));

	/* Suppress compiler warning */
	i = 0;

	if (n_encoders == 0)
	{
		/* No encoder found */
		vsave.softcut.encoder_names = NULL;
	}

	else if (n_encoders < 0)
	{
		/* Tooltip shows available (but disabled) encoders */

		buf[0] = cff_charconv_utf8_gettext(softcut_0);
		buf[1] = cff_charconv_utf8_gettext(softcut_1);

		ptr = vsave.softcut.encoder_names;

		i = 2;

		while ((str = *(ptr++)) != NULL)
		{
			buf[i] = g_strconcat("   ", str, NULL);
			i++;
		}

		buf[i] = cff_charconv_utf8_gettext(softcut_2);
		i++;
		buf[i] = NULL;
		
		/* malloc'ed */
		free(vsave.softcut.encoder_names);
		vsave.softcut.encoder_names = NULL;

	} /* if (n_encoders == 0) .. else if (n_encoders < 0) */


	bbox = GuiVBoxNew(FALSE, 1);

	vsave.softcut.option =
			cff_gui_build_option_menu(hbox, bbox,
			gettext_nop("Soft cut"),
			gettext_nop("No"),
			(const gchar **) vsave.softcut.encoder_names);

	vsave.softcut.both_gops = cff_guitl_vbox_add_check_button_gettext(bbox,
					gettext_nop("Both GOPs"));

	gtk_widget_set_sensitive(vsave.softcut.both_gops, FALSE);

	if (vsave.softcut.encoder_names != NULL)
	{
		GuiSignalConnect(vsave.softcut.option, "changed",
				cff_gui_save_softcut_option_changed_cb,
						vsave.softcut.both_gops);
	}

	else if (n_encoders < 0)
	{
		/* Tooltip shows available (but disabled) encoders */

		str = g_strjoinv("\n", buf);

		i--;
		CFF_CHARCONV_UTF8_FREE(buf[i], softcut_2);

		while (--i > 1)
		{
			g_free(buf[i]);
		}

		CFF_CHARCONV_UTF8_FREE(buf[1], softcut_1);
		CFF_CHARCONV_UTF8_FREE(buf[0], softcut_0);

		gtk_widget_set_tooltip_text(vsave.softcut.option, str);
		g_free(str);

	} /* if (vsave.softcut.encoder_names != NULL) .. else if (n_encoders < 0) */	


	/* Extra */

	notebook = gtk_notebook_new();
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), GTK_POS_TOP);
	GuiContainerSetBorderWidth(notebook, 4);

	frame = cff_guitl_frame_new_add_child_gettext(gettext_nop("Extra"),
								0, notebook);
	GuiVBoxPackStart(vbox, frame, FALSE, FALSE, 2);


	/* General-Tab */

	bbox = cff_gui_notebook_append_page(notebook, gettext_nop("General"));


	/* Check button */

	vsave.ts_keep_start = cff_guitl_vbox_add_check_button_gettext(bbox,
					gettext_nop("Keep start timestamps"));

	/* Option menu */

	vsave.common.format.u.list =
			cff_ffmpeg_get_format_names(vsave.common.video);

	ptr = buf;

	if ((format_info = vsave.common.format.u.list) != NULL)
	{
		while (format_info->name != NULL)
		{
			if (ptr - buf < (sizeof(buf) / sizeof(gchar*)) - 1)
			{
				*(ptr++) = g_strconcat(format_info->name, " ",
							format_info->codecs, NULL);
			}

			if (format_info->codecs != NULL)
			{
				/* malloc'ed */
				free(format_info->codecs);
				format_info->codecs = NULL;
			}

			format_info++;
		}

	} /* if ((format_info = vsave.common.format.u.list) != NULL) */

	*ptr = NULL;

	vsave.common.format.option =
			cff_gui_build_option_menu(bbox, NULL,
				gettext_nop("Format"),
				gettext_nop("Auto"),
				(const gchar **) buf);

	if (vsave.common.format.u.list != NULL)
	{
		GuiSignalConnect(vsave.common.format.option, "changed",
			cff_gui_save_optionmenu_callback, &(vsave.common));
	}


	/* Free strings in array */

	ptr = buf;

	while ((str = *(ptr++)) != NULL)
	{
		g_free(str);
	}


	/* Media-Tabs */

	cff_gui_build_media_option(notebook, &vsave, kCFFMediaVideo);
	cff_gui_build_media_option(notebook, &vsave, kCFFMediaAudio);
	cff_gui_build_media_option(notebook, &vsave, kCFFMediaOther);


	/* Last used streams (if any) */

	newpos_ptr = NULL;
	pos_old    = 0;

	if (save_prev->stream_idx != NULL && save_prev->nb_streams > 0)
	{
		newpos_ptr = newpos_buf;

		if (save_prev->nb_streams > sizeof(newpos_buf) / sizeof(gint))
		{
			newpos_ptr = g_malloc(sizeof(gint) * save_prev->nb_streams);
		}
	}


	/* Put streams in the available or output list */

	got_video = FALSE;
	got_audio = FALSE;

	i = 0;

	do
	{
		gint	media, stream_idx;
		gchar	*str, buf[256];

		if (cff_ffmpeg_get_stream_info(video, i,
					&stream_idx, buf, sizeof(buf)) < 0)
		{
			continue;
		}

		media	= kCFFMediaOther;
		list	= &(vsave.avail.list);
		str	= strchr(buf, ' ');

		if (str != NULL)
		{
			if (g_ascii_strncasecmp(&str[1], "video", 5) == 0)
			{
				media = kCFFMediaVideo;
			}

			else if (g_ascii_strncasecmp(&str[1], "audio", 5) == 0)
			{
				media = kCFFMediaAudio;
			}

			if (media == kCFFMediaVideo)
			{
				if (got_video == FALSE)
				{
					got_video = TRUE;
					list = &(vsave.out.list);
				}
			}

			else if (media == kCFFMediaAudio)
			{
				if (got_audio == FALSE)
				{
					got_audio = TRUE;
					list = &(vsave.out.list);
				}
			}

		} /* if (str != NULL) */

		if (newpos_ptr != NULL)
		{
			gint pos_new;

			list	= &(vsave.avail.list);
			pos_new	= 0;

			do
			{
				if (stream_idx == save_prev->stream_idx[pos_new])
				{
					/*
						Mapping the new position of
						each child to its old position
						before the re-ordering
					*/

					newpos_ptr[pos_new] = pos_old++;
					list = &(vsave.out.list);

					break;
				}
			}
			while (++pos_new < save_prev->nb_streams);

		} /* if (newpos_ptr != NULL) */

		gtk_list_store_append(list->store, &iter);

		str = cff_tl_replace_non_utf8_chars_in_str(buf);
		gtk_list_store_set(list->store, &iter,
				0, str, 1, stream_idx, 2, media, -1);
	}
	while(++i < video->nb_streams);


	/* Set last used options */

	if (newpos_ptr != NULL)
	{
		gint lastpos;

		list	= &(vsave.out.list);
		i	= 0;
		lastpos	= 0;

		do
		{
			gint pos =  newpos_ptr[i];

			if (lastpos > pos)
			{
				gtk_list_store_reorder(list->store, newpos_ptr);
				break;
			}

			lastpos = pos;
		}
		while(++i < save_prev->nb_streams);

		if (newpos_ptr != newpos_buf)
		{
			g_free(newpos_ptr);
		}

	} /* if (newpos_ptr != NULL) */


	if (save_prev->format_idx)
	{
		GuiOptionMenuSetActive(vsave.common.format.option, save_prev->format_idx);
	}

	if (vsave.softcut.encoder_names != NULL)
	{
		if (save_prev->encoder_idx)
		{
			GuiOptionMenuSetActive(vsave.softcut.option, save_prev->encoder_idx);

			if (save_prev->both_gops)
			{
				GuiCheckButtonSetActive(vsave.softcut.both_gops, TRUE);
			}
		}

	} /* if (vsave.softcut.encoder_names != NULL) */

	if (save_prev->shortest)
	{
		GuiCheckButtonSetActive(vsave.shortest, TRUE);
	}

	if (save_prev->stream_sync &&
		cff_ffmpeg_video_cut_list_distance(vsave.common.video) != 0)
	{
		GuiCheckButtonSetActive(vsave.stream_sync, TRUE);
	}

	if (save_prev->ts_keep_start)
	{
		GuiCheckButtonSetActive(vsave.ts_keep_start, TRUE);
	}

	i = kCFFMediaVideo;

	do
	{
		if (save_prev->media[i].bsfname_idx)
		{
			GuiOptionMenuSetActive(vsave.media[i].bsf_option,
						save_prev->media[i].bsfname_idx);
		}

		if (save_prev->media[i].ts_from_duration)
		{
			GuiCheckButtonSetActive(vsave.media[i].ts_from_duration, TRUE);
		}

		if (save_prev->media[i].ts_adjust)
		{
			GuiCheckButtonSetActive(vsave.media[i].ts_adjust, TRUE);
		}
	}
	while (++i < kCFFMediaTotal);


	/* Select the first row, set signal func, set buttons (in)sensitive */

	cff_gui_save_list_selection_init(&(vsave.avail.list), &vsave);
	cff_gui_save_list_selection_init(&(vsave.out.list), &vsave);
	GuiWidgetShowAll(vbox);

	response = GuiDialogRun(vsave.common.dialog);

	if (response != GTK_RESPONSE_ACCEPT ||
				cff_gui_save_video(&vsave) == FALSE)
	{
		gint n;

		GuiDialogClose(vsave.common.dialog, response);

		if (vsave.common.format.u.list != NULL)
		{
			/* malloc'ed */
			free(vsave.common.format.u.list);
		}

		n = kCFFMediaVideo;

		do
		{
			cff_mediaopts_t *mediaopts = &(vsave.media[n]);

			if (mediaopts->bsf_list != NULL)
			{
				/* malloc'ed */
				free(mediaopts->bsf_list);
			}
		}
		while (++n < kCFFMediaTotal);

		if (vsave.softcut.encoder_names != NULL)
		{
			/* malloc'ed */
			free(vsave.softcut.encoder_names);
		}

		g_free(vsave.common.format.extension_orig);
		g_free(vsave.common.filechooser_utf8str);

	} /* if (response != GTK_RESPONSE_ACCEPT ||
				cff_gui_save_video(vsave) == FALSE) */

	if (mbar != NULL && *mbar != NULL)
	{
		GuiWidgetUnsetStateSelected(*mbar);
	}

} /* static void cff_gui_file_save(GtkWidget*, gpointer*) */

#ifdef HAVE_GMENU
static void cff_gui_file_image(GSimpleAction *a, GVariant *p, gpointer *data)
#else
static void cff_gui_file_image(GtkWidget *w, gpointer *data)
#endif
{
	gint			fmt_num;
	GtkResponseType		response;
	gchar			**long_names;
	GtkWidget		*hbox, *vbox, *bbox, *fbox, *frame, **mbar;
	cff_video_t		*video;
	cff_save_image_t	vimage;

	video = &gVideo;

	if (video->fname.utf8 == NULL)
	{
		return;
	}

	vimage.common.video = video;

	mbar = (GtkWidget **) data;

	if (mbar != NULL && *mbar != NULL)
	{
		GuiWidgetSetStateSelected(*mbar);
	}

	vbox = cff_gui_save_build_common(&(vimage.common),
					gettext_nop("Save Image"),
					kPropStrFileImagePath,
					kPropIntFileImageWinWidth);
	fbox = GuiVBoxNew(FALSE, 4);
	GuiContainerSetBorderWidth(fbox, 4);
	GuiVBoxPackStart(vbox, fbox, FALSE, FALSE, 4);

	/* Spin buttons */

	hbox = GuiHBoxNew(FALSE, 4);
	GuiContainerSetBorderWidth(hbox, 4);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, 2, hbox);
	GuiVBoxPackStart(fbox, frame, FALSE, FALSE, 4);

	bbox = GuiHBoxNew(FALSE, 4);
	GuiContainerSetBorderWidth(bbox, 4);
	GuiHBoxPackStart(hbox, bbox, TRUE, FALSE, 4);

	vimage.width.prev_value = video->vframe.frame_width;
	vimage.width.widget = cff_gui_image_spinbutton(bbox,
			video->vframe.frame_width, gettext_nop("Width"));

	vimage.height.prev_value = video->vframe.frame_height;
	vimage.height.widget = cff_gui_image_spinbutton(bbox,
			video->vframe.frame_height, gettext_nop("Height"));

	/* Check button */

	hbox = GuiHBoxNew(FALSE, 4);
	GuiContainerSetBorderWidth(hbox, 4);

	frame = cff_guitl_frame_new_add_child_gettext(NULL, 2, hbox);
	GuiVBoxPackStart(fbox, frame, FALSE, FALSE, 4);

	bbox = GuiHBoxNew(FALSE, 4);
	GuiHBoxPackStart(hbox, bbox, FALSE, FALSE, 4);

	hbox = bbox;
	bbox = GuiVBoxNew(FALSE, 4);
	GuiHBoxPackStart(hbox, bbox, FALSE, FALSE, 4);

	GuiContainerSetBorderWidth(bbox, 4);

	vimage.keep_dimension =
		cff_guitl_check_button_new_gettext(gettext_nop("Keep dimension"));
	GuiVBoxPackStart(bbox, vimage.keep_dimension, FALSE, FALSE, 4);
	GuiSignalConnect(vimage.keep_dimension,
		"toggled", cff_gui_image_keep_dimension_toggled, &vimage);

	vimage.keep_ratio =
		cff_guitl_check_button_new_gettext(gettext_nop("Keep aspect ratio"));
	GuiVBoxPackStart(bbox, vimage.keep_ratio, FALSE, FALSE, 4);

	vimage.gray = cff_guitl_check_button_new_gettext(gettext_nop("Gray"));
	GuiVBoxPackStart(bbox, vimage.gray, FALSE, FALSE, 4);
				
	/* Option menu */

	fmt_num = cff_ffmpeg_get_image_format_names(&long_names,
					&(vimage.common.format.u.names));

	vimage.common.format.option =
			cff_gui_build_option_menu(vbox, NULL,
				gettext_nop("Format"),
				gettext_nop("Auto"),
				(const gchar **) long_names);
	if (fmt_num > 0)
	{
		if (long_names != NULL)
		{
			/* malloc'ed */
			free(long_names);
		}

		GuiSignalConnect(vimage.common.format.option, "changed",
			cff_gui_save_optionmenu_callback, &(vimage.common));
	}

	GuiCheckButtonSetActive(vimage.keep_ratio, TRUE);
	GuiWidgetShowAll(vbox);

	response = GuiDialogRun(vimage.common.dialog);

	if (response != GTK_RESPONSE_ACCEPT ||
				cff_gui_save_image(&vimage) == FALSE)
	{
		GuiDialogClose(vimage.common.dialog, response);

		if (vimage.common.format.u.names != NULL)
		{
			/* malloc'ed */
			free(vimage.common.format.u.names);
		}

		g_free(vimage.common.format.extension_orig);
		g_free(vimage.common.filechooser_utf8str);

	} /* if (if (response != GTK_RESPONSE_ACCEPT ||
				cff_gui_save_image(vimage) == FALSE) */

	if (mbar != NULL && *mbar != NULL)
	{
		GuiWidgetUnsetStateSelected(*mbar);
	}

} /* static void cff_gui_file_image(GtkWidget*, gpointer*) */

#ifdef HAVE_GMENU
static void cff_gui_file_close(GSimpleAction *a, GVariant *p, gpointer *data)
#else
static void cff_gui_file_close(GtkWidget *w, gpointer *data)
#endif
{
	/* mbar == NULL => from 'cff_gui_file_open' */

	GtkWidget **mbar = (GtkWidget **) data;

	if (mbar != NULL && cff_gui_file_discard_changes(mbar) == FALSE)
	{
		return;
	}

	cff_ffmpeg_file_close(&gVideo);

	cff_gui_save_video_prev_settings_free();
	cff_gui_frame_info(&gCurrentVideoFrame, &gVideo);
	cff_gui_slider_adjustment_configure(0);
	cff_gui_video_button_set_sensitive(FALSE);

	GuiWidgetQueueDrawArea(gCurrentVideoFrame.drawingArea);
	GuiWidgetQueueDrawArea(gDrawingAreaBar);

	cff_guitl_window_set_title_filename(gWindata.mainwin, NULL);

} /* static void cff_gui_file_close(GtkWidget*, gpointer*) */

#ifdef HAVE_GMENU
static void cff_gui_file_info(GSimpleAction *a, GVariant *p, gpointer *data)
#else
static void cff_gui_file_info(GtkWidget *w, gpointer *data)
#endif
{
	char		*str;
	GtkWidget	*vbox, **mbar;

	vbox = GuiVBoxNew(FALSE, 2);
	str  = cff_ffmpeg_file_info(&gVideo);

	cff_gui_ffmpeg_info_label(NULL, str, vbox);

	if (str != NULL)
	{
		/* malloc'ed */
		free(str);
	}

	mbar = (GtkWidget **) data;
	cff_gui_scrolled_info_dialog(gettext_nop("Video File Info"),
				vbox, mbar, kPropIntFileInfoWinWidth);

} /* static void cff_gui_file_info(GtkWidget*, gpointer*) */


#ifdef HAVE_GMENU
static void cff_gui_file_save_rc(GSimpleAction *a, GVariant *p, gpointer d)
#else
static void cff_gui_file_save_rc(GtkWidget *w, gpointer d)
#endif
{

#if GTK_MAJOR_VERSION == 4

	/*
		GTK-4 store the dimension only if the window
		should be closed. If the main- and/or logwindow
		is/are open, store the dimension(s).
	*/

	cff_guitl_widget_size_to_prop(gWindata.mainwin,
						kPropIntMainWinWidth);
	cff_guitl_widget_size_to_prop(gLogWindow.window,
						kPropIntLogWinWidth);

#endif /* GTK_MAJOR_VERSION == 4 */

	cff_main_save_rc();

} /* static void cff_gui_file_save_rc(GtkWidget*, gpointer) */

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
static gboolean cff_gui_file_quit_delevnt(GtkWidget *w, GdkEvent *e, gpointer *data)
#else
static gboolean cff_gui_file_quit_delevnt(GtkWidget *w, gpointer *data)
#endif
{
	GtkWidget	**mbar;
	cff_video_t	*video;
	cff_logwindow_t	*log;
	cff_thread_t	*thread;

	if (w != gWindata.mainwin)
	{
		/* Already closed by 'cff_gui_file_quit' */
		return FALSE;
	}

	if (gIsBusy || GuiMainLevelGetBusy())
	{
		return TRUE;
	}

	mbar = (GtkWidget **) data;

	if (cff_gui_file_discard_changes(mbar) == FALSE)
	{
		return TRUE;
	}

	video = &gVideo;

	/* 'cff_main_save_rc' called from 'cff_main_atexit' */

	log = &gLogWindow;

	if (log->window != NULL)
	{
		int		save_flag;
		gchar		*logname;
		GtkWidget	*window;

		if (cff_gprop_get_bool(kPropBoolOptSaveConfig,
					&save_flag) == FALSE || save_flag)
		{
			/* Get log-level, will be set to 'quiet' */
			logname = cff_gprop_get_str(kPropStrOptLogLevel);
		}

		else
		{
			logname = NULL;
		}

		window = log->window;

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
		cff_gui_logwindow_free(NULL, NULL, log);
		gtk_widget_destroy(window);
#else
		/*
			GTK-4: 'cff_gui_logwindow_free' will
			be called from 'gtk_window_close'
		*/
		gtk_window_close(GTK_WINDOW(window));
#endif
		if (logname != NULL)
		{
			/* Restore log-level */
			cff_gprop_set_str(kPropStrOptLogLevel, logname);
			g_free(logname);
		}

	} /* if (log->window != NULL) */

	cff_ffmpeg_file_close(video);

	thread = &gThread;

	if (thread->thread != NULL)
	{
		/* Terminate */
		thread->func = NULL;
		GuiMutexUnlock(&(thread->mutex));
		g_thread_join(thread->thread);
		thread->thread = NULL;
	}

#ifdef HAVE_CAIRO
	if (gCurrentVideoFrame.surfaceRGB24 != NULL)
	{
		cairo_surface_destroy(gCurrentVideoFrame.surfaceRGB24);
		gCurrentVideoFrame.surfaceRGB24 = NULL;
	}
#endif

	return FALSE;

} /* static gboolean cff_gui_file_quit_delevnt(GtkWidget*, GdkEvent*, gpointer *) */

#ifdef HAVE_GMENU
static void cff_gui_file_quit(GSimpleAction *a, GVariant *p, gpointer *mbar)
#else
static void cff_gui_file_quit(GtkWidget *w, gpointer *mbar)
#endif
{
	GtkWidget *window = gWindata.mainwin;

	if (window != NULL)
	{

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
		if (cff_gui_file_quit_delevnt(window, NULL, mbar) == FALSE)
#else
		if (cff_gui_file_quit_delevnt(window, mbar) == FALSE)
#endif
		{
			GuiWindowClose(window);
			gWindata.mainwin = NULL;
		}

	} /* if (window != NULL) */

} /* static void cff_gui_file_quit(GtkWidget*, gpointer*) */


/*
	-------------
	 Menu 'Edit'
	-------------
*/

static void cff_gui_edit_undo_redo(cff_video_t *video, int action)
{
	int retv;

	if (action == 0)
	{
		retv = cff_ffmpeg_video_undo(video);
	}

	else
	{
		retv = cff_ffmpeg_video_redo(video);
	}

	if (retv < 0)
	{
		cff_guitl_beep();
	}

	else
	{
		cff_gui_video_button_cut_update_sensitive(video);
		GuiWidgetQueueDrawArea(gDrawingAreaBar);
	}

} /* static void cff_gui_edit_undo_redo(cff_video_t*, int) */

#ifdef HAVE_GMENU
static void cff_gui_edit_undo(GSimpleAction *a, GVariant *p, gpointer data)
#else
static void cff_gui_edit_undo(GtkWidget *w, gpointer data)
#endif
{
	cff_gui_edit_undo_redo((cff_video_t *) data, 0);

} /* static void cff_gui_edit_undo(GtkWidget*, gpointer) */

#ifdef HAVE_GMENU
static void cff_gui_edit_redo(GSimpleAction *a, GVariant *p, gpointer data)
#else
static void cff_gui_edit_redo(GtkWidget *w, gpointer data)
#endif
{
	cff_gui_edit_undo_redo((cff_video_t *) data, 1);

} /* static void cff_gui_edit_redo(GtkWidget*, gpointer) */

#ifdef HAVE_GMENU
static void cff_gui_edit_load(GSimpleAction *a, GVariant *p, gpointer *data)
#else
static void cff_gui_edit_load(GtkWidget *w, gpointer *data)
#endif
{
	gint		n;
	gchar		*fname;
	GFile		*file;
	GtkWidget	**mbar;
	cff_video_t	*video;

	mbar  = (GtkWidget **) data;
	video = &gVideo;

	if (video->dirty &&
		cff_guitl_dialog_warning_2btn(
			gettext_nop("Add list to existing cutpoints?"),
						cff_gui_window_get_topwin(),
								mbar) == FALSE)
	{
		/* FALSE => ! GTK_RESPONSE_ACCEPT */
		return;
	}

	file = cff_gui_file_selection_new_gettext(gettext_nop("Load Cut List"),
					NULL, NULL, mbar, kPropStrEditCutListPath);
	if (file == NULL)
	{
		return;
	}

	/* Load cut list ... */

	fname = g_file_get_path(file);
	g_object_unref(G_OBJECT(file));

	n = -1;

	if (fname != NULL)
	{
		FILE *fd;

		fd = g_fopen(fname, "r");
		g_free(fname);

		if (fd != NULL)
		{
			n = cff_ffmpeg_video_cut_list_load(&gVideo, fd);
			fclose(fd);
		}

	} /* if (fname != NULL) */

	if (n < 0)
	{
		cff_guitl_beep();
		cff_guitl_dialog_error(gettext_nop("Cannot load cut list!"), NULL);
	}

	else
	{
		cff_gui_video_button_cut_update_sensitive(video);
		GuiWidgetQueueDrawArea(gDrawingAreaBar);

		if (n > 0)
		{
			gchar *str = g_strdup_printf("%d %s",
				n, (n == 1) ?
				gettext("item was discarded!") :
					gettext("items were discarded!"));
			cff_guitl_dialog_warning(str, mbar);
			g_free(str);
		}

	} /* if (n < 0) .. else */

} /* static void cff_gui_edit_load(GtkWidget*, gpointer*) */

#ifdef HAVE_GMENU
static void cff_gui_edit_save(GSimpleAction *a, GVariant *p, gpointer *data)
#else
static void cff_gui_edit_save(GtkWidget *w, gpointer *data)
#endif
{
	gint		retv;
	gchar		*dot, *fname;
	GtkWidget	**mbar;
	GFile		*file;
	cff_video_t	*video;

	mbar  = (GtkWidget **) data;
	video = &gVideo;

	if (video->fname.utf8 == NULL)
	{
		return;
	}

	dot = strrchr(video->fname.utf8, '.');

	if (dot != NULL)
	{
		/* Truncate */
		*dot = '\0';
	}

	fname = g_strconcat(video->fname.utf8, ".cff", NULL);

	if (dot != NULL)
	{
		/* Restore */
		*dot = '.';
	}

	file = cff_gui_file_selection_new_gettext(gettext_nop("Save Cut List"),
				fname, NULL, mbar, kPropStrEditCutListPath);
	g_free(fname);

	if (file == NULL)
	{
		return;
	}

	/* Save cut list ... */

	fname = g_file_get_path(file);
	g_object_unref(G_OBJECT(file));

	retv = -1;

	if (fname != NULL)
	{
		FILE *fd;

		fd = g_fopen(fname, "w");
		g_free(fname);

		if (fd != NULL)
		{
			retv = cff_ffmpeg_video_cut_list_save(&gVideo, fd);
			fclose(fd);
		}
	}

	if (retv < 0)
	{
		cff_guitl_beep();
		cff_guitl_dialog_error(gettext_nop("Cannot save cut list!"), NULL);
	}

} /* static void cff_gui_edit_save(GtkWidget*, GtkWidget**) */

#ifdef HAVE_GMENU
static void cff_gui_edit_clear(GSimpleAction *a, GVariant *p, gpointer *data)
#else
static void cff_gui_edit_clear(GtkWidget *w, gpointer *data)
#endif
{
	GtkWidget	**mbar;
	cff_video_t	*video;

	mbar  = (GtkWidget **) data;
	video = &gVideo;

	if (video->cut.num == 0 ||
		cff_guitl_dialog_warning_2btn(
			gettext_nop("Clear cut list?"),
				cff_gui_window_get_topwin(), mbar) == FALSE)
	{
		/* FALSE => ! GTK_RESPONSE_ACCEPT */
		return;
	}

	cff_ffmpeg_video_cut_list_clear(video);

	cff_gui_video_button_cut_set_sensitive(TRUE);
	GuiWidgetQueueDrawArea(gDrawingAreaBar);

} /* static void cff_gui_edit_clear(GtkWidget*, gpointer*) */

#ifdef HAVE_GMENU
static void cff_gui_edit_add(GSimpleAction *a, GVariant *p, gpointer data)
#else
static void cff_gui_edit_add(GtkWidget *w, gpointer data)
#endif
{
	if (cff_ffmpeg_video_cut_add((cff_video_t *) data) < 0)
	{
		cff_guitl_beep();
	}

	else
	{
		cff_gui_video_button_cut_set_sensitive(FALSE);
		GuiWidgetQueueDrawArea(gDrawingAreaBar);
	}

} /* static void cff_gui_edit_add(GtkWidget*, gpointer) */

#ifdef HAVE_GMENU
static void cff_gui_edit_add_clicked(GtkWidget *w, gpointer data)
{
	if (gIsBusy == 0 && ! GuiMainLevelGetBusy())
	{
		cff_gui_edit_add(NULL, NULL, data);
	}

} /* static void cff_gui_edit_add_clicked(GtkWidget*, gpointer) */
#endif

#ifdef HAVE_GMENU
static void cff_gui_edit_remove(GSimpleAction *a, GVariant *p, gpointer data)
#else
static void cff_gui_edit_remove(GtkWidget *w, gpointer data)
#endif
{
	if (cff_ffmpeg_video_cut_remove((cff_video_t *) data) < 0)
	{
		cff_guitl_beep();
	}

	else
	{
		cff_gui_video_button_cut_set_sensitive(TRUE);
		GuiWidgetQueueDrawArea(gDrawingAreaBar);
	}

} /* static void cff_gui_edit_remove(GtkWidget*, gpointer) */

#ifdef HAVE_GMENU
static void cff_gui_edit_remove_clicked(GtkWidget *w, gpointer data)
{
	if (gIsBusy == 0 && ! GuiMainLevelGetBusy())
	{
		cff_gui_edit_remove(NULL, NULL, data);
	}

} /* static void cff_gui_edit_remove_clicked(GtkWidget*, gpointer) */
#endif

#ifdef HAVE_GMENU
static void cff_gui_edit_frame_display(GSimpleAction *a, GVariant *parm, gpointer data)
#else
static void cff_gui_edit_frame_display(GtkWidget *w, gpointer data)
#endif
{
	enum cff_VideoFrame	next;
	cff_frame_seek_t	frame_seek;

#ifdef HAVE_GMENU
	next = g_variant_get_uint32(parm);
#else
	next = GPOINTER_TO_INT(data);
#endif

	frame_seek.type		= kFrameSeekCutpoint;
	frame_seek.value.e_next	= next;
	cff_gui_video_frame(&frame_seek);

} /* static void cff_gui_edit_frame_display(GtkWidget*, gpointer) */

#ifdef HAVE_GMENU
static void cff_gui_edit_program_open(GSimpleAction *a, GVariant *parm, gpointer data)
#else
static void cff_gui_edit_program_open(GtkWidget *w, gpointer data)
#endif
{
	cff_video_t *video = &gVideo;

	if (video->dirty &&
		cff_guitl_dialog_warning_2btn(
			gettext_nop("Clear cut list?"),
				cff_gui_window_get_topwin(),
					&gMenuBarTitle[kMenuBarEdit]) == FALSE)
	{
		/* FALSE => ! GTK_RESPONSE_ACCEPT */
		return;
	}

	cff_ffmpeg_video_cut_list_clear(video);

	cff_gui_save_video_prev_settings_free();
	cff_gui_video_button_set_sensitive(FALSE);
	GuiWidgetQueueDrawArea(gDrawingAreaBar);

#ifdef HAVE_GMENU
	data = GINT_TO_POINTER(g_variant_get_uint32(parm));
#endif
	g_idle_add_full(G_PRIORITY_LOW,
		(GSourceFunc) cff_gui_program_open, data, NULL);

} /* static void cff_gui_edit_program_open(GtkWidget*, gpointer) */


/*
	----------------
	 Menu 'Options'
	----------------
*/

#ifdef HAVE_GMENU
static void cff_gui_opt_accel_dialog(GSimpleAction *a, GVariant *parm, gpointer *data)
#else
static void cff_gui_opt_accel_dialog(GtkWidget *w, gpointer *data)
#endif
{
	/* Set/unset some shortcuts */

	GtkWidget **mbar = (GtkWidget **) data;

	if (mbar != NULL && *mbar != NULL)
	{
		GuiWidgetSetStateSelected(*mbar);
	}

	cff_accel_dialog_run();

	if (mbar != NULL && *mbar != NULL)
	{
		GuiWidgetUnsetStateSelected(*mbar);
	}

} /* static void cff_gui_opt_accel_dialog(GtkWidget*, gpointer*) */


#ifdef HAVE_GMENU

static void cff_gui_opt_toggle_bool(GSimpleAction *action, GVariant *p, gpointer data)
{
	int	active, prop;
	guint	n;
	GVariant *old_state, *new_state;

	old_state = g_action_get_state(G_ACTION(action));
	active    = ! g_variant_get_boolean(old_state);
	g_variant_unref(old_state);

	new_state = g_variant_new_boolean(active);

	g_action_change_state(G_ACTION(action), new_state);
	prop = GPOINTER_TO_INT(data);

	cff_gprop_set_bool(prop, active);

	n = prop - kPropBoolOptKeyFrame;

	if (n <= kPropBoolOptPFrame)
	{
		if (action == gCheckmenuFrameAction[n])
		{
			/* Menu clicked */
			action = gPopupCheckmenuFrameAction[n];
		}

		else
		{
			/* Popup-Menu clicked */
			action = gCheckmenuFrameAction[n];
		}

		g_action_change_state(G_ACTION(action), new_state);

	} /* if (n <= kPropBoolOptPFrame) */

} /* static void cff_gui_opt_toggle_bool(GSimpleAction*, GVariant*, gpointer) */

#else /* !  HAVE_GMENU */

static void cff_gui_opt_toggle_bool(GtkWidget *mi, gpointer data)
{
	static gboolean clicked = FALSE;

	int	active, prop;
	guint	n;

	if (clicked)
	{
		/* Already handled */
		clicked = FALSE;
		return;
	}

	active = GuiCheckMenuItemGetActive(GTK_CHECK_MENU_ITEM(mi));
	prop = GPOINTER_TO_INT(data);

	cff_gprop_set_bool(prop, active);

	n = prop - kPropBoolOptKeyFrame;

	if (n <= kPropBoolOptPFrame)
	{
		if (mi == gCheckmenuFrameAction[n])
		{
			/* Menu clicked */
			mi = gPopupCheckmenuFrameAction[n];
		}

		else
		{
			/* Popup-Menu clicked */
			mi = gCheckmenuFrameAction[n];
		}

		/* Following function emits the activate-signal */
		clicked = TRUE;

		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi), active);

	} /* if (n <= kPropBoolOptPFrame) */

} /* static void cff_gui_opt_toggle_bool(GtkWidget*, gpointer) */

#endif /* #ifdef HAVE_GMENU .. #else */


#ifdef HAVE_GMENU
static void cff_gui_opt_logwindow_menu_selected(GSimpleAction *a, GVariant *parm, gpointer data)
#else
static void cff_gui_opt_logwindow_menu_selected(GtkWidget *mi, gpointer data)
#endif
{
	int		logindex;
	const gchar	*logname;

#ifdef HAVE_GMENU
	logindex = g_variant_get_uint32(parm);
#else
	logindex = GPOINTER_TO_INT(data);
#endif

	logname = cff_ffmpeg_get_loglevel_name(logindex);

	if (logindex == 0)
	{
		GuiLogWindowFree(NULL, NULL, &gLogWindow);
		cff_gui_logwindow_set_title(gLogWindow.window, logname);
	}

	else
	{
		if (cff_gui_logwindow_open(logname) == TRUE)
		{
			cff_gprop_set_str(kPropStrOptLogLevel, logname);
		}
	}

} /* static void cff_gui_opt_logwindow_menu_selected(GtkWidget*, gpointer) */


/*
	-------------
	 Menu 'Help'
	-------------
*/

#ifdef HAVE_GMENU
static void cff_gui_help_ffmpeg(GSimpleAction *a, GVariant *p, gpointer *data)
#else
static void cff_gui_help_ffmpeg(GtkWidget *w, gpointer *data)
#endif
{
	int		i;
	char		*str;
	const char	*title;
	GtkWidget	*vbox, **mbar;

	vbox = GuiVBoxNew(FALSE, 0);
	i    = 0;

	do
	{
		switch (i)
		{
			case 0:
				str	= cff_ffmpeg_get_libs_info();
				title	= gettext_nop("Libs");
				break;
			case 1:
				str	= cff_ffmpeg_get_formats_info();
				title	= gettext_nop("Formats");
				break;
			case 2:
				str	= cff_ffmpeg_get_codecs_info();
				title	= gettext_nop("Codecs");
				break;
			default: /*case 3: */
				str	= cff_ffmpeg_get_bitstream_filters_info();
				title	= gettext_nop("Bitstream Filters");
				break;

		} /* switch (i) */

		cff_gui_ffmpeg_info_label(title, str, vbox);

		if (str)
		{
			/* malloc'ed */
			free(str);
		}
	}
	while (++i < 4);

	mbar = (GtkWidget **) data;
	cff_gui_scrolled_info_dialog(gettext_nop("CutterFF: FFmpeg"),
				vbox, mbar, kPropIntHelpFFmpegWinWidth);

} /* static void cff_gui_help_ffmpeg(GtkWidget*, gpointer*) */

#ifdef HAVE_GMENU
static void cff_gui_help_cutterff(GSimpleAction *a, GVariant *p, gpointer *data)
#else
static void cff_gui_help_cutterff(GtkWidget *w, gpointer *data)
#endif
{
	static const gchar *cff_help_str[] =
	{
		gettext_nop("Navigation"),
		gettext_nop("Button '<' or Key 'Left' | Button '>' or Key 'Right'"),
		gettext_nop("Previous or next frame"),
		gettext_nop("Button '<<' or Key 'Down' | Button '>>' or Key 'Up'"),
		gettext_nop("Half second back or forward"),
		gettext_nop("Button '<<<' or Key 'Page Down' | Button '>>>' or Key 'Page Up'"),
		gettext_nop("Five seconds back or forward"),
		gettext_nop("Button '|<' or Key 'Home' | Button '>|' or Key 'End'"),
		gettext_nop("Previous or next cutpoint or start or end of video"),
		gettext_nop("With shift key or the right mouse button instead of the left"),
		gettext_nop("Button '<' or Key 'Left' | Button '>' or Key 'Right'"),
		gettext_nop("One minute back or forward"),
		gettext_nop("Button '<<' or Key 'Down' | Button '>>' or Key 'Up'"),
		gettext_nop("Five minutes back or forward"),
		gettext_nop("Button '<<<' or Key 'Page Down' | Button '>>>' or Key 'Page Up'"),
		gettext_nop("Ten minutes back or forward"),
		gettext_nop("Button '|<' or Key 'Home' | Button '>|' or Key 'End'"),
		gettext_nop("Start or end of video"),
		gettext_nop("A mouseclick on the position bar will jump to this postion."),
		gettext_nop("Note: the times above depends on the video and may differ!")
	};

	gint		i;
	gchar		*markup, **dst, *str[sizeof(cff_help_str) / sizeof(gchar*)];
	const gchar	**src;
	GtkWidget	*vbox, *label, **mbar;

	vbox  = GuiVBoxNew(FALSE, 0);
	label = gtk_label_new(NULL);
	GuiVBoxPackStart(vbox, label, TRUE, TRUE, 4);

	i   = sizeof(cff_help_str) / sizeof(gchar*);
	src = cff_help_str;
	dst = str;

	do
	{
		*(dst++) = cff_charconv_utf8_gettext(*(src++));
	}
	while (--i > 0);

	markup = g_markup_printf_escaped(
		"<span size=\"xx-large\" weight=\"ultrabold\" stretch=\"ultraexpanded\"><i><u>%s</u></i></span>\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<u><b>%s</b></u>\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n"
		"%s\n\n"
		"<b>%s</b>\n\n"
		"%s",
		str[0],
		str[1],
		str[2],
		str[3],
		str[4],
		str[5],
		str[6],
		str[7],
		str[8],
		str[9],
		str[10],
		str[11],
		str[12],
		str[13],
		str[14],
		str[15],
		str[16],
		str[17],
		str[18],
		str[19]);

	gtk_label_set_markup(GTK_LABEL(label), markup);

	i   = sizeof(cff_help_str) / sizeof(gchar*);
	src = cff_help_str;
	dst = str;

	do
	{
		gchar		*d;
		const gchar	*s;

		s = *(src++);
		d = *(dst++);
		CFF_CHARCONV_UTF8_FREE(d, s);
	}
	while (--i > 0);

	mbar = (GtkWidget **) data;
	cff_gui_scrolled_info_dialog(gettext_nop("CutterFF: Help"),
				vbox, mbar, kPropIntHelpCutterFFWidth);
	
} /* static void cff_gui_help_cutterff(GtkWidget*, gpointer*) */

#ifdef HAVE_GMENU
static void cff_gui_help_about(GSimpleAction *a, GVariant *p, gpointer *data)
#else
static void cff_gui_help_about(GtkWidget *w, gpointer *data)
#endif
{
	static const gchar cff_about_str_0[] =
			gettext_nop("A small programm for cuttting videos");
	static const gchar cff_about_str_1[] =
			gettext_nop("Distributed under GNU-GPL 2 or later");
	static const gchar cff_about_str_2[] = gettext_nop("Version");
	static const gchar cff_about_str_3[] = gettext_nop("Author:");

	/* main.c */
	extern const gchar *gCffMainCutterffVersionString;

	GtkWidget	*frame, *label, **mbar;
	gint		major, minor, micro;
	gchar		*markup, *str[4];

	label = gtk_label_new(NULL);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, 4, label);

	str[0] = cff_charconv_utf8_gettext(cff_about_str_0);
	str[1] = cff_charconv_utf8_gettext(cff_about_str_1);
	str[2] = cff_charconv_utf8_gettext(cff_about_str_2);
	str[3] = cff_charconv_utf8_gettext(cff_about_str_3);

#if GTK_MAJOR_VERSION == 2
	major = GTK_MAJOR_VERSION;
	minor = GTK_MINOR_VERSION;
	micro = GTK_MICRO_VERSION;
#else
	major = gtk_get_major_version();
	minor = gtk_get_minor_version();
	micro = gtk_get_micro_version();
#endif

	markup = g_markup_printf_escaped(
		"<span size=\"xx-large\" weight=\"ultrabold\""
			" stretch=\"ultraexpanded\"><i><u>%s</u></i></span>\n\n"
		"<tt>%s\n%s <b>%s</b> %s <b>%s-%i.%i.%i</b></tt>\n\n"
		"%s\n\n"
		"%s %s (%u-bit)\n\n"
		"%s %s %s\n"
		"%s",

		"CutterFF",
		str[0],
		"using", "FFmpeg", "and", "GTK+", major, minor, micro,
		str[1],
		str[2], gCffMainCutterffVersionString,
			(unsigned int) (8 * sizeof(char*)),
		str[3],
		/* Harald F&ouml;rster */
		"Harald F" "\xC3\xB6" "rster",
		"<harald_foerster@users.sourceforge.net>",
		"URL: https://cutterff.sourceforge.io/");

	gtk_label_set_markup(GTK_LABEL(label), markup);
	g_free(markup);

	CFF_CHARCONV_UTF8_FREE(str[3], cff_about_str_3);
	CFF_CHARCONV_UTF8_FREE(str[2], cff_about_str_2);
	CFF_CHARCONV_UTF8_FREE(str[1], cff_about_str_1);
	CFF_CHARCONV_UTF8_FREE(str[0], cff_about_str_0);

	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
	GuiMiscSetPadding(label, 15);

	mbar = (GtkWidget **) data;
	cff_guitl_dialog_okay_run(gettext_nop("CutterFF: About"),
							frame, mbar, -1);

} /* static void cff_gui_help_about(GtkWidget*, gpointer*) */


/*
	----------------
	 Menu Functions
	----------------
*/

static void cff_gui_adjust_popup(GtkWidget *popmenu, gpointer data)
{
	/* Show signal */

	guint		item;
	GuiMenuAction	**menu;
	cff_video_t	*video;
	gboolean	*ptr, popup_sensitive[kPopupItemsTotal];

	memset(popup_sensitive, '\0', sizeof(popup_sensitive));

	video = data;

	if (video->vframe.ic != NULL)
	{
		if (video->vframe.undo[0] != NULL)
		{
			popup_sensitive[kPopupEditUndo] = TRUE;
		}

		if (video->vframe.redo[0] != NULL)
		{
			popup_sensitive[kPopupEditRedo] = TRUE;
		}

		if (video->cut.list == NULL || video->cut.num != 1 ||
			video->cut.list[0].frame != video->vframe.dts_stop)
		{
			popup_sensitive[kPopupFileSave] = TRUE;
		}

		popup_sensitive[kPopupEditLoad] = TRUE;

		if (video->cut.num > 0)
		{
			popup_sensitive[kPopupEditSave]  = TRUE;
			popup_sensitive[kPopupEditClear] = TRUE;
			popup_sensitive[kPopupEditGoto]  = TRUE;
		}

	} /* if (video->ic != NULL) */

	menu = &gPopupAction[kPopupItemsTotal];
	ptr  = &popup_sensitive[kPopupItemsTotal];
	item = kPopupItemsTotal;

	do
	{
		GuiMenuItemSetSensitive(*(--menu), *(--ptr));
	}
	while (--item != 0);

} /* static void cff_gui_adjust_popup(GtkWidget*, gpointer) */


#ifdef HAVE_GMENU
gboolean cff_gui_adjust_menu(GSimpleAction *action, GVariant *p, gpointer data)
#else
gboolean cff_gui_adjust_menu(GtkWidget *action, gpointer data)
#endif
{
	guint		item;
	GuiMenuAction	**menu;
	cff_video_t	*video;
	gboolean	*ptr, menu_sensitive[kMenuItemsTotal];

	if (gIsBusy != 0 || GuiMainLevelGetBusy())
	{
		return TRUE;
	}

	video = data;

#ifdef HAVE_GMENU

	static gint state = 0;

	if (video != NULL)
	{
		/* Menubar */

		state = ~state;

		if (state == 0)
		{
			return TRUE;
		}
	}

#endif /* HAVE_GMENU */

	menu = &gMenuAction[kMenuItemsTotal];
	item = kMenuItemsTotal;

	if (video == NULL)
	{
		/* Called from 'cff_gaccel.c' */

		do
		{
			if (*(--menu) == action)
			{
				/* Menu item found */
				break;
			}
		}
		while (--item > 0);

		if (item == 0)
		{
			/* No menu item to enable/disable */
			return FALSE;
		}

		video = &gVideo;

	} /* if (video == NULL) */

	memset(menu_sensitive, '\0', sizeof(menu_sensitive));

	if (video->inFormatCtx != NULL)
	{
		menu_sensitive[kMenuFileClose] = TRUE;
		menu_sensitive[kMenuFileInfo]  = TRUE;
	}

	if (video->vframe.ic != NULL)
	{
		if (video->vframe.undo[0] != NULL)
		{
			menu_sensitive[kMenuEditUndo] = TRUE;
		}

		if (video->vframe.redo[0] != NULL)
		{
			menu_sensitive[kMenuEditRedo] = TRUE;
		}

		if (video->cut.list == NULL || video->cut.num != 1 ||
			video->cut.list[0].frame != video->vframe.dts_stop)
		{
			menu_sensitive[kMenuFileSave] = TRUE;
		}

		menu_sensitive[kMenuFileImage] = TRUE;
		menu_sensitive[kMenuEditLoad]  = TRUE;

		if (video->cut.num > 0)
		{
			menu_sensitive[kMenuEditSave]  = TRUE;
			menu_sensitive[kMenuEditClear] = TRUE;
			menu_sensitive[kMenuEditGoto]  = TRUE;

			if (cff_ffmpeg_video_cut_find_current(video) >= 0)
			{
				menu_sensitive[kMenuEditRemove] = TRUE;
			}
		}

		if (menu_sensitive[kMenuEditRemove] == FALSE)
		{
			menu_sensitive[kMenuEditAdd] = TRUE;
		}

		if (video->inFormatCtx != NULL &&
			video->program.index != NULL &&
					video->program.num > 1)
		{
			menu_sensitive[kMenuEditProgram] = TRUE;
		}

	} /* if (video->ic != NULL) */

	menu = &gMenuAction[kMenuItemsTotal];
	ptr  = &menu_sensitive[kMenuItemsTotal];
	item = kMenuItemsTotal;

	do
	{
		GuiMenuItemSetSensitive(*(--menu), *(--ptr));
	}
	while (--item != 0);

	return FALSE;

} /* gboolean cff_gui_adjust_menu(GtkWidget*, gpointer) */

static gboolean _cff_gui_destroy_submenu(GuiMenu *submenu)
{

#ifdef HAVE_GMENU

	GtkWidget *widget = gWindata.mainwin;

	if (submenu == gSubmenuPopup)
	{
		widget = gPopupMenuWidget;
	}

	/* Remove action group from widget */
	gtk_widget_insert_action_group(widget, "sub", NULL);

	/* Remove all items in the menu */
	g_menu_remove_all(submenu);
#else
	GList *children;

	while ((children = GuiMenuShellGetChildren(submenu)) != NULL)
	{
		gtk_widget_destroy(GTK_WIDGET(children->data));
	}

#endif /* #ifdef HAVE_GMENU .. #else */

	return Gui_SOURCE_REMOVE;

} /* static gint _cff_gui_destroy_submenu(GuiMenu*) */

static void cff_gui_destroy_submenu(GuiMenu *submenu, gpointer data)
{
	/*
		Destroy a little bit later, else the
		'activate' - signal will not be emitted
	*/

	g_idle_add_full(G_PRIORITY_LOW,
		(GSourceFunc) _cff_gui_destroy_submenu, submenu, NULL);

} /* static void cff_gui_destroy_submenu(GuiMenu*, gpointer) */


#ifdef HAVE_GMENU

#if GTK_MAJOR_VERSION == 4

static gboolean _cff_gui_submenu_wait(gpointer data)
{
	return Gui_SOURCE_REMOVE;

} /* static gboolean _cff_gui_submenu_wait(gpointer) */

static void cff_gui_submenu_wait(void)
{
	g_idle_add_full(G_PRIORITY_LOW,
		(GSourceFunc) _cff_gui_submenu_wait,
			NULL, (GDestroyNotify) cff_guitl_main_level_quit);

	cff_guitl_main_level_start();

} /* static void cff_gui_submenu_wait(void) */

#endif /* GTK_MAJOR_VERSION == 4 */

static gboolean cff_gui_submenu_is_empty(GSimpleAction *action, GMenu *submenu)
{
	gboolean is_enabled = g_action_get_enabled(G_ACTION(action));

	if (g_menu_model_get_n_items(G_MENU_MODEL(submenu)) != 0)
	{
		if (is_enabled == FALSE)
		{
			/* Destroy now */
			_cff_gui_destroy_submenu(submenu);
		}

		else
		{
			/* Destroy when idle */
			cff_gui_destroy_submenu(submenu, NULL);
		}

		return TRUE;

	} /* if (g_menu_model_get_n_items(G_MENU_MODEL(submenu)) != 0) */

#if GTK_MAJOR_VERSION == 4
	/* Wait for '_cff_gui_destroy_submenu' and/or draw the submenu */
	cff_gui_submenu_wait();
#endif

	if (is_enabled == FALSE)
	{
		GMenuItem *mi;

		mi = cff_guitl_menu_item_new_gettext("Empty");
		g_menu_item_set_detailed_action(mi, "sub.subx");
		g_menu_append_item(submenu, mi);
		g_object_unref(mi);

		return TRUE;

	} /* if (is_enabled == FALSE) */

	return FALSE;

} /* static gboolean cff_gui_submenu_is_empty(GSimpleAction*, GMenu*) */

static GMenuItem *cff_gui_build_submenu_item(const gchar *str, GCallback func, gint idx, gint disable, GSimpleActionGroup **ptr)
{
	GMenuItem *item = g_menu_item_new(str, NULL);

	if (idx == disable)
	{
		/* Invalid action => disabled */
		g_menu_item_set_detailed_action(item, "sub.subx");
		return item;
	}

	g_menu_item_set_action_and_target(item, "sub.sub", "u", idx);

	if (*ptr == NULL)
	{
		GSimpleAction		*action;
		GSimpleActionGroup	*group;
		GVariantType		*vtype;

		group = g_simple_action_group_new();
		*ptr  = group;

		/* G_VARIANT_TYPE_UINT32 */
		vtype  = g_variant_type_new("u");
		action = g_simple_action_new("sub", vtype);

		g_action_map_add_action(G_ACTION_MAP(group), G_ACTION(action));
		gtk_widget_insert_action_group(gWindata.mainwin,
						"sub", G_ACTION_GROUP(group));

		GuiSignalConnect(action, "activate", func, NULL);

	} /* if (*ptr == NULL) */

	return item;

} /* static GMenuItem *cff_gui_build_submenu_item(const gchar*, GCallback, gint, gint, GSimpleActionGroup**) */

#else /* ! HAVE_GMENU */

static GtkWidget *cff_gui_build_submenu_item(const gchar *str, GCallback func, gint idx, gint disable)
{
	GtkWidget *item = gtk_menu_item_new_with_label(str);

	if (idx == disable)
	{
		gtk_widget_set_sensitive(item, FALSE);
	}

	else
	{
		GuiSignalConnect(item, "activate", func, GINT_TO_POINTER(idx));
	}

	gtk_widget_show(item);

	return item;

} /* static GtkWidget *cff_gui_build_submenu_item(const gchar*, GCallback, gint, gint) */

#endif /* #ifdef HAVE_GMENU .. #else */


#ifdef HAVE_GMENU
static void cff_gui_build_frame_submenu(GSimpleAction *a, GVariant *p, gpointer data)
#else
static void cff_gui_build_frame_submenu(GtkWidget *submenu, gpointer data)
#endif
{
	unsigned int		i, current;
	cff_video_t		*video;

#ifdef HAVE_GMENU

	GMenu			*submenu;
	GSimpleActionGroup	*group;

	submenu	= data;

	if (cff_gui_submenu_is_empty(a, submenu))
	{
		return;
	}

	group	= NULL;
	video	= &gVideo;
#else
	video	= data;
#endif
	current = cff_ffmpeg_video_cut_find_current(video);

	i = 0;

	do
	{
		unsigned int	frame_number;
		GuiMenuItem	*mi;
		gchar		buf[64];

		frame_number = cff_ffmpeg_get_cut_frame_number(video, i);

		if (frame_number == (unsigned int) -1)
		{
			continue;
		}

		g_snprintf(buf, sizeof(buf), "%u", frame_number);

		mi = GuiBuildSubmenuItem(buf,
			cff_gui_edit_frame_display, i,
					current, &group);

		GuiMenuShellAppend(submenu, mi);

#ifdef HAVE_GMENU
		g_object_unref(mi);
#endif
	}
	while (++i < video->cut.num);


#ifdef HAVE_GMENU

	if (submenu == gSubmenuPopup)
	{
		/* Move the action group into the popup widget */
		gtk_widget_insert_action_group(gPopupMenuWidget,
						"sub", G_ACTION_GROUP(group));
		gtk_widget_insert_action_group(gWindata.mainwin, "sub", NULL);
	}

#endif /* HAVE_GMENU */

} /* static void cff_gui_build_frame_submenu(GtkWidget*, gpointer) */

#ifdef HAVE_GMENU
static void cff_gui_build_program_submenu(GSimpleAction *a, GVariant *p, gpointer data)
#else
static void cff_gui_build_program_submenu(GtkWidget *submenu, gpointer data)
#endif
{
	unsigned int		i;
	cff_video_t		*video;

#ifdef HAVE_GMENU

	GMenu			*submenu;
	GSimpleActionGroup	*group;

	submenu	= data;

	if (cff_gui_submenu_is_empty(a, submenu))
	{
		return;
	}

	group = NULL;
	video = &gVideo;
#else
	video = data;
#endif

	i = 0;

	do
	{
		int		prgnum;
		const char	*name;
		GuiMenuItem	*mi;
		gchar		buf[64];

		name = cff_ffmpeg_get_program_name(video, i, &prgnum);

		if (name != NULL)
		{
			if (cff_tl_is_ascii_str(name) == FALSE)
			{
				if (strlen(name) >= sizeof(buf))
				{
					name = NULL;
				}

				else
				{
					strcpy(buf, name);
					name = cff_tl_replace_non_utf8_chars_in_str(buf);
				}
			}

		} /* if (name != NULL) */

		if (name == NULL)
		{
			if (prgnum < 0)
			{
				prgnum = i + 1;
			}

			g_snprintf(buf, sizeof(buf), "%d", prgnum);
			name = buf;

		} /* if (name == NULL) */

		mi = GuiBuildSubmenuItem(name,
			cff_gui_edit_program_open, i,
				video->program.selected, &group);

		GuiMenuShellAppend(submenu, mi);

#ifdef HAVE_GMENU
		g_object_unref(mi);
#endif
	}
	while (++i < video->program.num);

} /* static void cff_gui_build_program_submenu(GtkWidget*, gpointer) */

#ifdef HAVE_GMENU
static void cff_gui_build_loglevel_submenu(GSimpleAction *a, GVariant *p, gpointer data)
#else
static void cff_gui_build_loglevel_submenu(GtkWidget *submenu, gpointer data)
#endif
{
	int			i, current;

#ifdef HAVE_GMENU

	GMenu			*submenu;
	GSimpleActionGroup	*group;

	submenu = data;

	if (g_menu_model_get_n_items(G_MENU_MODEL(submenu)) != 0)
	{
		cff_gui_destroy_submenu(submenu, NULL);
		return;
	}

#if GTK_MAJOR_VERSION == 4
	/* Wait for '_cff_gui_destroy_submenu' and/or draw the submenu */
	cff_gui_submenu_wait();
#endif

	group = NULL;

#endif /* HAVE_GMENU */

	if (gLogWindow.window == NULL ||
		(current = cff_ffmpeg_get_current_loglevel_index()) < 0)
	{
		current = 0;
	}

	i = 0;

	for (;;)
	{
		GuiMenuItem	*mi;
		const gchar	*logname;

		logname = cff_ffmpeg_get_loglevel_name(i);

		if (logname == NULL)
		{
			break;
		}

		mi = GuiBuildSubmenuItem(logname,
			cff_gui_opt_logwindow_menu_selected,
						i, current, &group);
		GuiMenuShellAppend(submenu, mi);

#ifdef HAVE_GMENU
		g_object_unref(mi);
#endif
		i++;

	} /* for (;;) */

} /* static void cff_gui_build_loglevel_submenu(GtkWidget*, gpointer) */

#ifdef HAVE_GMENU
static gboolean cff_gui_mouse_button_show_popup_menu(GtkGesture *gesture,
		gint n_press, gdouble x, gdouble y, GtkWidget *popup_menu)
#else
static gboolean cff_gui_mouse_button_show_popup_menu(GtkWidget *widget,
					GdkEvent *event, GtkWidget *popup_menu)
#endif
{
	if (gIsBusy != 0 || GuiMainLevelGetBusy())
	{
		return TRUE;
	}

#ifdef HAVE_GMENU

	gtk_gesture_set_state(gesture, GTK_EVENT_SEQUENCE_CLAIMED);

#if GTK_MAJOR_VERSION == 3
	gtk_menu_popup_at_pointer(GTK_MENU(popup_menu), NULL);
#else
	gtk_popover_set_pointing_to(GTK_POPOVER(popup_menu),
				&(const GdkRectangle){x, y, 1, 1 });
	gtk_popover_popup(GTK_POPOVER(popup_menu));
#endif

#else /* ! HAVE_GMENU */

	guint		button;
	guint32		time;
	GdkEventButton	*bevent;

	bevent = (GdkEventButton *) event;

	if (bevent == NULL)
	{
		time	= GDK_CURRENT_TIME;
		button	= 0;
	}

	else
	{
		button = bevent->button;

		if (bevent->type != GDK_BUTTON_PRESS || button != 3)
		{
			return FALSE;
		}

		time = bevent->time;

	} /* if (bevent == NULL) .. else */

	gtk_menu_popup(GTK_MENU(popup_menu), NULL, NULL, NULL, NULL, button, time);

#endif /* #if GTK_MAJOR_VERSION > 3 .. #else */

	return TRUE;

} /* static gboolean cff_gui_mouse_button_show_popup_menu(GtkWidget*,
							GdkEvent*, GtkWidget*) */

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

static gboolean cff_gui_sf10_show_popup_menu(GtkWidget *widget, GtkWidget *popup_menu)
{

#ifdef HAVE_GMENU
/	/* GTK_MAJOR_VERSION == 3 */
	gtk_menu_popup_at_pointer(GTK_MENU(popup_menu), NULL);
	return TRUE;
#else
	return cff_gui_mouse_button_show_popup_menu(widget, NULL, popup_menu);
#endif

} /* static gboolean cff_gui_sf10_show_popup_menu(GtkWidget*, GtkWidget*) */

#endif /* #if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 */


#ifdef HAVE_GMENU

#if GTK_MAJOR_VERSION == 3

static void cff_gui_build_set_menubar_title(GtkWidget *mb_title, guint *i)
{
	/*
		If a modal dialog was opened, the state of the
		menubar item should be set to prelight (selected)
		This is NOT documented !!!
	*/

	if (*i < kMenuBarTotal)
	{
		gMenuBarTitle[*i] = mb_title;
		(*i)++;
	}

} /* static void cff_gui_build_set_menubar_title(GtkWidget*, guint*) */

#endif /* GTK_MAJOR_VERSION == 3 */


static void cff_gui_build_menu_items(const cff_guiMenuItem_t *item, GMenu *menu, GtkWidget *window, GSimpleActionGroup *group)
{
	GMenu *section = menu;

	for (;;)
	{
		gint		i;
		GMenu		*submenu;
		GMenuItem	*mi;
		gpointer	parameter;

		/* Supress compiler warning */
		parameter = NULL;

		switch (item->type)
		{
			case kMTypeItem:
			case kMTypeCheckItem:
			case kMTypeSubmenuItem:
				mi = cff_guitl_menu_item_new_gettext(item->label);

				if (item->type == kMTypeSubmenuItem)
				{
					submenu = g_menu_new();
					g_menu_item_set_submenu(mi, G_MENU_MODEL(submenu));
					g_object_unref(submenu);
					parameter = submenu;

					/* Context menu */
					if (window == NULL)
					{
						gSubmenuPopup = submenu;
						break;
					}

					i = 0;

					do
					{
						if (gSubmenuMbar[i] == NULL)
						{
							gSubmenuMbar[i] = submenu;
							break;
						}
					}
					while (++i < 3);
				}

				else
				{
					parameter = item->data;
				}

				break;

			case kMTypeSeparator:
				section = g_menu_new();
				mi = g_menu_item_new_section(NULL, G_MENU_MODEL(section));
				g_object_unref(section);
				break;

			default:
				return;

		} /* switch (item->type) */

		if (item->callback != NULL)
		{
			const gchar	*instance;
			GSimpleAction	*action;
			GVariant	*variant;
			gchar		action_name[256];

			cff_guitl_create_action_name(action_name,
					sizeof(action_name) - 1, item->label);

#if GTK_MAJOR_VERSION == 4
			/* Don't display shortcuts in context menu */
			if (window == NULL)
			{
				action_name[0] = 'p';
				action_name[1] = 'o';
				action_name[2] = 'p';
			}
#endif
			if (item->type != kMTypeSubmenuItem)
			{
				g_menu_item_set_detailed_action(mi, action_name);
				instance = "activate";
			}

			if (item->type == kMTypeSubmenuItem)
			{
				g_menu_item_set_attribute(mi,
					"submenu-action", "s", action_name);

				instance = "change-state";
				variant  = g_variant_new_boolean(TRUE);
				action   = g_simple_action_new_stateful(
					&action_name[CFF_GACTION_PREFIX_LENGTH],
								NULL, variant);
			}

			else if (item->type == kMTypeCheckItem)
			{
				int active, prop;

				prop = GPOINTER_TO_INT(item->data);

				if (cff_gprop_get_bool(prop, &active) == FALSE)
				{
					if (prop == kPropBoolOptDelete ||
						prop == kPropBoolOptCursor ||
							prop == kPropBoolOptSaveConfig)
					{
						active = TRUE;
					}

					else
					{
						active = FALSE;
					}
				}

				variant = g_variant_new_boolean(active);
				action  = g_simple_action_new_stateful(
					&action_name[CFF_GACTION_PREFIX_LENGTH],
								NULL, variant);
			}

			else
			{
				action = g_simple_action_new(
					&action_name[CFF_GACTION_PREFIX_LENGTH],
									NULL);

			} /* if (item->type == kMTypeSubmenuItem) .. else if .. else */


			g_action_map_add_action(G_ACTION_MAP(group),
							G_ACTION(action));

			if (item->accel_id != kAccelKeyNone)
			{
				cff_accel_bind_widget(item->accel_id,
						item->label, action, window);
			}

			GuiSignalConnect(action, instance,
						item->callback, parameter);
			if (item->storage)
			{
				*(item->storage) = action;
			}

		} /* if (item->callback != NULL) */

		if (item->type == kMTypeSeparator)
		{
			g_menu_append_item(menu, mi);
		}

		else
		{
			g_menu_append_item(section, mi);
		}

		g_object_unref(mi);
		item++;

	} /* for (;;) */

} /* static void cff_gui_build_menu_items(const cff_guiMenuItem_t*, GMenu*, GtkWidget*, GSimpleActionGroup*) */

#else /* ! HAVE_GMENU */

static void cff_gui_build_menu_items(const cff_guiMenuItem_t *item, GtkWidget *menu, GtkWidget *window)
{
	for (;;)
	{
		int		active, prop;
		GtkWidget	*mi, *submenu;

		switch (item->type)
		{
			case kMTypeItem:
			case kMTypeSubmenuItem:
				mi = cff_guitl_menu_item_new_gettext(item->label);

				if (item->type == kMTypeItem)
				{
					break;
				}

				submenu = gtk_menu_new();
				gtk_menu_item_set_submenu(GTK_MENU_ITEM(mi), submenu);
				gtk_widget_realize(submenu);

				if (item->callback)
				{
					GuiSignalConnect(submenu, "show",
						item->callback, item->data);
					GuiSignalConnect(submenu, "hide",
						cff_gui_destroy_submenu, NULL);
				}

				break;

			case kMTypeCheckItem:
				mi = cff_guitl_check_menu_item_new_gettext(item->label);
				prop = GPOINTER_TO_INT(item->data);

				if (cff_gprop_get_bool(prop, &active) == FALSE)
				{
					if (prop == kPropBoolOptDelete ||
							prop == kPropBoolOptCursor ||
								prop == kPropBoolOptSaveConfig)
					{
						active = TRUE;
					}

					else
					{
						active = FALSE;
					}
				}

				gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mi), active);
				break;

			case kMTypeSeparator:
				mi = GuiMenuSeparatorNew();
				break;

			default:
				return;
		}

		gtk_menu_shell_append(GTK_MENU_SHELL(menu), mi);
		gtk_widget_show(mi);

		if (item->accel_id != kAccelKeyNone)
		{
			cff_accel_bind_widget(item->accel_id, item->label, mi, window);
		}

		if (item->callback && item->type != kMTypeSubmenuItem)
		{
			GuiSignalConnect(mi, "activate", item->callback, item->data);
		}

		if (item->storage)
		{
			*(item->storage) = mi;
		}

		item++;

	} /* for (;;) */

} /* static void cff_gui_build_menu_items(const cff_guiMenuItem_t*, GtkWidget*, GtkWidget*) */

#endif /* #ifdef HAVE_GMENU .. #else */


static GtkWidget *cff_gui_build_menu(GtkWidget *window)
{
	static const struct build_menu_array
	{
		const gchar		*label;
		const cff_guiMenuItem_t	*entries;
		gboolean		adjust_menu;

	} build_menu_array[] = {

		{gettext_nop("File"), cff_gui_menuFile, TRUE},
		{gettext_nop("Edit"), cff_gui_menuEdit, TRUE},
		{gettext_nop("Options"), cff_gui_menuOptions, FALSE},
		{gettext_nop("Help"), cff_gui_menuHelp, FALSE}
	};

#ifdef HAVE_GMENU
	const gchar		*group_name;
	GMenu			*model;
	GSimpleAction		*action;
	GSimpleActionGroup	*group;
	GVariant		*variant;
#if GTK_MAJOR_VERSION == 4
	GtkWidget		*mbar_title, **mbtitleptr;
#endif
#else
	GtkWidget		**mbtitleptr;
#endif
	gint			n;
	GtkWidget		*mbar, *popup;
	GuiMenu			*menu;

	const struct build_menu_array *build_tab;

#ifdef HAVE_GMENU
	model = g_menu_new();
	group = NULL;
#else
	mbtitleptr = gMenuBarTitle;
#endif
	mbar = GuiMenuBarNew(model);

	/* Ignore this event if busy */
	cff_gui_signal_connect_key_ignore(mbar);
	cff_gui_signal_connect_mouse_button_ignore(mbar);

#if GTK_MAJOR_VERSION == 2
	GuiMenuBarDisableShadow(mbar);
#endif

	build_tab = build_menu_array;
	n         = kMenuBarFile;

	do
	{
		GuiMenuItem *mb_title;

		mb_title = cff_guitl_menu_item_new_gettext(build_tab->label);
		menu     = GuiMenuNew();
		GuiMenuItemSetSubmenu(mb_title, menu);

#ifdef HAVE_GMENU

		if (build_tab->adjust_menu)
		{
			g_menu_item_set_attribute(mb_title,
				"submenu-action", "s", "app.cffadjustmenu");

			if (group == NULL)
			{
				group	= g_simple_action_group_new();
				variant	= g_variant_new_boolean(TRUE);
				action	= g_simple_action_new_stateful("cffadjustmenu",
									NULL, variant);
				g_action_map_add_action(G_ACTION_MAP(group),
								G_ACTION(action));

				gtk_widget_insert_action_group(window,
							CFF_GACTION_GROUP_STRING,
								G_ACTION_GROUP(group));

				GuiSignalConnect(action, "change-state",
						cff_gui_adjust_menu, &gVideo);

			} /* if (group == NULL) */

		} /* if (build_tab->adjust_menu) */

		g_menu_append_item(model, mb_title);

#else /* ! HAVE_GMENU */

		GuiSignalConnect(mb_title, "destroy",
					gtk_widget_destroyed, mbtitleptr);

		if (build_tab->adjust_menu)
		{
			GuiSignalConnect(mb_title, "activate",
					cff_gui_adjust_menu, &gVideo);
		}

		gtk_widget_realize(menu);
		gtk_menu_shell_append(GTK_MENU_SHELL(mbar), mb_title);

		*(mbtitleptr++) = mb_title;

#endif /* #ifdef HAVE_GMENU .. #else */

		GuiBuildMenuItems(build_tab->entries, menu, window, group);

		build_tab++;
	}
	while (++n < kMenuBarTotal);


	/* Popup Menu */

	menu = GuiMenuNew();

#ifdef HAVE_GMENU

	popup = GuiPopupMenuNewFromModel(G_MENU_MODEL(menu));

#if GTK_MAJOR_VERSION == 3
	group_name = CFF_GACTION_GROUP_STRING;
#else
	gtk_widget_set_parent(popup, window);

	GuiSignalConnectSwapped(window, "destroy",
				gtk_widget_unparent, popup);
	group_name = "pop";
#endif
	group = g_simple_action_group_new();
	gtk_widget_insert_action_group(popup,
			group_name, G_ACTION_GROUP(group));

#else /* ! HAVE_GMENU */

	popup = menu;
	gtk_widget_realize(menu);

#endif /* #ifdef HAVE_GMENU .. #else */


	GuiBuildMenuItems(cff_gui_menuPopup, menu, NULL, group);

#if GTK_MAJOR_VERSION == 2
	gtk_widget_add_events(window, GDK_BUTTON_PRESS_MASK);
#endif

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
	GuiSignalConnect(window, "popup-menu",
				cff_gui_sf10_show_popup_menu, popup);
#else
	gtk_popover_set_autohide(GTK_POPOVER(popup), TRUE);
	gtk_popover_set_has_arrow(GTK_POPOVER(popup), FALSE);
#endif

	gPopupMenuWidget = popup;

	GuiSignalConnect(popup, "show", cff_gui_adjust_popup, &gVideo);

	GuiSignalConnectMouseButton(window,
		kMouseButtonRight, kMouseButtonPress,
			cff_gui_mouse_button_show_popup_menu, popup);

#ifdef HAVE_GMENU

	/*
		If a modal dialog was opened, the state of the
		menubar item should be set to prelight (selected)
		The stuff below is NOT documented !!!
	*/

	n = kMenuBarFile;

#if GTK_MAJOR_VERSION == 3
	gtk_container_foreach(GTK_CONTAINER(mbar),
		(GtkCallback) cff_gui_build_set_menubar_title, &n);
#else
	gtk_window_set_handle_menubar_accel(GTK_WINDOW(window), FALSE);
	mbtitleptr = gMenuBarTitle;
	mbar_title = gtk_widget_get_first_child(mbar);

	while (mbar_title != NULL && n < kMenuBarTotal)
	{
		*(mbtitleptr++) = mbar_title;
		mbar_title = gtk_widget_get_next_sibling(mbar_title);
		n++;
	}
#endif

#endif /* HAVE_GMENU */

	return mbar;

} /* static GtkWidget *cff_gui_build_menu(GtkWidget*) */


/*
	------
	 INIT
	------
*/


static void cff_gui_noop(void)
{
	return;

} /* static void cff_gui_noop(void) */


#if ! defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2

static GdkGC *cff_gui_set_gdk_color(const GtkWidget *widget, const char *str)
{
	GdkGC    *context;
	GdkColor color;

	gdk_color_parse(str, &color);
	context	= gdk_gc_new((GdkDrawable *) widget->window);
	gdk_gc_set_rgb_fg_color(context, &color);

	return context;

} /* static GdkGC *cff_gui_set_gdk_color(const GtkWidget*, const char*) */

#endif /* ! defined(HAVE_CAIRO) && GTK_MAJOR_VERSION == 2 */


static GtkWidget *cff_gui_hbuttonbox_new(GtkWidget *hbox)
{
	GtkWidget *bbox, *frame;

	bbox = GuiHBoxNew(TRUE, 3);
	GuiContainerSetBorderWidth(bbox, 3);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, 0, bbox);
	GuiHBoxPackStart(hbox, frame, TRUE, FALSE, 3);

	return bbox;

} /* static GtkWidget *cff_gui_hbuttonbox_new(GtkWidget*) */


#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

static gint cff_gui_get_scale_margin(GtkWidget *scale)
{
	gint width;

	gtk_widget_style_get(scale, "slider-length", &width, NULL);
	width /= 2;

	if (width <= 0)
	{
		/* Default values */
		width = 31 / 2;
	}

	return width;

} /* static gint cff_gui_get_scale_margin(GtkWidget*) */

#endif /* GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 */


#ifdef HAVE_GMENU
void cff_gui_init(GtkApplication *app, const cff_init_param_t *param)
#else
void cff_gui_init(const cff_init_param_t *param)
#endif
{
	gint			i, width;
	GtkWidget		*button, *drawbar, *frame, *menu_bar,
				*bbox, *dbox, *hbox, *sbox, *vbox, *v1box,
				*widget, *window;
	cff_slider_t		*slider;
	cff_video_frame_t	*current_frame;


#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

	gint			new_width;

#ifndef G_OS_WIN32
	GdkPixbuf		*pixbuf;
#endif

#if GTK_MAJOR_VERSION == 3
	gboolean		minor_version_lt_20;
#endif

#endif /* #if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 */


#if GTK_MAJOR_VERSION == 2

#ifdef HAVE_CAIRO
	GdkColor		*color;
#else
	GdkGC			**gcontext;
#endif

#else /* GTK_MAJOR_VERSION != 2 */

	GdkRGBA			*color;

#endif /* #if GTK_MAJOR_VERSION == 2 .. #else */


	g_log_set_handler("Gdk", G_LOG_LEVEL_MASK,
					(GLogFunc) cff_gui_noop, NULL);

	/* Main window */

	gWindata.topwin = NULL;
	gWindata.slist  = NULL;

	vbox = GuiVBoxNew(FALSE, 0);

	window = cff_guitl_window_new_gettext(NULL, vbox, 2);
	gWindata.mainwin = window;
	cff_guitl_window_setup(window, kPropIntMainWinWidth);

#ifdef HAVE_GMENU
	gtk_application_add_window(app, GTK_WINDOW(window));
#endif

#ifdef G_OS_WIN32
	/* On Windows a new window seems always appear at top left */
	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
#endif

	GuiSignalConnect(window, gDeleteEventStr,
				cff_gui_file_quit_delevnt, NULL);

#ifndef HAVE_GMENU
	/* Terminate 'CutterFF' without 'GtkApplication' */
	GuiSignalConnect(window, "destroy", gtk_main_quit, NULL);
#endif

	/* Goto frame */
	cff_gui_signal_connect_key_pressed(window,
				(GCallback) cff_gui_video_frame_key, NULL);


#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

	gtk_widget_realize(window);
	cff_gui_mainwindow_cursor_new(window, NULL, &gCursor);

	GuiSignalConnect(window, "screen-changed",
				cff_gui_mainwindow_cursor_new, &gCursor);

#ifndef G_OS_WIN32

	/* Application icon */

	pixbuf = gdk_pixbuf_new_from_xpm_data((const char **) cutterff_xpm);
	gtk_window_set_icon(GTK_WINDOW(window), pixbuf);
#endif

#else /* GTK_MAJOR_VERSION == 4 */

	gCursor.window	= window;
	gCursor.busy	= gdk_cursor_new_from_name("wait", NULL);

#endif /* #if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 .. #else */


	/* Menu and popup */

	menu_bar = cff_gui_build_menu(window);
	GuiVBoxPackStart(vbox, menu_bar, FALSE, FALSE, 0);


	/* CSS */

#if GTK_MAJOR_VERSION == 3

	minor_version_lt_20 = FALSE;

	if (gtk_get_major_version() == 3 && gtk_get_minor_version() < 20)
	{
		minor_version_lt_20 = TRUE;

		/* Widget style provider */

		cff_gui_css_load_from_data(menu_bar,
				"GtkMenuBar,"
				"menubar {"
				"  -GtkMenuBar-shadow-type: None;"
				"}",
				GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);
	}

#elif GTK_MAJOR_VERSION == 4

	/* Global style provider */

	cff_gui_css_load_from_data(NULL,
				"*.dialog-action-area button {"
				"  min-width: 64px;"
				"  margin: 3px;"
				"  padding-left: 6px;"
				"  padding-right: 6px;"
				"}",
				GTK_STYLE_PROVIDER_PRIORITY_APPLICATION);

#endif /* #if GTK_MAJOR_VERSION == 3 #elif GTK_MAJOR_VERSION == 4 */


	current_frame = &gCurrentVideoFrame;


	/* Video picture */

	widget = gtk_drawing_area_new();
	current_frame->drawingArea = widget;

	gtk_widget_set_size_request(widget, 400, 225);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, 0, widget);
	GuiVBoxPackStart(vbox, frame, TRUE, TRUE, 5);

	GuiSignalConnect(widget, gSizeAllocationStr,
				cff_gui_draw_frame_sizealloc_cb, &gVideo);

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
	GuiSignalConnect(widget, gExposeEventStr,
				cff_gui_draw_frame, current_frame);
#else
	gtk_drawing_area_set_draw_func(GTK_DRAWING_AREA(widget),
			(GtkDrawingAreaDrawFunc) cff_gui_draw_frame,
							current_frame, NULL);
#endif


	/* Info Bar */

	hbox = GuiHBoxNew(FALSE, 0);
	GuiVBoxPackStart(vbox, hbox, FALSE, FALSE, 0);

	v1box = GuiVBoxNew(FALSE, 1);
	GuiHBoxPackStart(hbox, v1box, TRUE, TRUE, 8);

	hbox = GuiHBoxNew(TRUE, 2);
	GuiContainerSetBorderWidth(hbox, 1);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, 0, hbox);
	GuiVBoxPackStart(v1box, frame, FALSE, FALSE, 2);

	widget = gtk_label_new(NULL);
	current_frame->info = GTK_LABEL(widget);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 1);

	widget = gtk_label_new(NULL);
	current_frame->time = GTK_LABEL(widget);
	GuiHBoxPackStart(hbox, widget, FALSE, FALSE, 1);

	cff_gui_frame_info(current_frame, &gVideo);


	/* Slider Bar */

	dbox = GuiVBoxNew(FALSE, 0);
	GuiContainerSetBorderWidth(dbox, 5);
	frame = cff_guitl_frame_new_add_child_gettext(NULL, 0, dbox);
	GuiVBoxPackStart(v1box, frame, FALSE, FALSE, 2);

	drawbar = gtk_drawing_area_new();
	gDrawingAreaBar = drawbar;

	GuiSignalConnectMouseButton(drawbar,
		kMouseButtonLeft, kMouseButtonPress,
			cff_gui_draw_bar_button, &gVideo);

#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3
	gtk_widget_add_events(drawbar, GDK_BUTTON_PRESS_MASK);
	GuiSignalConnect(drawbar, gExposeEventStr,
					cff_gui_draw_bar, &gVideo);
#else
	gtk_drawing_area_set_draw_func(GTK_DRAWING_AREA(drawbar),
		(GtkDrawingAreaDrawFunc) cff_gui_draw_bar, &gVideo, NULL);
#endif

	slider = &gSlider;
	slider->scale = GuiHScaleNew(NULL);

	gtk_scale_set_draw_value(GTK_SCALE(slider->scale), FALSE);
	gtk_widget_set_direction(slider->scale, GTK_TEXT_DIR_LTR);
	gtk_range_set_restrict_to_fill_level(GTK_RANGE(slider->scale), FALSE);
	slider->adjust = gtk_range_get_adjustment(GTK_RANGE(slider->scale));

	/* Specifies whether (or not) a widget can own the input focus */
	GuiWidgetSetCanFocus(slider->scale, FALSE);
	gtk_widget_set_receives_default(slider->scale, FALSE);


#if GTK_MAJOR_VERSION == 2

	width = cff_gui_get_scale_margin(slider->scale);

#else /* GTK_MAJOR_VERSION > 2 */

#if GTK_MAJOR_VERSION == 3

	if (minor_version_lt_20)
	{
		width = cff_gui_get_scale_margin(slider->scale);
	}

	else
#endif
	{

	/*
		GTK-3 + GTK-4:
		Don't know how to get the style properties,
		so I create here my own scale slider!
		The center of the slider should match the
		drawbar (start, stop, cutpoints) above.
	*/

	width = 10;

	cff_gui_css_load_from_data(slider->scale,
				"scale,"
				"scale.horizontal {"
				"  min-height: 0px;"
				"  min-width:  0px;"
				"  border:  0px;"
				"  margin: 10px;"
				"  padding: 0px;"
#if GTK_MAJOR_VERSION == 4
				"  padding-top: 5px;"
#endif
				"}"
				"scale slider,"
				"scale.horizontal slider {"
				"  min-height: 20px;"
				"  min-width:  20px;"
				"}"
				"scale trough,"
				"scale.horizontal trough {"
				"  min-height: 6px;"
				"  border:  0px;"
				"  margin:  0px;"
				"  padding: 0px;"
				"}",
				GTK_STYLE_PROVIDER_PRIORITY_USER);

	} /* if (minor_version_lt_20) .. else */

#endif /* #if GTK_MAJOR_VERSION == 2 .. #else */


	sbox = GuiHBoxNew(FALSE, 0);
	GuiContainerSetBorderWidth(sbox, width / 2);

	gtk_widget_set_size_request(drawbar, -1, 16);
	GuiHBoxPackStart(sbox, drawbar, TRUE, TRUE, width / 2);

	GuiVBoxPackStart(dbox, sbox, FALSE, FALSE, 0);
	GuiVBoxPackStart(dbox, slider->scale, FALSE, FALSE, 0);

	cff_gui_slider_adjustment_configure(0);

	slider->button    = 0;
	slider->is_slider = FALSE;

	GuiSignalConnectMouseButton(slider->scale,
		kMouseButtonLeft, kMouseButtonPress | kMouseButtonRelease,
						cff_gui_slider_button, slider);

	GuiSignalConnect(slider->scale, "change-value",
				cff_gui_video_frame_slider, slider);

#if GTK_MAJOR_VERSION == 2

#ifdef HAVE_CAIRO
	color = &gColor[kColorBlack];
	gdk_color_parse("Black",     color++);
	gdk_color_parse("DarkGrey",  color++);
	gdk_color_parse("LightGrey", color++);
	gdk_color_parse("Green",     color++);
	gdk_color_parse("Red",       color);
#else
	gcontext = &gContext[kColorBlack];
	*(gcontext++) = cff_gui_set_gdk_color(window, "Black");
	*(gcontext++) = cff_gui_set_gdk_color(window, "DarkGrey");
	*(gcontext++) = cff_gui_set_gdk_color(window, "LightGrey");
	*(gcontext++) = cff_gui_set_gdk_color(window, "Green");
	*gcontext     = cff_gui_set_gdk_color(window, "Red");
#endif
#else
	color = &gRGBA[kColorBlack];
	gdk_rgba_parse(color++, "Black");
	gdk_rgba_parse(color++, "DarkGrey");
	gdk_rgba_parse(color++, "LightGrey");
	gdk_rgba_parse(color++, "Green");
	gdk_rgba_parse(color,   "Red");

#endif /* #if GTK_MAJOR_VERSION == 2 .. #else */


	/* Button Bar */

	hbox = GuiHBoxNew(FALSE, 0);
	GuiVBoxPackStart(v1box, hbox, FALSE, FALSE, 5);

	bbox = cff_gui_hbuttonbox_new(hbox);


	/* Goto frame */

	i = 0;

	do
	{
		button = gtk_button_new_with_label(gButtonVideoFrameLabel[i]);
		gui_button_video_frame[i] = button;
		GuiHBoxPackStart(bbox, button, TRUE, TRUE, 1);
		GuiWidgetSetCanFocus(button, FALSE);

		GuiSignalConnectMouseButton(button,
			kMouseButtonAny, kMouseButtonPress,
				cff_gui_video_frame_button, GINT_TO_POINTER(i));
	}
	while (++i < kButtonVideoFrameTotal);


	/* Padding */

	bbox = GuiHBoxNew(FALSE, 0);
	GuiHBoxPackStart(hbox, bbox, FALSE, FALSE, 5);


	/* Add or remove cutpoints */

	bbox = cff_gui_hbuttonbox_new(hbox);

	i = kButtonVideoCutAdd;

	for (;;)
	{
		const gchar *label;
		GCallback   func;

		if (i == kButtonVideoCutAdd)
		{
			label = "+";
			func  = (GCallback) GuiEditAddCutFunc;
		}

		else
		{
			label = "-";
			func  = (GCallback) GuiEditRemoveCutFunc;
		}
		
		button = cff_guitl_hbox_add_button(bbox, label, func, &gVideo);
		gButtonVideoCut[i] = button;
		GuiWidgetSetCanFocus(button, FALSE);

		/* Ignore this event if busy */
		cff_gui_signal_connect_mouse_button_ignore(button);

		if (i == kButtonVideoCutAdd)
		{
			i = kButtonVideoCutRemove;
		}

		else
		{
			break;
		}

	} /* for (;;) */

	
	/* Test filter extensions */

	cff_gui_file_filter_validate_prop(kPropStrFileVideoFilter);
	cff_gui_file_filter_validate_prop(kPropStrFileImageFilter);


	/* Open a thread (default: gThreads.max == 0 => disable */

	if (param->threads == 0)
	{
		/* Auto => limited by number of CPUs */
		gThread.max = UINT_MAX;
	}

	else if (param->threads > 1)
	{
		/* Max */
		gThread.max = param->threads;
	}

	/* This is the main thread */
	gThread.num = 1;

	cff_gui_thread_new(NULL, NULL, NULL);


	if (param->fname != NULL)
	{
		/* Log window is on the top */
		GuiWidgetShowAll(window);
	}


	/* Open the log window if level is not 'quiet' */

	if (param->loglevel == NULL)
	{
		gchar		*logname;
		const gchar	*logquiet;

		logquiet = cff_ffmpeg_get_loglevel_name(0);
		logname  = cff_gprop_get_str(kPropStrOptLogLevel);

		if (logname != NULL)
		{
			if (strcmp(logname, logquiet) != 0)
			{
				cff_gui_logwindow_open(logname);
			}

			g_free(logname);
		}

	} /* if (param->loglevel == NULL) */

	if (param->fname == NULL)
	{
		/* Main window is on the top */
		GuiWidgetShowAll(window);
		cff_gui_video_button_set_sensitive(FALSE);
		cff_guitl_window_set_title_filename(window, NULL);
	}

	else
	{
		GFile *file;

		/*
			On Windows, this function expects its
			argument to be in UTF-8 (see 'cff_main.c')
		*/

		file = g_file_new_for_commandline_arg(param->fname);

#ifdef G_OS_WIN32
		/* dup'ed in 'cff_main.c' */
		g_free((gchar *) param->fname);
#endif
		/* Open the file */
		g_idle_add_full(G_PRIORITY_LOW,
			(GSourceFunc) cff_gui_video_open, file, NULL);

	} /* if (param->fname == NULL) .. else */


#if GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3

#if GTK_MAJOR_VERSION == 3
	if (minor_version_lt_20)
#endif
	{
		new_width = cff_gui_get_scale_margin(slider->scale);

		if (new_width != width)
		{
			new_width /= 2;
			GuiContainerSetBorderWidth(sbox, new_width);
			gtk_box_set_spacing(GTK_BOX(sbox), new_width);
		}

	} /* if (minor_version_lt_20) */

#endif /* GTK_MAJOR_VERSION == 2 || GTK_MAJOR_VERSION == 3 */

} /* void cff_gui_init(const cff_init_param_t*) */

/* EOF */
